/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include "action_pl.h"
#include "object.h"
#include "view.h"


ActionPoint::ActionPoint(const String &name, Object *object)
    : the_object(object), the_name(name)
{
}

ActionPoint::~ActionPoint()
{
    angle_lines.deleteAll();
}

Object *ActionPoint::object() const
{
    return the_object;
}

String ActionPoint::name() const
{
    return the_name;
}

Metrum3 ActionPoint::pos() const
{
    return the_object->handlePos(this);
}

List<ActionLine *> ActionPoint::lines()
{
    if (angle_lines.isEmpty())
	angle_lines << new AngleActionLine(0, this)
		    << new AngleActionLine(90, this);

    List<ActionLine *> result = the_object->handleLines(this);
    foreach (ActionLine *angle_line, angle_lines)
	result << angle_line;

    return result;
}

void ActionPoint::move(const Metrum3 &pos)
{
    the_object->moveHandle(pos, this);
}

/****************************************************************************/

ScreenDist ActionLine::distance(const ScreenPos &p, View *view) const
{
    ScreenDist parameter, distance;

    p.computeParameters(view->map(base()),
			view->map(tangent()), parameter, distance);

    return distance;
}

Metrum3 ActionLine::nearest(const ScreenPos &p, View *view) const
{
    ScreenPos bs = view->map(base());
    ScreenPos ts = view->map(tangent());
    ScreenDist parameter, distance;

    p.computeParameters(bs, ts, parameter, distance);

    return base() + tangent() * parameter / ts.length();
}

bool ActionLine::doesIntersect(ActionLine *line) const
{
    Metrum3 t0 = tangent();
    Metrum3 t1 = line->tangent();

    if (tangent().isNull() || line->tangent().isNull())
	return false;

    /* Are they not parallel [i.e. cross product (t0, t1) != 0] ? */

    if (t0.y() * t1.z() == t0.z() * t1.y() &&
	t0.z() * t1.x() == t0.x() * t1.z() &&
	t0.x() * t1.y() == t0.y() * t1.x())
	return false;

    /* Are they in a plane [i.e. triple product (t0, t1, b1 - b0) = 0] ? */

    Metrum3 b = line->base() - base();

    return b.x() * (t0.y() * t1.z() - t0.z() * t1.y()) +
	   b.y() * (t0.z() * t1.x() - t0.x() * t1.z()) +
	   b.z() * (t0.x() * t1.y() - t0.y() * t1.x()) == 0;
}

Metrum3 ActionLine::intersection(ActionLine *line) const
{
    /* assuming someone already has verified that they _do_ intersect */

    /* (x y z) = norm (cross product (t0, t1)) */

    Metrum3 t0 = tangent();
    Metrum3 t1 = line->tangent();

    double x = t0.y() * t1.z() - t0.z() * t1.y();
    double y = t0.z() * t1.x() - t0.x() * t1.z();
    double z = t0.x() * t1.y() - t0.y() * t1.x();

    double l = sqrt(x * x + y * y + z * z);
    x /= l; y /= l; z /= l;

    /* return b0 + t0 * triple product (b1 - b0, t1, x y z)
		      / triple product (t0,      t1, x y z) */

    Metrum3 b = line->base() - base();

    double u = x * (b.y() * t1.z() - b.z() * t1.y()) +
	       y * (b.z() * t1.x() - b.x() * t1.z()) +
	       z * (b.x() * t1.y() - b.y() * t1.x());

    double v = x * (t0.y() * t1.z() - t0.z() * t1.y()) +
	       y * (t0.z() * t1.x() - t0.x() * t1.z()) +
	       z * (t0.x() * t1.y() - t0.y() * t1.x());

    return base() + (u / v) * tangent();
}

/****************************************************************************/

AngleActionLine::AngleActionLine(int angle, ActionPoint *point)
    : ActionLine(), the_point(point), the_angle(angle)
{
}

String AngleActionLine::name() const
{
    if (the_angle == 0)
	return _("Horizontal");
    if (the_angle == 90)
	return _("Vertical");

    return _("Guide at %1 degrees").arg(int_to_str(the_angle));
}

Metrum3 AngleActionLine::base() const
{
    return the_point->pos();
}

Metrum3 AngleActionLine::tangent() const
{
    return Metrum3(llrint(cos(the_angle * M_PI / 180)),
		   llrint(sin(the_angle * M_PI / 180)), 0);
}

/****************************************************************************/

ObjectActionLine::ObjectActionLine(const String &name, Object *object)
    : ActionLine(), the_object(object), the_name(name)
{
}

String ObjectActionLine::name() const
{
    return the_name;
}

Metrum3 ObjectActionLine::base() const
{
    return the_object->lineBase(this);
}

Metrum3 ObjectActionLine::tangent() const
{
    return the_object->lineTangent(this);
}
