/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "gen_linear.h"
#include "hub.h"
#include "icons.h"
#include "drawing.h"
#include "action.h"
#include "painter.h"
#include "view.h"


LinearMajorMode::LinearMajorMode(Hub &hub) : MajorMode(hub),
					     startMinorMode(hub, *this),
					     endMinorMode(hub, *this)
{
    setMinorMode(&startMinorMode);

/** comboBox = new QComboBox;
    comboBox->addItem(_("Continuous lines"));
    comboBox->addItem(_("Shared first point"));

    QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(comboBox);
    layout->addStretch(); **/
}

void LinearMajorMode::accept(MinorMode *minor_mode)
{
    if (minor_mode == &startMinorMode)
	setMinorMode(&endMinorMode);

    if (minor_mode == &endMinorMode) {
	Object *object = buildObject(startMinorMode.point(),
				     endMinorMode.point());
	hub.drawing()->performActions(
	    List<Action *>() << new NewObjectAction(object),
	    icons::add, _("New %1").arg(object->name()));

	if (endMinorMode.released)
	    cancel();
	else if (/** comboBox->currentIndex() **/ 0 == 0) {
	    startMinorMode.setPoint(endMinorMode.point());
	    setMinorMode(&endMinorMode);
	}
    }
}

void LinearMajorMode::cancel()
{
    setMinorMode(&startMinorMode, true);
}

void LinearMajorMode::paint(Painter &painter)
{
    MajorMode::paint(painter);

    painter.setPen(0x7f, 0x7f, 0x7f);
    if (minorMode() == &endMinorMode && hub.pointerActive())
	painter.paintLine(startMinorMode.point(), endMinorMode.point());

    painter.setBlackPen();
    if (minorMode() == &endMinorMode || hub.pointerActive())
	painter.paintPoint(startMinorMode.point());
    if (minorMode() == &endMinorMode && hub.pointerActive())
	painter.paintPoint(endMinorMode.point());
}

List<ActionLine *> LinearMajorMode::lines(ActionPoint *handle)
{
    return List<ActionLine *>();
}

/****************************************************************************/

LinearObject::LinearObject(const Metrum3 &p0, const Metrum3 &p1,
			   Drawing *drawing)
    : Object(drawing), p0(p0), p1(p1),
      startHandle(_("endpoint"), this),
      midpointHandle(_("midpoint"), this),
      endHandle(_("endpoint"), this),
      startNormal(_("endpoint normal"), this),
      midpointNormal(_("midpoint normal"), this),
      endNormal(_("endpoint normal"), this),
      tangent(_("tangent"), this)
{
}

void LinearObject::paint(Painter &painter, const Embedding &embedding)
{
    painter.setBlackPen();
    painter.paintLine(embedding.transform(p0), embedding.transform(p1));
}

void LinearObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			       Painter &painter)
{
    Metrum3 hp0(p0), hp1(p1);

    if (handle == &startHandle)
	hp0 = pos;
    if (handle == &midpointHandle) {
	Metrum3 delta = p1 - p0;
	hp0 = pos - delta / 2;
	hp1 = hp0 + delta;
    }
    if (handle == &endHandle)
	hp1 = pos;

    painter.setPen(0x80, 0x80, 0x80);
    painter.paintLine(hp0, hp1);
}

List<ActionPoint *> LinearObject::handles()
{
    return List<ActionPoint *>() << &startHandle
				 << &midpointHandle
				 << &endHandle;
}

Metrum3 LinearObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &startHandle)
	return p0;
    if (handle == &midpointHandle)
	return (p0 + p1) / 2;
    if (handle == &endHandle)
	return p1;

    return Metrum3();
}

List<ActionLine *> LinearObject::handleLines(const ActionPoint *handle)
{
    if (handle == &startHandle)
	return List<ActionLine *>() << &tangent << &startNormal;

    if (handle == &midpointHandle)
	return List<ActionLine *>() << &tangent << &midpointNormal;

    if (handle == &endHandle)
	return List<ActionLine *>() << &tangent << &endNormal;

    return List<ActionLine *>();
}

void LinearObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &startHandle)
	p0 = pos;
    if (handle == &midpointHandle) {
	Metrum3 delta = p1 - p0;
	p0 = pos - delta / 2;
	p1 = p0 + delta;
    }
    if (handle == &endHandle)
	p1 = pos;
}

Metrum3 LinearObject::lineBase(const ActionLine *line) const
{
    if (line == &startNormal)
	return p0;
    if (line == &midpointNormal || line == &tangent)
	return (p0 + p1) / 2;
    if (line == &endNormal)
	return p1;

    return Metrum3();
}

Metrum3 LinearObject::lineTangent(const ActionLine *line) const
{
    if (line == &tangent)
	return p1 - p0;

    if (line == &startNormal ||
	line == &midpointNormal ||
	line == &endNormal)
	return Metrum3((p1 - p0).y(), (p0 - p1).x(), 0);

    return Metrum3(1, 0, 0);
}

ScreenDist LinearObject::proximity(const ScreenPos &point, View *view,
				   const Embedding &embedding) const
{
    ScreenPos s0 = view->map(embedding.transform(p0));
    ScreenPos s1 = view->map(embedding.transform(p1));
    ScreenDist parameter, distance;

    point.computeParameters(s0, s1 - s0, parameter, distance);

    if (parameter < 0)
	return (point - s0).length();

    if (parameter > (s1 - s0).length())
	return (point - s1).length();

    return distance;
}

bool LinearObject::within(const ScreenRect &rect, View *view,
			  bool partially, const Embedding &embedding) const
{
    ScreenPos s0 = view->map(embedding.transform(p0));
    ScreenPos s1 = view->map(embedding.transform(p1));

    if (!partially)
	return rect.contains(s0) && rect.contains(s1);

    return rect.contains(s0) || rect.contains(s1); //TODO
}
