/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <strings.h>
#include "instance.h"
#include "util/c++11.h"
#include "hub.h"
#include "icons.h"
#include "shared.h"
#include "dirmodel.h"
#include "object.h"
#include "action.h"
#include "view.h"
#include "majormode.h"
#include "views.h"
#include "actionsets.h"
#include "tabs.h"
#include "viewport.h"

using namespace Toolkit;


class Instance;

class InstanceHub final : public Hub
{
    Instance *const p;
    void updateHandle();
    void updateDrawingStatus();
    void updateView();
    void updateMajorMode();
    void updateMinorMode();
    void updateSelectedObjects();
    void updateStatusMessage();
    void updateProjectDirectories();
public:
    InstanceHub(Instance *p);
    ~InstanceHub();
    bool pointerActive() const;
    ScreenPos pointer() const;
    void requestClose();
    void update();

    String selectedLibraryComponent() const;
};


class Instance final : public ActionSet, Destructible
{
public:
    Instance(Connection &conn);
    void actionTriggered(Toolkit::Action *action);

    InstanceHub hub;

    ProjectDirView project_dir_view;
    ScrollPane project_dir_pane;
    VContainer project_dir_container;

    ObjectsView objects_view;
    ScrollPane objects_pane;

    HistoryView1 history_view;
    ScrollPane history_pane;
    VContainer history_container;

    ToolboxView toolbox_view;
    ScrollPane toolbox_pane;

    LibraryView library_view;
    ScrollPane library_pane;

    ScrollPane properties_pane;

    FileActionSet file_action_set;
    DrawingActionSet drawing_action_set;
    PropertiesActionSet properties_action_set;
    PopupActionSet popup_action_set;

    Toolkit::Action show_origin_action;
    Toolkit::Action show_grid_action;
    Toolkit::Action crosshairs_action;
    Toolkit::Action fullscreen_action;
    Toolkit::Action tab_bar_action;
    Toolkit::Action docks_action;
    Toolkit::Action new_window_action;
    Toolkit::Action close_window_action;
    Toolkit::Action quit_action;

    VContainer root_widget;
    MenuBarBuilder menu_bar_builder;
    Docks::Manager dock_manager;
    VContainer main_widget;
    MainTabBar tab_bar;
    ScrollPane pane;
    Viewport viewport;
    ViewportNavigator viewport_navigator;
    StatusBar status_bar;

    Docks::Dock project_dir_dock, objects_dock, history_dock,
		mode_dock, library_dock, properties_dock;

    /* must be defined after all actions, action sets and docks */
    ToolbarBuilder project_dir_toolbar_builder;
    ToolbarBuilder history_toolbar_builder;
    MenuBuilder file_menu_builder;
    MenuBuilder edit_menu_builder;
    MenuBuilder view_menu_builder;
    MenuBuilder docks_menu_builder;
    MenuBuilder popup_menu_builder;
    MenuBuilder properties_menu_builder;

    /* properties_menu_builder passed at construction time */
    MyPropertiesView properties_view;

    InputWindowHook hook;
};

class InitializationTimer : public Timer
{
    Instance *const p;
    int tick;
public:
    InitializationTimer(Instance *p);
    void timeout();
};

static X::Window make_window(Connection &conn, int x, int y,
			     unsigned int width, unsigned int height,
			     unsigned int border_width, int argc, char *argv[])
{
    X::Window window;
    Atom protocols[] = { conn.atoms().WM_DELETE_WINDOW };

    window = conn.screen().rootWindow().createWindow(
	x, y, width, height, border_width, conn.visualInfo().depth,
	InputOutput, conn.visual(), 0, NULL);

    window.utf8SetWMProperties(_("Xi"), _("Xi"), argv, argc, 0, 0, 0);

    window.setWMProtocols(protocols, 1);

    window.selectInput(KeyPressMask | KeyReleaseMask |
		       ButtonPressMask | ButtonReleaseMask |
		       EnterWindowMask | LeaveWindowMask |
		       PointerMotionMask | ExposureMask |
		       VisibilityChangeMask | StructureNotifyMask |
		       FocusChangeMask);
    window.map();

    return window;
}

Instance::Instance(Connection &conn)
    : hub(this),

      project_dir_view(conn, hub),
      project_dir_pane(conn),
      project_dir_container(conn),

      objects_view(conn, hub),
      objects_pane(conn),

      history_view(conn, hub),
      history_pane(conn),
      history_container(conn),

      toolbox_view(conn, hub),
      toolbox_pane(conn),

      library_view(conn, hub, 0),
      library_pane(conn),

      properties_pane(conn),

      file_action_set(hub, conn),
      drawing_action_set(hub),
      properties_action_set(hub, properties_view),
      popup_action_set(hub),

      show_origin_action(*this), show_grid_action(*this),
      crosshairs_action(*this),
      fullscreen_action(*this), tab_bar_action(*this),

      docks_action(*this),
      new_window_action(*this), close_window_action(*this), quit_action(*this),

      root_widget(conn),
      menu_bar_builder(conn),
      dock_manager(conn),
      main_widget(conn),
      tab_bar(conn, hub, tab_bar_action),
      pane(conn),
      viewport(conn, hub),
      viewport_navigator(conn, hub, 1000.),
      status_bar(conn),

      project_dir_dock(
	  dock_manager, _("Project directories"), icons::tree_leaf,
	      &project_dir_container, "project-dirs",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),
      objects_dock(
	  dock_manager, _("Objects"), icons::tree_leaf,
	      &objects_pane, "objects",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),
      history_dock(
	  dock_manager, _("History"), icons::tree_leaf,
	      &history_container, "history",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),
      mode_dock(
	  dock_manager, _("Modes"), icons::tools, &toolbox_pane, "modes",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),
      library_dock(
	  dock_manager, _("Library"), icons::tree_leaf,
	      &library_pane, "library",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),
      properties_dock(
	  dock_manager, _("Properties"), icons::tree_leaf,
	      &properties_pane, "properties",
	  conn.screen().blackPixel(), conn.colors().viewBackground.pixel),

      project_dir_toolbar_builder(conn),
      history_toolbar_builder(conn),
      file_menu_builder(conn),
      edit_menu_builder(conn),
      view_menu_builder(conn),
      docks_menu_builder(conn),
      popup_menu_builder(conn),
      properties_menu_builder(conn),
      properties_view(conn, properties_menu_builder.menu()),

      hook(conn, make_window(conn, 0, 0, 1080, 720, 1, 0, 0), this,
		 0, 0, 1080, 720, 1)
{
    project_dir_toolbar_builder.addAction(file_action_set.newFileAction);
    project_dir_toolbar_builder.addAction(file_action_set.newFolderAction);
    project_dir_toolbar_builder.addSeparator();
    project_dir_toolbar_builder.addAction(file_action_set.saveAction);
    project_dir_toolbar_builder.addAction(file_action_set.saveAllAction);
    project_dir_toolbar_builder.addSpacer();
    project_dir_toolbar_builder.addAction(file_action_set.commitAction);

    project_dir_view.setRootInterior(shared.projectDirs());
    project_dir_pane.mainControl()->setWidget(&project_dir_view);
    project_dir_container.addWidget(&project_dir_toolbar_builder.toolbar);
    project_dir_container.addWidget(&project_dir_pane);

    objects_pane.mainControl()->setWidget(&objects_view);

    history_toolbar_builder.addAction(drawing_action_set.undoAction);
    history_toolbar_builder.addAction(drawing_action_set.redoAction);

    history_pane.mainControl()->setWidget(&history_view);
    history_container.addWidget(&history_toolbar_builder.toolbar);
    history_container.addWidget(&history_pane);

    toolbox_view.setRootInterior(hub.toolbox());
    toolbox_pane.mainControl()->setWidget(&toolbox_view);

    library_view.setRootInterior(shared.library());
    library_pane.mainControl()->setWidget(&library_view);

    properties_menu_builder.addAction(properties_action_set.selectAllAction);
    properties_menu_builder.addAction(properties_action_set.resetAction);
    properties_view.setRootInterior(hub.attributes());
    properties_pane.mainControl()->setWidget(&properties_view);

    hook.registerActionSet(&file_action_set);
    hook.registerActionSet(&drawing_action_set);
    hook.registerActionSet(&properties_action_set);
    hook.registerActionSet(&popup_action_set);
    hook.registerActionSet(tab_bar.actionSet());
    hook.registerActionSet(this);

    show_origin_action.setText(_("Show o&rigin"));
    show_origin_action.setHotkey(_("o o"));

    show_grid_action.setText(_("Show &grid"));
    show_grid_action.setHotkey(_("o g"));

    crosshairs_action.setText(_("&Crosshairs cursor"));
    crosshairs_action.setHotkey(_("o c"));
    crosshairs_action.setIcon(icons::crosshairs);

    fullscreen_action.setText(_("&Fullscreen"));
    fullscreen_action.setHotkey(_("F11"));
    fullscreen_action.setIcon(icons::fullscreen);

    tab_bar_action.setText(_("Show &tab bar"));
    tab_bar_action.setHotkey(_("o t"));
    tab_bar_action.setActive(true);

    docks_action.setText(_("&Dock windows"));
    docks_action.setMenu(docks_menu_builder.menu());

    new_window_action.setText(_("&New window"));
    new_window_action.setHotkey(_("C-n"));

    close_window_action.setText(_("&Close window"));

    quit_action.setText(_("&Quit"));
    quit_action.setHotkey(_("C-q"));
    quit_action.setIcon(icons::quit);

    file_menu_builder.addAction(file_action_set.recentFilesAction);
    file_menu_builder.addSeparator();
    file_menu_builder.addAction(file_action_set.saveAction);
    file_menu_builder.addAction(file_action_set.printAction);
    file_menu_builder.addAction(file_action_set.sendAction);
    file_menu_builder.addSeparator();
    file_menu_builder.addAction(new_window_action);
    file_menu_builder.addAction(close_window_action);
    file_menu_builder.addSeparator();
    file_menu_builder.addAction(quit_action);

    edit_menu_builder.addAction(drawing_action_set.undoAction);
    edit_menu_builder.addAction(drawing_action_set.redoAction);

    view_menu_builder.addAction(drawing_action_set.zoomInAction);
    view_menu_builder.addAction(drawing_action_set.zoomOutAction);
    view_menu_builder.addSeparator();
    view_menu_builder.addAction(show_origin_action);
    view_menu_builder.addAction(show_grid_action);
    view_menu_builder.addAction(crosshairs_action);
    view_menu_builder.addSeparator();
    view_menu_builder.addAction(fullscreen_action);
    view_menu_builder.addSeparator();
    view_menu_builder.addAction(tab_bar_action);
    view_menu_builder.addAction(docks_action);

    project_dir_dock.action.setHotkey(_("F3"));
    hook.registerActionSet(&project_dir_dock);

    docks_menu_builder.addAction(project_dir_dock.action);
    docks_menu_builder.addAction(objects_dock.action);
    docks_menu_builder.addAction(history_dock.action);
    docks_menu_builder.addAction(mode_dock.action);
    docks_menu_builder.addAction(library_dock.action);
    docks_menu_builder.addAction(properties_dock.action);

    popup_menu_builder.addAction(popup_action_set.deselectModeAction);

    menu_bar_builder.addMenu(_("&File"), file_menu_builder.menu());
    menu_bar_builder.addMenu(_("&Edit"), edit_menu_builder.menu());
    menu_bar_builder.addMenu(_("&View"), view_menu_builder.menu());

    viewport.contextMenu = popup_menu_builder.menu();

    pane.mainControl()->setWidget(&viewport);
    pane.navigatorControl()->setWidget(&viewport_navigator);

    main_widget.addWidget(&tab_bar);
    main_widget.addWidget(&pane);

    dock_manager.mainControl()->setWidget(&main_widget);
    dock_manager.setConfiguration(
	"(((project-dirs) 0.33, (modes) 0.33, (library) 0.33) 0.25) "
	"(((objects) 0.5, (history, properties) 0.5) 0.25) () () ()");

    root_widget.addWidget(menu_bar_builder.menuBar());
    root_widget.addWidget(dock_manager.widget());
    root_widget.addWidget(&status_bar);

    hook.control()->setWidget(&root_widget);
    hook.setFocusHandler(dock_manager.focusHandler());
    hook.setStatusBar(&status_bar);

    new InitializationTimer(this);
}

static void toggle_fullscreen(Connection &conn, X::Window window)
{
    XEvent xev;
    bzero(&xev, sizeof xev);
    xev.type = ClientMessage;
    xev.xclient.window = window;
    xev.xclient.message_type = conn.atoms().ewmh.WM_STATE;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 2;
    xev.xclient.data.l[1] = conn.atoms().ewmh.WM_STATE_FULLSCREEN;
    xev.xclient.data.l[2] = 0;

    XSendEvent(conn.display(),
	       DefaultRootWindow(static_cast<Display *>(conn.display())),
	       False, SubstructureNotifyMask, &xev);
}

void Instance::actionTriggered(Toolkit::Action *action)
{
    if (action == &show_origin_action) {
	bool flag = !viewport.showOrigin();
	viewport.setShowOrigin(flag);
	show_origin_action.setActive(flag);
	return;
    }
    if (action == &show_grid_action) {
	bool flag = !viewport.showGrid();
	viewport.setShowGrid(flag);
	show_grid_action.setActive(flag);
	return;
    }
    if (action == &crosshairs_action) {
	bool flag = !viewport.crosshairsCursor();
	viewport.setCrosshairsCursor(flag);
	crosshairs_action.setActive(flag);
	return;
    }
    if (action == &fullscreen_action) {
	toggle_fullscreen(hook.conn, hook.window());
	return;
    }
    if (action == &tab_bar_action) {
	if (dock_manager.mainControl()->widget() != &main_widget) {
	    dock_manager.mainControl()->setWidget(&main_widget);
	    main_widget.addWidget(&pane);
	    tab_bar_action.setActive(true);
	} else {
	    main_widget.removeWidget(&pane);
	    dock_manager.mainControl()->setWidget(&pane);
	    tab_bar_action.setActive(false);
	}
	return;
    }
    if (action == &new_window_action) {
	instantiate(hook.conn);
	return;
    }
    if (action == &close_window_action) {
	hub.requestClose();
	return;
    }
    if (action == &quit_action) {
	shared.requestQuit();
	return;
    }

/* viewport.showGrid()
   viewport.useCrosshairsCursor()
   window_action_set.tabBarAction.active()
   tab_bar.action_set.insert_after_current_tab
   geometry
   file_action_set.recentFiles() */
}

InitializationTimer::InitializationTimer(Instance *p) : p(p), tick(0)
{
    set(0, 0, false);
}

void InitializationTimer::timeout()
{
    String message;

    if (shared.moreInitialization(message))
	p->status_bar.setMessage(message);
    else
	switch (tick++) {
	case 0:
	    p->status_bar.setMessage(_("Loading recent file..."));
	    break;
	case 1:
	    p->file_action_set.visitSomeFile();
	    p->status_bar.setMessage(_("Setting up user interface..."));
	    break;
	case 2:
	    p->hub.deselectMajorMode();
	    p->status_bar.setMessage(_("Ready"));
	    /* fallthrough */
	default:
	    delete this;
	    return;
	}

    set(0, 0, false);
}

/****************************************************************************/

InstanceHub::InstanceHub(Instance *p) : p(p)
{
}

InstanceHub::~InstanceHub()
{
}

void InstanceHub::updateHandle()
{
    String title = _("%1 - Xi").arg(handle()->path().split('/').last());
    p->hook.window().utf8SetWMProperties(title, title, 0, 0, 0, 0, 0);

    p->project_dir_view.scrollTo(p->project_dir_view.selected());
    p->project_dir_view.invalidate();
    p->tab_bar.selectNewTab(handle());
    p->file_action_set.handleChanged();
    p->objects_view.setRootInterior(drawing()->rootContext()->interior());
    p->history_view.setRootInterior(&drawing()->historyInterior);

    updateDrawingStatus();
    updateView();
}

void InstanceHub::updateDrawingStatus()
{
    std::vector<Item *> path;
    for (ActionGroup *ag = drawing()->mostRecent(); ag; ag = ag->parent())
	path.insert(path.begin(), ag);
    p->history_view.setLatestShadow(
	static_cast<HistoryViewShadow *>(
	    p->history_view.rootShadow()->lookup(path)));

    p->history_view.scrollTo(p->history_view.selected());
    p->history_view.invalidate();

    p->file_action_set.drawingContentsChanged();
    p->drawing_action_set.updateActions();
    p->viewport.invalidate();
}

void InstanceHub::updateView()
{
    p->viewport.requestLayout();

    /* This is deferred, so the viewport dimensions will not be set
     * until the next iteration of the event loop.  This may me too
     * late (since there may be more scrollbar drag events queued), so
     * we need to set the viewport dimensions explicitly here. */

    view()->setViewportDimensions(p->viewport.width(),
				  p->viewport.height());
}

void InstanceHub::updateMajorMode()
{
    p->toolbox_view.scrollTo(p->toolbox_view.selected());
    p->toolbox_view.invalidate();

    if (majorMode() == componentMajorMode()) {
	p->dock_manager.setFocussedDock(&p->library_dock);

	if (p->library_view.selected())
	    p->library_view.scrollTo(p->library_view.selected());
    }
    p->library_view.invalidate();
}

void InstanceHub::updateMinorMode()
{
    p->viewport.updateMinorMode();
}

void InstanceHub::updateSelectedObjects()
{
    p->objects_view.invalidate();
    p->viewport.invalidate();
}

void InstanceHub::updateStatusMessage()
{
    p->status_bar.setMessage(majorMode()->minorMode()->statusMessage());
}

void InstanceHub::updateProjectDirectories()
{
    // TODO
}

bool InstanceHub::pointerActive() const
{
    return p->viewport.pointerActive();
}

ScreenPos InstanceHub::pointer() const
{
    return p->viewport.pointer();
}

void InstanceHub::requestClose()
{
/** List<Handle> unsaved_handles;
    List<String> list;

#ifdef RELEASE
    foreach (Handle handle, handles)
	if (handle.drawing()->isModified()) {
	    unsaved_handles.append(handle);
	    if (!list.contains(handle.path()))
		list.append(handle.path());
	}

    if (!unsaved_handles.isEmpty()) {
	QMessageBox::StandardButton button = QMessageBox::warning(
	    hub()->mainWindow(), _("Unsaved changes"),
	    String(_("The following files have been modified:\n") +
		   join_strings(list, '\n') +
		   _("\nDo you want to save the changes?")),
	    QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);

	if (button == QMessageBox::Save)
	    foreach (Handle handle, unsaved_handles)
		if (handle.drawing()->isModified())
		    handle.save();

	if (button == QMessageBox::Cancel)
	    return false;
    }
#endif
    return true; **/

    p->hook.close();
}

void InstanceHub::update()
{
    p->viewport.invalidate();
}

String InstanceHub::selectedLibraryComponent() const
{
    if (!p->library_view.selected())
	return String();

    return static_cast<Item3 *>(p->library_view.selected()->item())->path();
}

/****************************************************************************/

void instantiate(Toolkit::Connection &conn)
{
    new Instance(conn);
}
