/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "tabs.h"
#include "icons.h"
#include "hub.h"
#include "dirmodel.h"


TabActionSet::TabActionSet(MainTabBar &tabBar)
    : tab_bar(tabBar), insert_after_current(false),
      nextAction(*this), previousAction(*this),
      closeAction(*this), moveLeftAction(*this), moveRightAction(*this),
      closeOthersAction(*this), insertAfterCurrentAction(*this)
{
    nextAction.setText(_("Go to next tab"));
    nextAction.setHotkey(_("C-Tab"));

    previousAction.setText(_("Go to previous tab"));
    previousAction.setHotkey(_("C-S-Tab"));

    closeAction.setText(_("&Hide tab"));
    closeAction.setHotkey(_("C-w"));

    moveLeftAction.setText(_("Move tab &left"));
    moveLeftAction.setIcon(icons::backward);

    moveRightAction.setText(_("Move tab &right"));
    moveRightAction.setIcon(icons::forward);

    closeOthersAction.setText(_("Hide &other tabs"));
    closeOthersAction.setHotkey(_("C-W"));
    closeOthersAction.setIcon(icons::clear);

    insertAfterCurrentAction.setText(_("Insert after &current tab"));
    // insertAfterCurrentAction.setCheckable(true);
}

void TabActionSet::insertTab(Item *tab)
{
    if (tab_bar.tabs.cachedIndexOf(tab) != NPOS)
	return;

    std::vector<Item *> tabs;
    tabs.push_back(tab);

    if (insert_after_current)
	tab_bar.tabs.replace(tab_bar.selectedTabIndex() + 1, 0, tabs);
    else
	tab_bar.tabs.replace(tab_bar.tabs.count(), 0, tabs);
}

void TabActionSet::actionTriggered(Toolkit::Action *action)
{
    if (action == &nextAction) {
	unsigned int count = tab_bar.tabs.count(),
		     index = tab_bar.selectedTabIndex() + 1;
	tab_bar.selectTab(tab_bar.tabs.cachedItem(index % count));
	return;
    }

    if (action == &previousAction) {
	unsigned int count = tab_bar.tabs.count(),
		     index = tab_bar.selectedTabIndex() + count - 1;
	tab_bar.selectTab(tab_bar.tabs.cachedItem(index % count));
	return;
    }

    if (action == &closeAction) {
	unsigned int index = tab_bar.selectedTabIndex();
	if (index == NPOS)
	    return;

	tab_bar.tabs.replace(index, 1, std::vector<Item *>());

	update();
	return;
    }

    if (action == &moveLeftAction) {
	unsigned int index = tab_bar.selectedTabIndex();
	if (index == NPOS || index == 0)
	    return;

	std::vector<Item *> items;
	items.push_back(tab_bar.tabs.cachedItem(index));
	items.push_back(tab_bar.tabs.cachedItem(index - 1));
	tab_bar.tabs.replace(index - 1, 2, items);

	update();
	return;
    }

    if (action == &moveRightAction) {
	unsigned int index = tab_bar.selectedTabIndex();
	if (index >= tab_bar.tabs.count() - 1)
	    return;

	std::vector<Item *> items;
	items.push_back(tab_bar.tabs.cachedItem(index + 1));
	items.push_back(tab_bar.tabs.cachedItem(index));
	tab_bar.tabs.replace(index, 2, items);

	update();
	return;
    }

    if (action == &closeOthersAction) {
	std::vector<Item *> items;
	items.push_back(tab_bar.selectedTab());
	tab_bar.tabs.setItems(items);

	update();
	return;
    }

    if (action == &insertAfterCurrentAction) {
	insert_after_current = !insert_after_current;
	return;
    }
}

void TabActionSet::update()
{
    if (tab_bar.selectedTabIndex() == 0)
	moveLeftAction.disable(
	    _("Can't %1: %2").arg(_("move tab left"),
				  _("Already at first position")));
    else
	moveLeftAction.enable();

    if (tab_bar.selectedTabIndex() == tab_bar.tabs.count() - 1)
	moveRightAction.disable(
	    _("Can't %1: %2").arg(_("move tab right"),
				  _("Already at last position")));
    else
	moveRightAction.enable();

    if (tab_bar.tabs.count() == 1) {
	closeAction.disable(
	    _("Can't %1: %2").arg(_("hide tab"),
				  _("This is the only tab")));
	closeOthersAction.disable(
	    _("Can't %1: %2").arg(_("hide other tabs"),
				  _("There are no other tabs")));
    } else {
	closeAction.enable();
	closeOthersAction.enable();
    }
}

/****************************************************************************/

MainTabBar::MainTabBar(Toolkit::Connection &conn, Hub &hub,
		       Toolkit::Action &toggle_action)
    : TabBar(conn), hub(hub), action_set(*this), menu_builder(conn)
{
    setInterior(&tabs);

    menu_builder.addAction(action_set.closeAction);
    menu_builder.addSeparator();
    menu_builder.addAction(action_set.moveLeftAction);
    menu_builder.addAction(action_set.moveRightAction);
    menu_builder.addAction(action_set.closeOthersAction);
    menu_builder.addSeparator();
    menu_builder.addAction(toggle_action);
    menu_builder.addAction(action_set.insertAfterCurrentAction);
}

MainTabBar::~MainTabBar()
{
    setInterior(0);
}

Toolkit::ActionSet *MainTabBar::actionSet()
{
    return &action_set;
}

void MainTabBar::selectNewTab(Item *tab)
{
    action_set.insertTab(tab);
    selectTab(tab);
}

void MainTabBar::tabSelected()
{
    action_set.update();
    hub.setHandle(static_cast<Item3 *>(selectedTab()));
}

Toolkit::Menu *MainTabBar::invokeContextMenu(Shadow *shadow)
{
    if (!shadow)
	return 0;
    hub.setHandle(static_cast<Item3 *>(shadow->item()));
    return menu_builder.menu();
}
