/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include <assert.h>
#include <X11/cursorfont.h>
#include "viewport.h"
#include "hub.h"
#include "view.h"
#include "painter.h"
#include "majormode.h"
#include "object.h"
#include "action_pl.h"
#include "dirmodel.h"


ViewportNavigator::ViewportNavigator(Toolkit::Connection &conn,
				     Hub &hub, double scale)
    : Navigator(conn), hub(hub), scale(scale)
{
}

Toolkit::NavigatorData ViewportNavigator::getNavigatorData() const
{
    if (!hub.view())
	return Toolkit::NavigatorData();

    View::ScrollInformation scroll_info = hub.view()->getScrollInformation();

    unsigned int width = scroll_info.horzRange / scale;
    unsigned int height = scroll_info.vertRange / scale;

    X::Pixmap pixmap = window().createPixmap(
	width, height, conn.visualInfo().depth);

    gc().setForeground(conn.screen().whitePixel());
    pixmap.fillRectangle(gc(), 0, 0, width, height);

    gc().setForeground(conn.screen().blackPixel());
    pixmap.drawLine(gc(), 0, 0, width - 1, height - 1);
    pixmap.drawLine(gc(), 0, height - 1, width - 1, 0);

    return Toolkit::NavigatorData(scroll_info.horzPosition / scale,
				  scroll_info.vertPosition / scale,
				  scroll_info.horzPageSize / scale,
				  scroll_info.vertPageSize / scale,
				  width, height, pixmap);
}

void ViewportNavigator::navigateTo(int x, int y)
{
    hub.setView(hub.view()->scrollTo(x * scale, y * scale));
}

/****************************************************************************/

class ViewportScroller : public Toolkit::WidgetScroller
{
    Viewport &viewport;
public:
    ViewportScroller(Viewport &viewport) : WidgetScroller(viewport),
					   viewport(viewport) {
    }
    int lineSize() const {
	return 50;
    }
    void setRange(int range) {
	WidgetScroller::setRange(range);
    }
    void setPageSize(int pageSize) {
	WidgetScroller::setPageSize(pageSize);
    }
private:
    void updatePosition() {
	viewport.updateScrollPos();
    }
};

/****************************************************************************/

static XGCValues gc_defaults =
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

class ViewportPainter : public Painter
{
    Toolkit::Connection &conn;
    X::Pixmap pixmap;
    X::GC gc;
    unsigned long gc_valuemask;
    XGCValues gc_values;
    int origin_x, origin_y;
    ScreenRect visible_rect;
public:
    ViewportPainter(Hub &hub, Toolkit::Connection &conn,
		    X::Pixmap pixmap, X::GC gc,
		    int originX, int originY, const ScreenRect &visibleRect)
	: Painter(hub), conn(conn), pixmap(pixmap), gc(gc),
	  gc_valuemask(0), gc_values(gc_defaults),
	  origin_x(originX), origin_y(originY), visible_rect(visibleRect) {
    }

    void setBlackPen() {
	gc_values.foreground = conn.screen().blackPixel();
	change_gc(GCForeground);
	gc_values.foreground = 0;
    }
    void setWhitePen() {
	gc_values.foreground = conn.screen().whitePixel();
	change_gc(GCForeground);
	gc_values.foreground = 0;
    }
    void setPen16(unsigned int red, unsigned int green, unsigned int blue) {
	XColor color;
	color.red = red;
	color.green = green;
	color.blue = blue;
	conn.screen().defaultColormap().allocColor(&color);
	gc_values.foreground = color.pixel;
	change_gc(GCForeground);
	gc_values.foreground = 0;
    }
    void setPen(unsigned int red, unsigned int green, unsigned int blue) {
	setPen16(red * 0x101, green * 0x101, blue * 0x101);
    }
    void setDashedBlackPen(char dashes, int dash_offset) {
	gc_values.foreground = conn.screen().blackPixel();
	gc_values.line_style = LineOnOffDash;
	gc_values.dash_offset = dash_offset;
	gc_values.dashes = dashes;
	change_gc(GCForeground | GCLineStyle | GCDashOffset | GCDashList);
	gc_values.foreground = 0;
	gc_values.line_style = LineSolid;
	gc_values.dash_offset = 0;
	gc_values.dashes = 0;
    }
    void setDashedPen16(char dashes, int dash_offset, unsigned int red,
			unsigned int green, unsigned int blue) {
	XColor color;
	color.red = red;
	color.green = green;
	color.blue = blue;
	conn.screen().defaultColormap().allocColor(&color);
	gc_values.foreground = color.pixel;
	gc_values.line_style = LineOnOffDash;
	gc_values.dash_offset = dash_offset;
	gc_values.dashes = dashes;
	change_gc(GCForeground | GCLineStyle | GCDashOffset | GCDashList);
	gc_values.foreground = 0;
	gc_values.line_style = LineSolid;
	gc_values.dash_offset = 0;
	gc_values.dashes = 0;
    }
    void setDashedPen(char dashes, int dash_offset, unsigned int red,
		      unsigned int green, unsigned int blue) {
	setDashedPen16(dashes, dash_offset,
		       red * 0x101, green * 0x101, blue * 0x101);
    }

    ScreenRect visibleRect() const {
	return visible_rect;
    }
    void drawLine(int x0, int y0, int x1, int y1) {
	pixmap.drawLine(gc, origin_x + x0, origin_y - y0,
			    origin_x + x1, origin_y - y1);
    }
    void drawLine(const ScreenPos &p0, const ScreenPos &p1) {
	pixmap.drawLine(gc,
	    origin_x + llrint(p0.x()), origin_y - llrint(p0.y()),
	    origin_x + llrint(p1.x()), origin_y - llrint(p1.y()));
    }
    void drawRectangle(int x, int y, unsigned int width,
				     unsigned int height) {
	if (width != 0) {
	    if (height != 0)
		pixmap.drawRectangle(gc, origin_x + x, origin_y - y - height,
					 width, height);
	    else
		pixmap.drawLine(gc, origin_x + x, origin_y - y,
				    origin_x + x + width, origin_y - y);
	} else if (height != 0)
	    pixmap.drawLine(gc, origin_x + x, origin_y - y - height,
				origin_x + x, origin_y - y);
    }
    void fillRectangle(int x, int y, unsigned int width,
				     unsigned int height) {
	if (width != 0 && height != 0)
	    pixmap.fillRectangle(gc, origin_x + x, origin_y - y - height + 1,
				     width, height);
    }
    void drawArc(int x, int y, unsigned int width, unsigned int height,
		 int angle1, int angle2) {
	pixmap.drawArc(gc, origin_x + x, origin_y - y - height,
		       width, height, angle1, angle2);
    }

    void change_gc(unsigned long valuemask) {
	gc.change(valuemask | gc_valuemask, &gc_values);
	gc_valuemask = valuemask & ~GCDashList;
    }
};

/****************************************************************************/

static X::Cursor create_invisible_cursor(X::Display display,
					 X::Drawable drawable)
{
    char bits[] = { 0x00 };
    XColor color = { 0, 0, 0, 0, 0, 0 };
    X::Pixmap pixmap = drawable.createBitmapFromData(bits, 1, 1);
    X::Cursor cursor = display.createPixmapCursor(
	pixmap, pixmap, &color, &color, 0, 0);
    pixmap.free();
    return cursor;
}

Viewport::Viewport(Toolkit::Connection &conn, Hub &hub)
    : Widget(conn, horz_scroller = new ViewportScroller(*this),
		   vert_scroller = new ViewportScroller(*this)),
      hub(hub), show_origin(false), show_grid(false), crosshairs_cursor(false),
      buttons_down(0),
      no_cursor(create_invisible_cursor(conn.display(),
					conn.screen().rootWindow())),
      move_cursor(conn.display().createFontCursor(XC_fleur)),
      /*has_focus(false),*/ pointer_inside(false), pointer_grabbed(false),
      significantly_moved(false), its_us(false),
      buffer_width(0), buffer_height(0), buffer_state(0), buffer_view(0)
{
/*
    setBackgroundRole(QPalette::Base);
    setAutoFillBackground(true);
    setMouseTracking(true);
 */

    setFocussables(std::vector<Toolkit::Focussable *>(1, this));
}

Viewport::~Viewport()
{
    if (!buffer.isNull())
	buffer.free();

    move_cursor.free();
    no_cursor.free();

    delete horz_scroller;
    delete vert_scroller;
}

ScreenPos Viewport::screen_pos(int x, int y) const
{
    return ScreenPos(x - width() / 2, height() / 2 - y);
}

bool Viewport::pointerActive() const
{
    return pointer_inside || pointer_grabbed;
}

ScreenPos Viewport::pointer() const
{
    return pointerActive() ? last_pos : ScreenPos();
}

bool Viewport::showOrigin() const
{
    return show_origin;
}

void Viewport::setShowOrigin(bool showOrigin)
{
    if (showOrigin == show_origin)
	return;

    show_origin = showOrigin;

    if (!buffer.isNull()) {
	buffer.free();
	buffer = X::Pixmap();
    }
    invalidate();
}

bool Viewport::showGrid() const
{
    return show_grid;
}

void Viewport::setShowGrid(bool showGrid)
{
    if (showGrid == show_grid)
	return;

    show_grid = showGrid;

    if (!buffer.isNull()) {
	buffer.free();
	buffer = X::Pixmap();
    }
    invalidate();
}

bool Viewport::crosshairsCursor() const
{
    return crosshairs_cursor;
}

void Viewport::setCrosshairsCursor(bool crosshairsCursor)
{
    if (crosshairsCursor == crosshairs_cursor)
	return;

    crosshairs_cursor = crosshairsCursor;

    /* Crosshairs cursor is active exactly when:
       crosshairs_cursor && has_focus && (pointer_inside || pointer_grabbed) */

    if (/*!has_focus ||*/ !pointerActive())
	return;

    if (crosshairsCursor)
	window().defineCursor(no_cursor);
    else
	window().undefineCursor();

    invalidate();
}

void Viewport::updateMinorMode()
{
    entry_pos = last_pos;
    significantly_moved = false;

    MinorMode *mode = hub.majorMode()->minorMode();
    mode->invoke();
    mode->mouseMoveEvent(last_pos, buttons_down, 0);
    invalidate();
}

void Viewport::updateScrollPos()
{
    if (its_us)
	return;

    hub.setView(hub.view()->scrollTo(horzScroller()->position(),
				     vertScroller()->position()));
}

void Viewport::updateLayout()
{
    if (!hub.view())
	return;

    assert(!its_us);
    its_us = true;

    View::ScrollInformation scroll_info = hub.view()->getScrollInformation();
    horz_scroller->setRange(scroll_info.horzRange);
    horz_scroller->setPageSize(scroll_info.horzPageSize);
    horz_scroller->setPosition(scroll_info.horzPosition);
    vert_scroller->setRange(scroll_info.vertRange);
    vert_scroller->setPageSize(scroll_info.vertPageSize);
    vert_scroller->setPosition(scroll_info.vertPosition);

    its_us = false;
}

static double ceil1(double ofs, double st, double val)
{
    return ceil((val - ofs) / st) * st + ofs;
}

void Viewport::paint(const Toolkit::Region &region, X::Pixmap pixmap)
{
    if (!hub.drawing()) {
	gc().setForeground(conn.screen().whitePixel());
	pixmap.fillRectangle(gc(), x() - region.x, y() - region.y,
				   width(), height());
	return;
    }

    if (!region.overlaps(control()->dimensions()))
	return;

    ScreenRect visible_rect(screen_pos(0, height() - 1),
			    screen_pos(width() - 1, 0));

    if (buffer.isNull() || buffer_width != width()
			|| buffer_height != height()) {
	if (!buffer.isNull())
	    buffer.free();
	buffer_width = width();
	buffer_height = height();
	buffer = window().createPixmap(buffer_width, buffer_height,
				       conn.visualInfo().depth);
	goto redraw;
    }

    if (hub.drawing()->head() != buffer_state || hub.view() != buffer_view) {
    redraw:
	gc().setForeground(conn.screen().whitePixel());
	buffer.fillRectangle(gc(), 0, 0, buffer_width, buffer_height);

	ViewportPainter painter(hub, conn, buffer, gc(),
				buffer_width / 2,
				buffer_height / 2, visible_rect);

	if (show_grid) {
	    ScreenPos origin = hub.view()->map(Metrum3());
	    ScreenDist step = hub.view()->map(
		Metrum3(hub.drawing()->base(), 0., 0.)).x() - origin.x();

	    while (step < 20.)
		step *= hub.drawing()->baseBase();
	    while (step > 20. * hub.drawing()->baseBase())
		step /= hub.drawing()->baseBase();

	    /* half grid step if it is too large */
	    if (step > 20. * hub.drawing()->baseBase() / 2)
		step /= 2;

	    painter.setPen(0xf0, 0xf0, 0xf0);

	    for (ScreenDist x = ceil1(origin.x(), step, visible_rect.x0());
		 x <= visible_rect.x1(); x += step)
		painter.drawLine(x, visible_rect.y1(), x, visible_rect.y0());

	    for (ScreenDist y = ceil1(origin.y(), step, visible_rect.y0());
		 y <= visible_rect.y1(); y += step)
		painter.drawLine(visible_rect.x0(), y, visible_rect.x1(), y);
	}

	if (show_origin) {
	    ScreenPos origin = hub.view()->map(Metrum3());
	    painter.setPen(0xe8, 0xe8, 0xe8);

	    if (origin.x() >= visible_rect.x0() &&
		origin.x() <= visible_rect.x1())
		painter.drawLine(origin.x(), visible_rect.y1(),
				 origin.x(), visible_rect.y0());

	    if (origin.y() >= visible_rect.y0() &&
		origin.y() <= visible_rect.y1())
		painter.drawLine(visible_rect.x0(), origin.y(),
				 visible_rect.x1(), origin.y());
	}

	foreach (Object *ob, hub.drawing()->objects())
	    ob->paint(painter);

	buffer_state = hub.drawing()->head();
	buffer_view = hub.view();
    }

    Toolkit::Region r = region.intersection(control()->dimensions());
    X::Drawable::copyArea(buffer, pixmap, gc(), r.x - x(), r.y - y(),
			  r.width, r.height, r.x - region.x, r.y - region.y);

    ViewportPainter painter(hub, conn, pixmap, gc(),
			    x() + width() / 2 - region.x,
			    y() + height() / 2 - region.y, visible_rect);

    if (hub.majorMode()) {
	List<Object *> selected = hub.majorMode()->selected();

	foreach (Object *ob, hub.drawing()->objects())
	    if (selected.contains(ob))
		foreach (ActionPoint *p, ob->handles()) {
		    painter.setBlackPen();
		    painter.paintHandle(p->pos(), 0x00, 0xe0, 0x00);
		}

	hub.majorMode()->paint(painter);
	hub.majorMode()->minorMode()->paint(painter);
    }

    if (crosshairsCursor() && pointerActive()) {
	painter.setBlackPen();
	painter.drawLine(visible_rect.x0(), last_pos.y(),
			 visible_rect.x1(), last_pos.y());
	painter.drawLine(last_pos.x(), visible_rect.y1(),
			 last_pos.x(), visible_rect.y0());
    }

    gc().setForeground(conn.screen().blackPixel());
    unsigned int y_feed = 0;

    foreach (Message message, hub.handle()->messages) {
	X::FontSet font_set =
	    message.type == Message::Error ? conn.fontSetBold()
					   : conn.fontSet();
	String prefix;
	if (message.line != -1)
	    if (message.column != -1)
		prefix = _("%1:%2: ").arg(int_to_str(message.line + 1),
					  int_to_str(message.column + 1));
	    else
		prefix = _("%1: ").arg(int_to_str(message.line + 1));

	prefix += message.type == Message::Error   ? _("error: ") :
		  message.type == Message::Warning ? _("warning: ") :
						     _("note: ");
	pixmap.mbDrawString(font_set, gc(),
			    x() - region.x + 8,
			    y() - region.y + 8 + y_feed + font_set.ascent(),
			    prefix + message.str);
	y_feed += font_set.ascent() + font_set.descent();
    }
}

void Viewport::enterEvent(Time time, int x, int y, int x_root, int y_root,
			  unsigned int state)
{
    if (pointer_inside)
	return;

    pointer_inside = true;
    significantly_moved = true;
    if (crosshairsCursor() && /*has_focus &&*/ !pointer_grabbed)
	window().defineCursor(no_cursor);

    motionEvent(time, x, y, x_root, y_root, state, false);
    invalidate();
}

void Viewport::leaveEvent(Time time, int x, int y, int x_root, int y_root,
			  unsigned int state)
{
    if (!pointer_inside)
	return;

    pointer_inside = false;
    significantly_moved = true;
    if (crosshairsCursor() && /*has_focus &&*/ !pointer_grabbed)
	window().undefineCursor();
    invalidate();
}


void Viewport::motionEvent(Time time, int x, int y, int x_root, int y_root,
			   unsigned int state, char is_hint)
{
    if (!hub.drawing())
	return;

    ScreenPos current_pos = screen_pos(x - this->x(), y - this->y());

    if (current_pos == last_pos)
	return;

    if (buttons_down & (1 << Button2)) {
	hub.setView(hub.view()->move(current_pos - last_pos));
	last_pos = current_pos;
	invalidate();
	return;
    }

    last_pos = current_pos;

    if (!significantly_moved) {
	ScreenPos delta = current_pos - entry_pos;
	if (abs(delta.x()) >= 3 || abs(delta.y()) >= 3)
	    significantly_moved = true;
    }

    if (hub.majorMode())
	hub.majorMode()->minorMode()->mouseMoveEvent(
	    last_pos, buttons_down, state);
    //if (crosshairsCursor() /*&& has_focus*/)
    invalidate();
}

void Viewport::buttonPressEvent(Time time, int x, int y,
				int x_root, int y_root,
				unsigned int state, unsigned int button)
{
    if (button == Button4) {
	if (hub.drawing())
	    hub.setView(hub.view()->zoom(exp(.25),
					 screen_pos(x - this->x(),
						    y - this->y())));
	return;
    }
    if (button == Button5) {
	if (hub.drawing())
	    hub.setView(hub.view()->zoom(exp(-.25),
					 screen_pos(x - this->x(),
						    y - this->y())));
	return;
    }

    if (control())
	control()->focus(this);

    if (!hub.drawing())
	return;

    buttons_down |= 1 << button;
    pointer_grabbed = true;

    if (button == Button2) {
	if (!crosshairs_cursor)
	    window().defineCursor(move_cursor);
	return;
    }

    if (button == Button3) {
       conn.enterMenuMode(time, contextMenu);
       contextMenu->show(time, x_root, y_root, hook());
       return;
    }

    motionEvent(time, x, y, x_root, y_root, state, false);
    if (hub.majorMode())
	hub.majorMode()->minorMode()->mousePressEvent(
	    last_pos, button, buttons_down & ~(1 << button), state);
    entry_pos = last_pos;
    significantly_moved = false;
    invalidate();
}

void Viewport::buttonReleaseEvent(Time time, int x, int y,
				  int x_root, int y_root,
				  unsigned int state, unsigned int button)
{
    if (!hub.drawing())
	return;

    buttons_down &= ~(1 << button);
    if (buttons_down == 0) {
	pointer_grabbed = false;
	if (!pointer_inside)
	    window().undefineCursor();
    }

    if (button == Button2) {
	if (!crosshairs_cursor)
	    window().undefineCursor();
	return;
    }

    motionEvent(time, x, y, x_root, y_root, state, false);
    if (hub.majorMode())
	hub.majorMode()->minorMode()->mouseReleaseEvent(
	    last_pos, button, buttons_down, state, significantly_moved);
    invalidate();
}

void Viewport::keyPressEvent(Time time, int x, int y, int x_root, int y_root,
			     unsigned int state, unsigned int keycode,
			     const std::string &text, KeySym keysym)
{
}

void Viewport::keyReleaseEvent(Time time, int x, int y, int x_root, int y_root,
			       unsigned int state, unsigned int keycode)
{
}

void Viewport::focusInEvent()
{
/*
    has_focus = true;
    if (crosshairsCursor() && pointerActive())
	window().defineCursor(no_cursor);
*/
}

void Viewport::focusOutEvent()
{
/*
    has_focus = false;
    if (crosshairsCursor() && pointerActive())
	window().undefineCursor();
*/
}
