/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include "hub.h"
#include "icons.h"
#include "shared.h"
#include "dirmodel.h"
#include "view.h"

#include "sch_line.h"
#include "sch_picture.h"
#include "sch_box.h"
#include "sch_net.h"
#include "sch_circle.h"
#include "sch_text.h"
#include "sch_component.h"
#include "nb_majormode.h"

#include "nb_attribute.h"


class ToolboxModel
{
public:
    ToolboxModel(Hub &hub);
    MutableStandardInterior root_interior;
    StandardItem edit_modes, schematic_modes, neverball_modes;

    SelectMajorMode select_major_mode;
    DeleteMajorMode delete_major_mode;

    /* Schematic major modes */
    LineMajorMode line_major_mode;
    PictureMajorMode picture_major_mode;
    BoxMajorMode box_major_mode;
    CircleMajorMode circle_major_mode;
    TextMajorMode text_major_mode;
    NetMajorMode net_major_mode;
    ComponentMajorMode component_major_mode;

    /* Neverball major modes */
    PathMajorMode path_major_mode;
    BodyMajorMode body_major_mode;
    CoinMajorMode coin_major_mode;
    GoalMajorMode goal_major_mode;
    JumpMajorMode jump_major_mode;
    SwitchMajorMode switch_major_mode;
    BillboardMajorMode billboard_major_mode;
    UserMajorMode user_major_mode;
    ViewMajorMode view_major_mode;
};

ToolboxModel::ToolboxModel(Hub &hub)
    : select_major_mode(hub), delete_major_mode(hub),
      line_major_mode(hub), picture_major_mode(hub), box_major_mode(hub),
      circle_major_mode(hub), text_major_mode(hub), net_major_mode(hub),
      component_major_mode(hub),
      path_major_mode(hub), body_major_mode(hub), coin_major_mode(hub),
      goal_major_mode(hub), jump_major_mode(hub), switch_major_mode(hub),
      billboard_major_mode(hub), user_major_mode(hub), view_major_mode(hub)
{
    root_interior.addItem(&select_major_mode);
    root_interior.addItem(&edit_modes);
    root_interior.addItem(&schematic_modes);
    root_interior.addItem(&neverball_modes);

    edit_modes.setCaption(_("Edit"));
    edit_modes.setIcon(icons::generic);
    edit_modes.interior.addItem(&delete_major_mode);

    schematic_modes.setCaption(_("Schematic"));
    schematic_modes.setIcon(icons::generic);
    schematic_modes.interior.addItem(&line_major_mode);
    schematic_modes.interior.addItem(&picture_major_mode);
    schematic_modes.interior.addItem(&box_major_mode);
    schematic_modes.interior.addItem(&circle_major_mode);
    schematic_modes.interior.addItem(&text_major_mode);
    schematic_modes.interior.addItem(&net_major_mode);
    schematic_modes.interior.addItem(&component_major_mode);

    neverball_modes.setCaption(_("Neverball"));
    neverball_modes.setIcon(icons::neverball);
    neverball_modes.interior.addItem(&path_major_mode);
    neverball_modes.interior.addItem(&body_major_mode);
    neverball_modes.interior.addItem(&coin_major_mode);
    neverball_modes.interior.addItem(&goal_major_mode);
    neverball_modes.interior.addItem(&view_major_mode);
    neverball_modes.interior.addItem(&jump_major_mode);
    neverball_modes.interior.addItem(&switch_major_mode);
    neverball_modes.interior.addItem(&billboard_major_mode);
    neverball_modes.interior.addItem(&user_major_mode);
}

/****************************************************************************/

class AttributeInterior : public ItemInterior
{
    List<AbstractAttributeItem *> items;
    List<Item *> visible_items;

    Item *item(unsigned int index) const {
	return visible_items[index];
    }
    unsigned int indexOf(Item *item) const {
	return visible_items.find(item);
    }
public:
    AttributeInterior(const List<AbstractAttributeItem *> &items)
	: items(items) {
    }
    void showAttributes(const List<Attribute *> &attributes) {
	visible_items.clear();

	foreach (AbstractAttributeItem *item, items)
	    if (item->showAttributes(attributes))
		visible_items.append(item);

	replace(0, count(), visible_items.count());
    }
};

class AttributeCategory : public AbstractAttributeItem
{
    AttributeInterior interior;
public:
    AttributeCategory(const String &name, Icon *icon,
		      const List<AbstractAttributeItem *> &items)
	: interior(items) {
	setCaption(name);
	setIcon(icon);
	setInterior(&interior);
    }
    bool showAttributes(const List<Attribute *> &attributes) {
	interior.showAttributes(attributes);
	return interior.count() != 0;
    }
};

/****************************************************************************/

class AttributeModel
{
    SchematicAttributeItems schematic_attribute_items;
    NeverballAttributeItems neverball_attribute_items;
    AttributeCategory geometry, attributes, line, fill;
public:
    AttributeModel(Hub &hub);
    AttributeInterior root_interior;
};

AttributeModel::AttributeModel(Hub &hub)
    : schematic_attribute_items(hub),
      neverball_attribute_items(hub),
      geometry(_("Geometry"), 0, List<AbstractAttributeItem *>()
	  << &neverball_attribute_items.radius),
      attributes(_("Attributes"), 0, List<AbstractAttributeItem *>()
	  << &schematic_attribute_items.color
	  << &schematic_attribute_items.isBus
	  << &schematic_attribute_items.isPin
	  << &schematic_attribute_items.isInverted
	  << &schematic_attribute_items.textSize
	  << &schematic_attribute_items.visibility
	  << &schematic_attribute_items.showNameValue
	  << &schematic_attribute_items.angle
	  << &schematic_attribute_items.alignment
	  << &schematic_attribute_items.selectable
	  << &schematic_attribute_items.mirror
	  << &neverball_attribute_items.symbol
	  << &neverball_attribute_items.travelTime
	  << &neverball_attribute_items.path
	  << &neverball_attribute_items.enableFlag
	  << &neverball_attribute_items.model
	  << &neverball_attribute_items.value
	  << &neverball_attribute_items.timer
	  << &neverball_attribute_items.state
	  << &neverball_attribute_items.detail),
      line(_("Line style"), 0, List<AbstractAttributeItem *>()
	  << &schematic_attribute_items.lineWidth
	  << &schematic_attribute_items.capStyle
	  << &schematic_attribute_items.dashStyle
	  << &schematic_attribute_items.dashLength
	  << &schematic_attribute_items.dashSpace),
      fill(_("Fill style"), 0, List<AbstractAttributeItem *>()
	  << &schematic_attribute_items.fillType
	  << &schematic_attribute_items.fillWidth
	  << &schematic_attribute_items.fillAngle0
	  << &schematic_attribute_items.fillPitch0
	  << &schematic_attribute_items.fillAngle1
	  << &schematic_attribute_items.fillPitch1),
      root_interior(List<AbstractAttributeItem *>()
	   << &geometry << &attributes << &line << &fill)
{
}

/****************************************************************************/

Hub::Hub() : _handle(0), major_mode(0),
	     drawing_head(0), drawing_is_modified(false), minor_mode(0),
	     toolbox_model(new ToolboxModel(*this)),
	     attribute_model(new AttributeModel(*this))
{
    shared.registerHub(this);
}

Hub::~Hub()
{
    shared.unregisterHub(this);
    delete attribute_model;
    delete toolbox_model;
}

Item3 *Hub::handle() const
{
    return _handle;
}

Drawing *Hub::drawing() const
{
    return _handle ? _handle->drawing() : 0;
}

View *Hub::view() const
{
    if (!_handle)
	return 0;

    Drawing *drawing = _handle->drawing();
    View *&view = views[drawing];
    if (!view)
	view = new SimpleView(Metrum3(), 20., 0., drawing);

    /*  qreal sqrt_5 = sqrt(.5);
	view = new OrthogonalView(Metrum3(),
				  ScreenPos(sqrt_5 * 20, sqrt_5 * 10),
				  ScreenPos(-sqrt_5 * 20, sqrt_5 * 10),
				  ScreenPos(0, 20), drawing); */

    return view;
}

MajorMode *Hub::majorMode() const
{
    return major_mode;
}

ItemInterior *Hub::toolbox() const
{
    return &toolbox_model->root_interior;
}

ItemInterior *Hub::attributes() const
{
    return &attribute_model->root_interior;
}

void Hub::drawingStatusChanged()
{
    if (!_handle)
	return;

    Drawing *drawing = _handle->drawing();

    if (drawing->head() == drawing_head &&
	drawing->isModified() == drawing_is_modified)
	return;

    drawing_head = drawing->head();
    drawing_is_modified = drawing->isModified();
    updateDrawingStatus();
}

void Hub::setHandle(Item3 *handle)
{
    assert(handle->drawing());

    if (handle == _handle)
	return;

    if (majorMode())
	majorMode()->cancel();

    _handle = handle;
    updateHandle();
}

void Hub::setView(View *view)
{
    if (!_handle)
	return;

    views[_handle->drawing()] = view;
    updateView();
}

void Hub::setMajorMode(MajorMode *majorMode)
{
    if (majorMode == major_mode)
	return;

    if (major_mode)
	major_mode->cancel();
    major_mode = majorMode;

    updateMajorMode();
    minorModeChanged();
    selectedObjectsChanged();
}

void Hub::deselectMajorMode()
{
    if (majorMode() == &toolbox_model->select_major_mode)
	majorMode()->cancel();
    else
	setMajorMode(&toolbox_model->select_major_mode);
}

MajorMode *Hub::componentMajorMode() const
{
    return &toolbox_model->component_major_mode;
}

void Hub::selectObjects(const List<Object *> &objects)
{
    setMajorMode(&toolbox_model->select_major_mode);
    toolbox_model->select_major_mode.setSelected(objects);
}

void Hub::minorModeChanged()
{
    if (!majorMode() || majorMode()->minorMode() == minor_mode)
	return;

    minor_mode = majorMode()->minorMode();

    updateMinorMode();
    statusMessageChanged();
}

void Hub::selectedObjectsChanged()
{
    if (!majorMode() || majorMode()->selected() == selected_objects)
	return;

    selected_objects = majorMode()->selected();
    updateSelectedObjects();

    List<Attribute *> attrs;

    foreach (Object *ob, selected_objects)
	foreach (Attribute *attr, ob->attributes())
	    if (!attrs.contains(attr))
		attrs.append(attr);

    attribute_model->root_interior.showAttributes(attrs);
}

void Hub::statusMessageChanged()
{
    if (!minor_mode || minor_mode->statusMessage() == status_message)
	return;

    status_message = minor_mode->statusMessage();
    updateStatusMessage();
}
