/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "majormode.h"
#include "icons.h"
#include "hub.h"
#include "painter.h"
#include "action_pl.h"
#include "drawing.h"
#include "object.h"
#include "action.h"


MajorModeTimer::MajorModeTimer(MajorMode &majorMode) : major_mode(majorMode)
{
}

void MajorModeTimer::timeout()
{
    major_mode.timeout();
}


MajorMode::MajorMode(Hub &hub) : minor_mode(0), currently_snapped_point(0),
				 timer(*this), hub(hub)
{
    setCaption(_("Unknown mode"));
    setIcon(icons::generic);
}

MinorMode *MajorMode::minorMode() const
{
    return minor_mode;
}

void MajorMode::setMinorMode(MinorMode *mode, bool reset)
{
    if (reset)
	active_points.clear();
    currently_snapped_point = 0;

    /* Re-set the mode even if already selected */
    minor_mode = mode;
    hub.minorModeChanged();

    timer.unset();
}

List<Object *> MajorMode::selected() const
{
    return the_selected;
}

void MajorMode::setSelected(const List<Object *> &selected)
{
    if (selected == the_selected)
	return;

    the_selected = selected;
    hub.selectedObjectsChanged();
}

void MajorMode::paint(Painter &painter)
{
    painter.setPen(0xff, 0x00, 0x00);
    foreach (ActionPoint *point, activePoints())
	painter.paintActivePoint(point->pos());
}

List<ActionLine *> MajorMode::lines(ActionPoint * /*handle*/)
{
    return List<ActionLine *>();
}

List<ActionPoint *> MajorMode::activePoints() const
{
    return active_points;
}

void MajorMode::notifySnapped(ActionPoint *point)
{
    if (point != currently_snapped_point) {
	if (currently_snapped_point)
	    timer.unset();
	currently_snapped_point = point;
	if (currently_snapped_point)
	    timer.set(1, 0, false);
    }
}

void MajorMode::timeout()
{
    if (!active_points.contains(currently_snapped_point)) {
	foreach (Object *ob, hub.drawing()->objects())
	    foreach (ActionPoint *p, ob->handles())
		if (p->pos() == currently_snapped_point->pos())
		    active_points.append(p);
    } else {
	foreach (Object *ob, hub.drawing()->objects())
	    foreach (ActionPoint *p, ob->handles())
		if (p->pos() == currently_snapped_point->pos())
		    active_points.removeAll(p);
    }

    hub.update();
}

/****************************************************************************/

SelectMajorMode::SelectMajorMode(Hub &hub) : MajorMode(hub),
					     selectMinorMode(hub, *this),
					     grabHandleMinorMode(hub, *this)
{
    setMinorMode(&selectMinorMode);

/** QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(selectMinorMode.widget());
    layout->addStretch(); **/

    setCaption(_("Select"));
    setIcon(icons::cursor);
}

void SelectMajorMode::cancel()
{
    setSelected(List<Object *>());
    setMinorMode(&selectMinorMode, true);
}

void SelectMajorMode::setSelected(const List<Object *> &selected)
{
    MajorMode::setSelected(selected);
}

void SelectMajorMode::accept(MinorMode *minor_mode)
{
    if (minor_mode == &selectMinorMode) {
	if (selectMinorMode.handle()) {
	    grabHandleMinorMode.setHandle(selectMinorMode.handle());
	    setMinorMode(&grabHandleMinorMode);
	} else
	    setSelected(selectMinorMode.objects());
    }

    if (minor_mode == &grabHandleMinorMode) {
	hub.drawing()->performActions(
	    List<Action *>() << new MoveHandleAction(
		grabHandleMinorMode.handle(),
		grabHandleMinorMode.point()),
	    icons::generic, _("Move handle"));

	setMinorMode(&selectMinorMode, true);
    }
}

List<ActionLine *> SelectMajorMode::lines(ActionPoint *handle)
{
    if (minorMode() == &grabHandleMinorMode && selectMinorMode.handle())
	return selectMinorMode.handle()->lines();

    return List<ActionLine *>();
}

/****************************************************************************/

DeleteMajorMode::DeleteMajorMode(Hub &hub) : MajorMode(hub),
					     selectMinorMode(hub, *this)
{
    setMinorMode(&selectMinorMode);

/** QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(selectMinorMode.widget());
    layout->addStretch(); **/

    setCaption(_("Delete"));
    setIcon(icons::erase);
}

void DeleteMajorMode::cancel()
{
    setMinorMode(&selectMinorMode, true);
}

void DeleteMajorMode::accept(MinorMode *minor_mode)
{
    List<Object *> objects;
    List<Action *> actions;
    String name;

    if (selectMinorMode.handle())
	objects = List<Object *>() << selectMinorMode.handle()->object();
    else
	objects = selectMinorMode.objects();

    if (objects.count() == 1)
	name = _("Delete %1").arg(objects[0]->name());
    else
	name = _("Delete objects");

    foreach (Object *ob, objects)
	actions << new DeleteObjectAction(ob);

    hub.drawing()->performActions(actions, icons::remove, name);
}
