/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdio.h>
#include <iostream>
#include "nb_lump.h"
#include "nb_matrix.h"

#define SMALL 0.0005f

const String cpath = "/usr/share/games/neverball/";


/* The following code caches image sizes. Textures are referenced by
 * name, but their sizes are necessary when computing texture
 * coordinates. This code allows each file to be accessed only once
 * regardless of the number of surfaces refering to it. */

class CachedImageSize
{
public:
    CachedImageSize(const String &s, int w, int h);

    String s;
    int w;
    int h;
};


CachedImageSize::CachedImageSize(const String &s, int w, int h)
    : s(s), w(w), h(h)
{
}

static void size_image(const String &name, int &w, int &h)
{
    static List<CachedImageSize *> images;

    foreach (CachedImageSize *image, images)
	if (image->s == name) {
	    w = image->w;
	    h = image->h;
	    return;
	}

    /* TODO */
    w = 256;
    h = 256;

    images.append(new CachedImageSize(name, w, h));
}

/****************************************************************************/

Material::Material(const String &name)
{
    FILE *fin;

    f = name;

    a[0] = a[1] = a[2] = a[3] = 1.0f;
    d[0] = d[1] = d[2] = d[3] = 1.0f;
    s[0] = s[1] = s[2] = s[3] = 1.0f;
    e[0] = e[1] = e[2] = e[3] = 1.0f;
    h[0] = 0.0f;
    fl   = 0;

    String path = cpath + name;

    if ((fin = fopen(path.data(), "r")) != NULL) {
	fscanf(fin,
	       "%f %f %f %f "
	       "%f %f %f %f "
	       "%f %f %f %f "
	       "%f %f %f %f "
	       "%f %hd ",
	       d, d + 1, d + 2, d + 3,
	       a, a + 1, a + 2, a + 3,
	       s, s + 1, s + 2, s + 3,
	       e, e + 1, e + 2, e + 3,
	       h, &fl);
	fclose(fin);
    }
}

/****************************************************************************/

Vertex::Vertex(const Vector3 &p, Side *s0, Side *s1, Side *s2)
    : p(p), s0(s0), s1(s1), s2(s2)
{
}

void Vertex::move(const Vector3 &p)
{
    this->p -= p;
}

/****************************************************************************/

Edge::Edge(Vertex *i, Vertex *j) : i(i), j(j)
{
}

/****************************************************************************/

Side::Side(const Vector3 &p0, const Vector3 &p1, const Vector3 &p2,
	   short tu, short tv, short r,
	   float su, float sv, const String &s)
{
    static const Vector3 base[6][3] = {
	{ Vector3( 0,  0,  1), Vector3( 1,  0,  0), Vector3( 0, -1,  0) },
	{ Vector3( 0,  0, -1), Vector3( 1,  0,  0), Vector3( 0, -1,  0) },
	{ Vector3( 1,  0,  0), Vector3( 0,  0, -1), Vector3( 0, -1,  0) },
	{ Vector3(-1,  0,  0), Vector3( 0,  0, -1), Vector3( 0, -1,  0) },
	{ Vector3( 0,  1,  0), Vector3( 1,  0,  0), Vector3( 0,  0,  1) },
	{ Vector3( 0, -1,  0), Vector3( 1,  0,  0), Vector3( 0,  0,  1) },
    };

    float R[16];
    float k, _d = 0.f;
    short i, _n = 0;
    int   w, h;

    size_image(s, w, h);

    n = v_nrm(v_crs(p0 - p1, p2 - p1));

    d = v_dot(n, p1);

    for (i = 0; i < 6; i++)
	if ((k = v_dot(n, base[i][0])) >= _d)
	{
	    _d = k;
	    _n = i;
	}

    m_rot(R, base[_n][0], V_RAD(r));

    plane_u = m_vxfm(R, base[_n][1]);
    plane_v = m_vxfm(R, base[_n][2]);
    plane_p = m_vxfm(R, base[_n][1] * su * tu +
			base[_n][2] * sv * tv);

    plane_u *= 64.f / w;
    plane_v *= 64.f / h;

    plane_u *= 1.f / su;
    plane_v *= 1.f / sv;

    plane_m = new Material(s);
}

void Side::move(const Vector3 &p)
{
    d -= v_dot(n, p);
}

/* Test the location of a point with respect to a side plane. */

bool Side::fore_side(const Vector3 &p) const
{
    return v_dot(p, n) - d > +SMALL;
}

bool Side::on_side(const Vector3 &p) const
{
    float _d = v_dot(p, n) - d;

    return -SMALL < _d && _d < +SMALL;
}

/****************************************************************************/

TextureCoord::TextureCoord(float u0, float u1)
{
    u[0] = u0;
    u[1] = u1;
}

/****************************************************************************/

Geometry::Geometry(TextureCoord *ti, Side *si, Vertex *vi,
		   TextureCoord *tj, Side *sj, Vertex *vj,
		   TextureCoord *tk, Side *sk, Vertex *vk,
		   Material *m) : m(m),
				  ti(ti), tj(tj), tk(tk),
				  si(si), sj(sj), sk(sk),
				  vi(vi), vj(vj), vk(vk)
{
}
