/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef NB_OBJECT_H
#define NB_OBJECT_H

#include "util/metrum3.h"
#include "object.h"
#include "action_pl.h"

class XMLElement;
class Side;
class Vector3;
class Vertex;
class Edge;
class Geometry;

class PathObject : public Object
{
public:
    PathObject(Metrum3 pos, Drawing *drawing);
    PathObject(XMLElement *element, Drawing *drawing);

    String symbol() const;
    Metrum3 pos() const;
    PathObject *path() const;
    void setPath(PathObject *path);

    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    List<ActionLine *> handleLines(const ActionPoint *handle);
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    Metrum3 lineBase(const ActionLine *line) const;
    Metrum3 lineTangent(const ActionLine *line) const;
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    String the_symbol;
    Metrum3 the_pos;
    double travelTime;
    PathObject *the_path;
    bool enable;

    ActionPoint handle;
    ObjectActionLine tangent, normal;
};

class BodyObject : public Object
{
public:
    BodyObject(Metrum3 pos, Drawing *drawing);
    BodyObject(XMLElement *element, Drawing *drawing);
    String name() const;
    ObjectContext *children() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos;
    String model;
    PathObject *path;
    ObjectContext context;

    ActionPoint handle;
};

class LumpHandle
{
public:
    LumpHandle(Side *s0, Side *s1, Side *s2, Object *object);
    bool matches(Side *t0, Side *t1, Side *t2);

    ActionPoint handle;
    Side *s0, *s1, *s2;
};

class LumpObject : public Object
{
public:
    LumpObject(bool detail, Drawing *drawing);
    LumpObject(XMLElement *element, Drawing *drawing);

    bool detail() const;
    void setDetail(bool detail);

    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

    List<Side *> s;
    void clip();

private:
    void move(const Vector3 &p);
    bool ok_vert(const Vector3 &p) const;

    Metrum3 where_do_these_thingies_meet(Side *si, Side *sj, Side *sk,
					 Drawing *drawing) const;


    void clip_vert(Side *si, Side *sj, Side *sk);
    void clip_edge(Side *si, Side *sj);
    void clip_geom(Side *sp);

    bool is_detail;

    List<Vertex *> v;
    List<Edge *> e;
    List<Geometry *> g;

    List<LumpHandle *> handle_buf;
};

class CoinObject : public Object
{
public:
    CoinObject(Metrum3 pos, Drawing *drawing);
    CoinObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos;
    int value;

    ActionPoint handle;
};

class GoalObject : public Object
{
public:
    GoalObject(const Metrum3 &pos, Drawing *drawing);
    GoalObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos;
    Metrum radius;

    ActionPoint centerHandle, perimeterHandle;
};

class JumpObject : public Object
{
public:
    JumpObject(const Metrum3 &pos, const Metrum3 &target, Drawing *drawing);
    JumpObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    List<ActionLine *> handleLines(const ActionPoint *handle);
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    Metrum3 lineBase(const ActionLine *line) const;
    Metrum3 lineTangent(const ActionLine *line) const;
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos, target;
    Metrum radius;

    ActionPoint centerHandle, perimeterHandle, targetHandle;
    ObjectActionLine tangent, startNormal, targetNormal;
};

class SwitchObject : public Object
{
public:
    SwitchObject(const Metrum3 &pos, Drawing *drawing);
    SwitchObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos;
    Metrum radius;
    PathObject *path;
    double timer;
    int state;

    ActionPoint centerHandle, perimeterHandle;
};

class BillboardObject : public Object
{
public:
    BillboardObject(const Metrum3 &pos, Drawing *drawing);
    BillboardObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    List<ActionLine *> handleLines(const ActionPoint *handle);
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    Metrum3 lineBase(const ActionLine *line) const;
    Metrum3 lineTangent(const ActionLine *line) const;
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    int fl;
    /* String material; */
    double repeatTimeInterval;
    /* double widthCoefficients[3];
       double heightCoefficients[3];
       double zRotationCoefficients[3]; */
    Metrum3 pos;
};

class UserObject : public Object
{
public:
    UserObject(const Metrum3 &pos, Drawing *drawing);
    UserObject(XMLElement *element, Drawing *drawing);
    String name() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos;
    Metrum radius;

    ActionPoint centerHandle, perimeterHandle;
};

class ViewObject : public Object
{
public:
    ViewObject(const Metrum3 &pos, const Metrum3 &target, Drawing *drawing);
    ViewObject(XMLElement *element, Drawing *drawing);
    String name() const;
    Icon *icon() const;
    void paint(Painter &painter, const Embedding &embedding);
    void paintShadow(const Metrum3 &pos, const ActionPoint *handle,
		     Painter &painter);

    List<ActionPoint *> handles();
    Metrum3 handlePos(const ActionPoint *handle) const;
    List<ActionLine *> handleLines(const ActionPoint *handle);
    void moveHandle(const Metrum3 &pos, const ActionPoint *handle);
    Metrum3 lineBase(const ActionLine *line) const;
    Metrum3 lineTangent(const ActionLine *line) const;
    ScreenDist proximity(const ScreenPos &point, View *view,
			 const Embedding &embedding) const;
    bool within(const ScreenRect &rect, View *view, bool partially,
		const Embedding &embedding) const;
    void write(XMLWriter *writer);
    List<Attribute *> attributes() const;
    Variant attribute(const Attribute *attribute) const;
    Variant attributeDefault(const Attribute *attribute) const;
    void setAttribute(const Variant &value, const Attribute *attribute);

private:
    Metrum3 pos, target;

    ActionPoint posHandle, targetHandle;
    ObjectActionLine tangent, posNormal, targetNormal;
};

#endif
