/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <config.h>
#include "read_z.h"
#include "messages.h"

#ifdef HAVE_LIBZ

#include <stdio.h>
#include <assert.h>
#include <zlib.h>

#define BUFSIZ_OUT BUFSIZ


class ZSink : public Shared::Sink
{
    Shared::Sink *const sink;
    Messages &messages;
    z_stream strm;
    enum { Uninitialized, Reading, End, Freed } state;
public:
    ZSink(Shared::Sink *sink, bool gzip_format, Messages &messages);
    ~ZSink();
    bool read(const void *buf, ssize_t count);
};

ZSink::ZSink(Shared::Sink *sink, bool gzip_format, Messages &messages)
    : sink(sink), messages(messages), state(Uninitialized)
{
    strm.next_in = Z_NULL;
    strm.avail_in = 0;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;

    int ret = inflateInit2(&strm, gzip_format ? 15 + 16 : 15);
    /* 15 is the default value, +16 indicates gzip format */

    if (strm.msg)
	messages.error(_("%1 says: %2").arg("inflateInit2", strm.msg));

    switch (ret) {
    case Z_OK:
	state = Reading;
	break;
    case Z_MEM_ERROR:
	messages.error("inflateInit2: " + _("not enough memory"));
	break;
    case Z_VERSION_ERROR:
	messages.error("inflateInit2: " + _("incompatible zlib "
					    "library version"));
	break;
    case Z_STREAM_ERROR:
	messages.error("inflateInit2: " + _("parameters are invalid"));
	break;
    default:
	messages.error("inflateInit2: " + _("error #%1").arg(int_to_str(ret)));
    }
}

ZSink::~ZSink()
{
    if (state == Reading || state == End) {
	/* clean up in case reading was not finished normally */
	inflateEnd(&strm);
    }
}

bool ZSink::read(const void *buf, ssize_t count)
{
    assert(state != Freed);

    if (state == Uninitialized)
	return false;

    if (count == 0) {
	if (state == Reading)
	    messages.error(_("unexpected end of stream"));

	int ret = inflateEnd(&strm);
	if (strm.msg)
	    messages.error(_("%1 says: %2").arg("inflateEnd", strm.msg));
	state = Freed;

	switch (ret) {
	case Z_OK:
	    /* nice */
	    break;
	case Z_STREAM_ERROR:
	    messages.error("inflateEnd: " + _("inconsistent stream state"));
	    return false;
	default:
	    messages.error("inflateEnd: " + _("error #%1").arg(
						int_to_str(ret)));
	    return false;
	}
	return true;
    }

    strm.next_in = (z_const Bytef *)buf;
    strm.avail_in = count;

    Bytef out[BUFSIZ_OUT];

    while (strm.avail_in != 0) {
	if (state == End) {
	    messages.error(_("trailing garbage"));
	    return false;
	}

	strm.next_out = (Bytef *)out;
	strm.avail_out = sizeof out;

	int ret = inflate(&strm, Z_NO_FLUSH);
	if (strm.msg)
	    messages.error(_("%1 says: %2").arg("inflate", strm.msg));

	switch (ret) {
	case Z_OK:
	    /* nice */
	    break;
	case Z_STREAM_END:
	    state = End;
	    break;
	case Z_NEED_DICT:
	    messages.error("inflate: " + _("preset dictionary is needed"));
	    return false;
	case Z_DATA_ERROR:
	    messages.error("inflate: " + _("corrupted input data"));
	    return false;
	case Z_STREAM_ERROR:
	    messages.error("inflate: " + _("inconsistent stream structure"));
	    return false;
	case Z_MEM_ERROR:
	    messages.error("inflate: " + _("not enough memory"));
	    return false;
	case Z_BUF_ERROR:
	    /* no progress is possible */
	    break;
	default:
	    messages.error("inflate: " + _("error #%1").arg(int_to_str(ret)));
	    return false;
	}

	if (strm.next_out != out && !sink->read(out, strm.next_out - out))
	    return false;
    }

    return true;
}

Shared::Sink *new_z_sink(Shared::Sink *sink, bool gzip_format,
			 Messages &messages)
{
    return new ZSink(sink, gzip_format, messages);
}

#else

Shared::Sink *new_z_sink(Shared::Sink *sink, bool /*gzip_format*/,
			 Messages &messages)
{
    messages.error(_("zlib support disabled during configuration"));
    delete sink;
    return NULL;
}

#endif /* HAVE_LIBZ */
