/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_attribute.h"
#include "icons.h"
#include "variant.h"
#include "xml_writer.h"
#include "drawing.h"
#include "../config.h"
#include "../gettext.h"

#define PN_(msgctxt, msgid) (msgid)

namespace SchematicEnum {
    List<String> boolean = List<String>()
	<< PN_("schematic boolean", "no")
	<< PN_("schematic boolean", "yes");
    List<String> capstyle = List<String>()
	<< PN_("schematic capstyle", "none")
	<< PN_("schematic capstyle", "square")
	<< PN_("schematic capstyle", "round");
    List<String> dashstyle = List<String>()
	<< PN_("schematic dashstyle", "solid")
	<< PN_("schematic dashstyle", "dotted")
	<< PN_("schematic dashstyle", "dashed")
	<< PN_("schematic dashstyle", "center")
	<< PN_("schematic dashstyle", "phantom");
    List<String> filltype = List<String>()
	<< PN_("schematic filltype", "hollow")
	<< PN_("schematic filltype", "fill")
	<< PN_("schematic filltype", "mesh")
	<< PN_("schematic filltype", "hatch")
	<< PN_("schematic filltype", "void");
    List<String> show_name_value = List<String>()
	<< PN_("schematic show_name_value", "name value")
	<< PN_("schematic show_name_value", "value")
	<< PN_("schematic show_name_value", "name");
    List<String> alignment = List<String>()
	<< PN_("schematic alignment", "lower left")
	<< PN_("schematic alignment", "middle left")
	<< PN_("schematic alignment", "upper left")
	<< PN_("schematic alignment", "lower middle")
	<< PN_("schematic alignment", "middle middle")
	<< PN_("schematic alignment", "upper middle")
	<< PN_("schematic alignment", "lower right")
	<< PN_("schematic alignment", "middle right")
	<< PN_("schematic alignment", "upper right");
    List<String> pintype = List<String>()
	<< PN_("schematic pintype", "normal")
	<< PN_("schematic pintype", "bus");
    List<String> color = List<String>()
	<< PN_("schematic color", "background")
	<< PN_("schematic color", "pin")
	<< PN_("schematic color", "net_endpoint")
	<< PN_("schematic color", "graphic")
	<< PN_("schematic color", "net")
	<< PN_("schematic color", "attribute")
	<< PN_("schematic color", "logic_bubble")
	<< PN_("schematic color", "dots_grid")
	<< PN_("schematic color", "detached_attribute")
	<< PN_("schematic color", "text")
	<< PN_("schematic color", "bus")
	<< PN_("schematic color", "select")
	<< PN_("schematic color", "boundingbox")
	<< PN_("schematic color", "zoom_box")
	<< PN_("schematic color", "stroke")
	<< PN_("schematic color", "lock")
	<< PN_("schematic color", "output_background")
	<< PN_("schematic color", "freestyle1")
	<< PN_("schematic color", "freestyle2")
	<< PN_("schematic color", "freestyle3")
	<< PN_("schematic color", "freestyle4")
	<< PN_("schematic color", "junction")
	<< PN_("schematic color", "mesh_grid_major")
	<< PN_("schematic color", "mesh_grid_minor");
}

SchematicAttributes *schematicAttributes;


static List<Item *> build_items(const List<String> &values,
				const char *msgctxt)
{
    List<Item *> items;
    foreach (String str, values)
	items.append(new StandardItem(pgettext_expr(msgctxt, str.data()),
				      icons::tree_leaf));
    return items;
}

static List<Item *> build_angle_items()
{
    List<Item *> items;
    items.append(new StandardItem(_("0 deg"), icons::tree_leaf));
    items.append(new StandardItem(_("90 deg"), icons::tree_leaf));
    items.append(new StandardItem(_("180 deg"), icons::tree_leaf));
    items.append(new StandardItem(_("270 deg"), icons::tree_leaf));
    return items;
}

SchematicAttributes::SchematicAttributes()
  : color(new Attribute(_("Color"), icons::color)),
    lineWidth(new Attribute(_("Width"), icons::line_width)),
    capStyle(new Attribute(_("Cap style"))),
    dashStyle(new Attribute(_("Dash style"), icons::line_style)),
    dashLength(new Attribute(_("Dash length"))),
    dashSpace(new Attribute(_("Dash spacing"))),
    fillType(new Attribute(_("Type"))),
    fillWidth(new Attribute(_("Width"))),
    fillAngle0(new Attribute(_("First angle"))),
    fillPitch0(new Attribute(_("First pitch/spacing"))),
    fillAngle1(new Attribute(_("Second angle"))),
    fillPitch1(new Attribute(_("Second pitch/spacing"))),
    isBus(new Attribute(_("Bus"))),
    isPin(new Attribute(_("Pin"))),
    isInverted(new Attribute(_("Inverted"))),
    textSize(new Attribute(_("Size"))),
    visibility(new Attribute(_("Visibility"))),
    showNameValue(new Attribute(_("Attribute visibility"))),
    angle(new Attribute(_("Angle"))),
    alignment(new Attribute(_("Alignment/origin"))),
    selectable(new Attribute(_("Selectable"))),
    mirror(new Attribute(_("Mirror"))),

    colorEnum(build_items(SchematicEnum::color,
			  "schematic color"), true),
    capStyleEnum(build_items(SchematicEnum::capstyle,
			     "schematic capstyle"), true),
    dashStyleEnum(build_items(SchematicEnum::dashstyle,
			      "schematic dashstyle"), true),
    fillTypeEnum(build_items(SchematicEnum::filltype,
			     "schematic filltype"), true),
    showNameValueEnum(build_items(SchematicEnum::show_name_value,
				  "schematic show_name_value"), true),
    angleEnum(build_angle_items(), true),
    alignmentEnum(build_items(SchematicEnum::alignment,
			      "schematic alignment"), true)
{
    // SchematicEnum::pintype
}

SchematicAttributeItems::SchematicAttributeItems(Hub &hub)
    : color(hub, *schematicAttributes->color,
	    schematicAttributes->colorEnum),
      lineWidth(hub, *schematicAttributes->lineWidth),
      capStyle(hub, *schematicAttributes->capStyle,
	       schematicAttributes->capStyleEnum),
      dashStyle(hub, *schematicAttributes->dashStyle,
		schematicAttributes->dashStyleEnum),
      dashLength(hub, *schematicAttributes->dashLength),
      dashSpace(hub, *schematicAttributes->dashSpace),
      fillType(hub, *schematicAttributes->fillType,
	       schematicAttributes->fillTypeEnum),
      fillWidth(hub, *schematicAttributes->fillWidth),
      fillAngle0(hub, *schematicAttributes->fillAngle0),
      fillPitch0(hub, *schematicAttributes->fillPitch0),
      fillAngle1(hub, *schematicAttributes->fillAngle1),
      fillPitch1(hub, *schematicAttributes->fillPitch1),
      isBus(hub, *schematicAttributes->isBus),
      isPin(hub, *schematicAttributes->isPin),
      isInverted(hub, *schematicAttributes->isInverted),
      textSize(hub, *schematicAttributes->textSize),
      visibility(hub, *schematicAttributes->visibility),
      showNameValue(hub, *schematicAttributes->showNameValue,
		    schematicAttributes->showNameValueEnum),
      angle(hub, *schematicAttributes->angle,
	    schematicAttributes->angleEnum),
      alignment(hub, *schematicAttributes->alignment,
		schematicAttributes->alignmentEnum),
      selectable(hub, *schematicAttributes->selectable),
      mirror(hub, *schematicAttributes->mirror)
{
}

/****************************************************************************/

SchematicLine::SchematicLine(Drawing *drawing)
    : drawing(drawing), width(0), cap_style(0), dash_style(0),
      dash_length(0), dash_space(0)
{
}

List<Attribute *> SchematicLine::attributes() const
{
    return List<Attribute *>() << schematicAttributes->lineWidth
			       << schematicAttributes->capStyle
			       << schematicAttributes->dashStyle
			       << schematicAttributes->dashLength
			       << schematicAttributes->dashSpace;
}

Variant SchematicLine::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->lineWidth)
	return Variant::fromMetrum(width);

    if (attribute == schematicAttributes->capStyle)
	return Variant::fromInt(cap_style);

    if (attribute == schematicAttributes->dashStyle)
	return Variant::fromInt(dash_style);

    if (attribute == schematicAttributes->dashLength) {
	if (dash_style == 0 || dash_style == 1)
	    return Variant::fromMetrum(0);

	return Variant::fromMetrum(dash_length);
    }

    if (attribute == schematicAttributes->dashSpace) {
	if (dash_style == 0)
	    return Variant::fromMetrum(0);

	return Variant::fromMetrum(dash_space);
    }

    return Variant();
}

Variant SchematicLine::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->lineWidth)
	return Variant::fromMetrum(0);
    if (attribute == schematicAttributes->capStyle)
	return Variant::fromInt(0);
    if (attribute == schematicAttributes->dashStyle)
	return Variant::fromInt(0);
    if (attribute == schematicAttributes->dashLength)
	return Variant::fromMetrum(0);
    if (attribute == schematicAttributes->dashSpace)
	return Variant::fromMetrum(0);

    return Variant();
}

void SchematicLine::setAttribute(const Variant &value,
				 const Attribute *attribute)
{
    if (attribute == schematicAttributes->lineWidth)
	width = value.toMetrum();
    if (attribute == schematicAttributes->capStyle)
	cap_style = value.toInt();
    if (attribute == schematicAttributes->dashStyle)
	dash_style = value.toInt();
    if (attribute == schematicAttributes->dashLength)
	dash_length = value.toMetrum();
    if (attribute == schematicAttributes->dashSpace)
	dash_space = value.toMetrum();
}

void SchematicLine::write(XMLWriter *writer)
{
    writer->writeAttribute("width", drawing->format(width));
    writer->writeAttribute("capstyle", SchematicEnum::capstyle[cap_style]);
    writer->writeAttribute("dashstyle", SchematicEnum::dashstyle[dash_style]);
    if (dash_style != 0 && dash_style != 1)
	writer->writeAttribute("dashlength", drawing->format(dash_length));
    if (dash_style != 0)
	writer->writeAttribute("dashspace", drawing->format(dash_space));
}

/****************************************************************************/

SchematicFill::SchematicFill(Drawing *drawing)
    : drawing(drawing), type(0), width(0),
      angle0(0), pitch0(0), angle1(0), pitch1(0)
{
}

List<Attribute *> SchematicFill::attributes() const
{
    return List<Attribute *>() << schematicAttributes->fillType
			       << schematicAttributes->fillWidth
			       << schematicAttributes->fillAngle0
			       << schematicAttributes->fillPitch0
			       << schematicAttributes->fillAngle1
			       << schematicAttributes->fillPitch1;
}

Variant SchematicFill::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->fillType)
	return Variant::fromInt(type);

    if (attribute == schematicAttributes->fillWidth) {
	if (type != 2 && type != 3)
	    return Variant::fromMetrum(0);
	return Variant::fromMetrum(width);
    }

    if (attribute == schematicAttributes->fillAngle0) {
	if (type != 2 && type != 3)
	    return Variant::fromInt(0);
	return Variant::fromInt(angle0);
    }

    if (attribute == schematicAttributes->fillPitch0) {
	if (type != 2 && type != 3)
	    return Variant::fromMetrum(0);
	return Variant::fromMetrum(pitch0);
    }

    if (attribute == schematicAttributes->fillAngle1) {
	if (type != 2)
	    return Variant::fromInt(0);
	return Variant::fromInt(angle1);
    }

    if (attribute == schematicAttributes->fillPitch1) {
	if (type != 2)
	    return Variant::fromMetrum(0);
	return Variant::fromMetrum(pitch1);
    }

    return Variant();
}

Variant SchematicFill::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->fillType)
	return Variant::fromInt(0);
    if (attribute == schematicAttributes->fillWidth)
	return Variant::fromMetrum(0);
    if (attribute == schematicAttributes->fillAngle0)
	return Variant::fromInt(0);
    if (attribute == schematicAttributes->fillPitch0)
	return Variant::fromMetrum(0);
    if (attribute == schematicAttributes->fillAngle1)
	return Variant::fromInt(0);
    if (attribute == schematicAttributes->fillPitch1)
	return Variant::fromMetrum(0);

    return Variant();
}

void SchematicFill::setAttribute(const Variant &value,
				 const Attribute *attribute)
{
    if (attribute == schematicAttributes->fillType)
	type = value.toInt();
    if (attribute == schematicAttributes->fillWidth)
	width = value.toMetrum();
    if (attribute == schematicAttributes->fillAngle0)
	angle0 = value.toInt();
    if (attribute == schematicAttributes->fillPitch0)
	pitch0 = value.toMetrum();
    if (attribute == schematicAttributes->fillAngle1)
	angle1 = value.toInt();
    if (attribute == schematicAttributes->fillPitch1)
	pitch1 = value.toMetrum();
}

void SchematicFill::write(XMLWriter *writer)
{
    writer->writeAttribute("filltype", SchematicEnum::filltype[type]);
    if (type == 2 || type == 3) {
	writer->writeAttribute("fillwidth", drawing->format(width));
	writer->writeAttribute("angle0", int_to_str(angle0));
	writer->writeAttribute("pitch0", drawing->format(pitch0));
    }
    if (type == 2) {
	writer->writeAttribute("angle1", int_to_str(angle1));
	writer->writeAttribute("pitch1", drawing->format(pitch1));
    }
}
