/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <math.h>
#include "sch_readsch.h"
#include "dirmodel.h"
#include "variant.h"
#include "sch_box.h"
#include "sch_circle.h"
#include "sch_component.h"
#include "sch_line.h"
#include "sch_net.h"
#include "sch_path.h"
#include "sch_picture.h"
#include "sch_text.h"

#define VERSION_20000220 20000220
#define VERSION_20000704 20000704
#define VERSION_20020825 20020825
#define VERSION_20030921 20030921

#define INVALID(what, val) \
    error(_("invalid " what " (%1)").arg(int_to_str(val)))


class PlaintextSink : public Shared::Sink
{
    unsigned int lineno;
    char *buf;
    size_t size, filled;
    void digest_line(char *ptr, size_t len);
    virtual void readLine(const char *s) = 0;
    virtual void done() = 0;
protected:
    Messages &messages;
public:
    PlaintextSink(Messages &messages) : lineno(0), buf(0), size(0), filled(0),
					messages(messages) {
	buf = (char *)malloc(size);
    }
    ~PlaintextSink() {
	free(buf);
    }
    bool read(const void *buf, ssize_t count);
};

bool PlaintextSink::read(const void *src, ssize_t count)
{
    if (count == 0) {
	if (filled != 0) {
	    messages.append(Message(Message::Warning, _("line not terminated"),
				    lineno, filled));
	    if (size < filled + 1) {
		size = filled + 1;
		buf = (char *)realloc(buf, size);
	    }
	    digest_line(buf, filled);
	}
	done();
	return true;
    }

    if (size < filled + count) {
	size = filled + count;
	buf = (char *)realloc(buf, size);
    }
    memcpy(buf + filled, src, count);
    filled += count;

    size_t used = 0;
    while (used < filled) {
	char *newline = (char *)memchr(buf + used, '\n', filled - used);
	if (newline == NULL) {
	    memmove(buf, buf + used, filled - used);
	    break;
	}

	digest_line(buf + used, newline - buf - used);
	used = newline - buf + 1;
    }
    filled -= used;
    return true;
}

void PlaintextSink::digest_line(char *buf, size_t len)
{
    if (len != 0 && buf[len - 1] == '\r')
	len--;
    buf[len] = '\0';

    if (strlen(buf) != len) {
	messages.append(Message(Message::Error, _("stray null byte"),
				lineno, strlen(buf)));
	lineno++;
	return;
    }

    const char *ptr = buf;
    mbstate_t ps;
    bzero(&ps, sizeof ps);

    if (mbsrtowcs(0, &ptr, 0, &ps) == (size_t)-1) {
	if (errno != EILSEQ)
	    warn("mbsrtowcs");
	messages.append(Message(Message::Error, _("data is not valid UTF-8"),
				lineno, ptr - buf));
	lineno++;
	return;
    }

    readLine(buf);
    lineno++;
}

/****************************************************************************/

class SchematicSink : public PlaintextSink
{
    Drawing *const drawing;

    unsigned int release_ver;
    unsigned int fileformat_ver;
    Object *object_to_attach_to;
    List<Object *> new_object_list;
    List<ObjectContext *> stack;
    List<NetObject *> pins;
    ComponentObject *last_complex;

    enum ParserState { Base, Text, PicturePath, PictureData, Path };
    TextObject *text_object;
    PictureObject *picture_object;
    Path1Object *path_object;
    unsigned int num_lines;

    ParserState state;

    ObjectContext *attribute_list;
    unsigned int lineno;
    List<String> encoded_picture;
    ComponentObject *complex_to_embed_to;

    void validate_color(int &color, Object *object);
    void validate_line(SchematicLine &line, Metrum width,
		       int cap_style, int dash_style,
		       Metrum dash_length, Metrum dash_space);
    void validate_fill(SchematicFill &fill, int type, Metrum width,
		       int angle0, Metrum pitch0,
		       int angle1, Metrum pitch1);

    LineObject *read_line(const char buf[]);
    NetObject *read_net(const char buf[]);
    NetObject *read_bus(const char buf[]);
    BoxObject *read_box(const char buf[]);
    PictureObject *read_picture(const char buf[]);
    CircleObject *read_circle(const char buf[]);
    ComponentObject *read_complex(const char buf[]);
    TextObject *read_text(const char buf[]);
    Path1Object *read_path(const char buf[]);
    NetObject *read_pin(const char buf[]);
    CircleObject *read_arc(const char buf[]);

    void check_symversion();
    void check_symversion1(const String &inside, bool inside_present,
			   const String &outside, bool outside_present,
			   const String &refdes, const String &name,
			   const String &complex_basename);
    void check_symversion2(double inside_value, bool inside_present,
			   double outside_value, bool outside_present,
			   const String &refdes, const String &name,
			   const String &complex_basename);
    void guess_pin_orientation();
    void error(const String &message);
    void error(unsigned int column, const String &message);
    void warn(const String &message);
    void warn(unsigned int column, const String &message);
    void readLine(const char *line);
    void done();
public:
    SchematicSink(Drawing *drawing, Messages &messages);
};

SchematicSink::SchematicSink(Drawing *drawing, Messages &messages)
    : PlaintextSink(messages), drawing(drawing),
      release_ver(0),
      fileformat_ver(0),
      object_to_attach_to(NULL),
      last_complex(NULL),
      text_object(NULL),
      picture_object(NULL),
      path_object(NULL),
      num_lines(0),
      state(Base),
      attribute_list(NULL),
      lineno(-1),
      complex_to_embed_to(NULL)
{
}

void SchematicSink::error(const String &message)
{
    messages.append(Message(Message::Error, message, lineno, -1));
}

void SchematicSink::error(unsigned int column, const String &message)
{
    messages.append(Message(Message::Error, message, lineno, column));
}

void SchematicSink::warn(const String &message)
{
    messages.append(Message(Message::Warning, message, lineno, -1));
}

void SchematicSink::warn(unsigned int column, const String &message)
{
    messages.append(Message(Message::Warning, message, lineno, column));
}

void SchematicSink::validate_color(int &color, Object *object)
{
    if (color < 0 || color >= SchematicEnum::color.size()) {
	INVALID("color", color);
	color = object->attributeDefault(schematicAttributes->color).toInt();
    }
}

void SchematicSink::validate_line(SchematicLine &line, Metrum width,
				  int cap_style, int dash_style,
				  Metrum dash_length, Metrum dash_space)
{
    if (width < 0) {
	error(_("negative line width not allowed (%1)")
	      .arg(int_to_str(width)));
	width = 0;
    }
    line.width = width;

    if (cap_style < 0 || cap_style >= SchematicEnum::capstyle.size()) {
	INVALID("cap style", cap_style);
	cap_style = 0;
    }
    line.cap_style = cap_style;

    if (dash_style < 0 || dash_style >= SchematicEnum::dashstyle.size()) {
	INVALID("dash style", dash_style);
	dash_style = 0;
    }
    line.dash_style = dash_style;

    if (dash_length < -1) {
	error(_("negative dash length not allowed (%1)")
	      .arg(int_to_str(dash_length)));
	dash_length = 0;
    }
    line.dash_length = dash_length;

    if (dash_space < -1) {
	error(_("negative dash space not allowed (%1)")
	      .arg(int_to_str(dash_space)));
	dash_space = 0;
    }
    line.dash_space = dash_space;
}

void SchematicSink::validate_fill(SchematicFill &fill, int type, Metrum width,
				  int angle0, Metrum pitch0,
				  int angle1, Metrum pitch1)
{
    if (type < 0 || type >= SchematicEnum::filltype.size()) {
	INVALID("fill type", type);
	type = 0;
    }
    fill.type = type;

    if (width < -1) {
	INVALID("fill width", width);
	width = 0;
    }
    fill.width = width;

    fill.angle0 = angle0;
    fill.pitch0 = pitch0;
    fill.angle1 = angle1;
    fill.pitch1 = pitch1;
}

LineObject *SchematicSink::read_line(const char buf[])
{
    int x1, y1;
    int x2, y2;
    int line_width, line_space, line_length;
    int line_end;
    int line_type;
    int color;

    if (release_ver <= VERSION_20000704) {
	/* The old geda file format, i.e. releases 20000704 and older,
	 * does not handle the line type and the filling - here
	 * filling is irrelevant.  They are set to default. */
	if (sscanf(buf, "L %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color) != 5) {
	    error(_("failed to parse line object"));
	    return NULL;
	}

	line_width = 0;
	line_end = 0;
	line_type = 0;
	line_length = -1;
	line_space = -1;
    } else {
	/* The current line format to describe a line is a space
	 * separated list of characters and numbers in plain ASCII on
	 * a single line.  The meaning of each item is described in
	 * the file format documentation. */
	if (sscanf(buf, "L %d %d %d %d %d %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color, &line_width, &line_end,
		   &line_type, &line_length, &line_space) != 10) {
	    error(_("failed to parse line object"));
	    return NULL;
	}
    }

    LineObject *line = new LineObject(Metrum3(x1, y1, 0),
				      Metrum3(x2, y2, 0), drawing);
    if (line->p0 == line->p1)
	warn(_("line has zero length"));
    validate_color(line->color = color, line);
    validate_line(line->line, line_width, line_end,
		  line_type, line_length, line_space);
    return line;
}

NetObject *SchematicSink::read_net(const char buf[])
{
    int x1, y1;
    int x2, y2;
    int color;

    if (sscanf(buf, "N %d %d %d %d %d\n",
	       &x1, &y1, &x2, &y2, &color) != 5) {
	error(_("failed to parse net object"));
	return NULL;
    }

/** if (toplevel->override_net_color != -1)
	color = toplevel->override_net_color; **/

    NetObject *net = new NetObject(Metrum3(x1, y1, 0),
				   Metrum3(x2, y2, 0), drawing);
    if (net->p0 == net->p1)
	warn(_("net has zero length"));
    validate_color(net->color = color, net);
    return net;
}

NetObject *SchematicSink::read_bus(const char buf[])
{
    int x1, y1;
    int x2, y2;
    int color;
    int ripper_dir;

    if (release_ver <= VERSION_20020825) {
	if (sscanf(buf, "U %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color) != 5) {
	    error(_("failed to parse bus object"));
	    return NULL;
	}
	ripper_dir = 0;
    } else {
	if (sscanf(buf, "U %d %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color, &ripper_dir) != 6) {
	    error(_("failed to parse bus object"));
	    return NULL;
	}
    }

/** if (toplevel->override_bus_color != -1)
	color = toplevel->override_bus_color; **/

    NetObject *bus = new NetObject(Metrum3(x1, y1, 0),
				   Metrum3(x2, y2, 0), drawing);
    if (bus->p0 == bus->p1)
	warn(_("bus has zero length"));

    validate_color(bus->color = color, bus);

    bus->is_bus = true;

    if (ripper_dir < -1 || ripper_dir > 1) {
	INVALID("bus ripper direction", ripper_dir);
	ripper_dir = 0;
    }
    /** ripper_dir **/

    return bus;
}

BoxObject *SchematicSink::read_box(const char buf[])
{
    int x1, y1;
    int width, height;
    int color;
    int box_width, box_space, box_length;
    int fill_width, angle1, pitch1, angle2, pitch2;
    int box_end;
    int box_type;
    int box_filling;

    if (release_ver <= VERSION_20000704) {
	/* The old geda file format, i.e. releases 20000704 and older,
	 * does not handle the line type and the filling of the box
	 * object. They are set to default. */
	if (sscanf(buf, "B %d %d %d %d %d\n",
		   &x1, &y1, &width, &height, &color) != 5) {
	    error(_("failed to parse box object"));
	    return NULL;
	}

	box_width = 0;
	box_end = 0;
	box_type = 0;
	box_length = -1;
	box_space = -1;

	box_filling = 0;
	fill_width = 0;
	angle1 = -1;
	pitch1 = -1;
	angle2 = -1;
	pitch2 = -1;
    } else {
	/* The current line format to describe a box is a space
	 * separated list of characters and numbers in plain ASCII on
	 * a single line. The meaning of each item is described in the
	 * file format documentation. */
	if (sscanf(buf, "B %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
		   &x1, &y1, &width, &height, &color, &box_width,
		   &box_end, &box_type, &box_length, &box_space, &box_filling,
		   &fill_width, &angle1, &pitch1, &angle2, &pitch2) != 16) {
	    error(_("failed to parse box object"));
	    return NULL;
	}
    }

    if (width == 0 || height == 0)
	warn(_("box has zero width/height"));

    BoxObject *box = new BoxObject(Metrum3(x1, y1, 0),
				   Metrum3(x1 + width,
					   y1 + height, 0), drawing);
    validate_color(box->color = color, box);
    validate_line(box->line, box_width, box_end,
		  box_type, box_length, box_space);
    validate_fill(box->fill, box_filling, fill_width,
		  angle1, pitch1, angle2, pitch2);
    return box;
}

PictureObject *SchematicSink::read_picture(const char buf[])
{
    int x1, y1;
    int width, height, angle;
    int mirrored, embedded;

    if (sscanf(buf, "G %d %d %d %d %d %d %d\n",
	       &x1, &y1, &width, &height, &angle, &mirrored, &embedded) != 7) {
	error(_("failed to parse picture definition"));
	return NULL;
    }

    if (width == 0 || height == 0)
	warn(_("picture has zero width/height"));

    PictureObject *picture = new PictureObject(
	Metrum3(x1, y1, 0),
	Metrum3(x1 + width,
		y1 + height, 0), drawing);

    if (angle != 0 && angle != 90 && angle != 180 && angle != 270) {
	error(_("invalid picture angle (%1 degrees)").arg(int_to_str(angle)));
	angle = 0;
    }
    picture->angle = angle / 90;

    if (mirrored != 0 && mirrored != 1) {
	INVALID("picture 'mirrored' flag", mirrored);
	mirrored = 0;
    }
    picture->mirrored = mirrored;

    if (embedded != 0 && embedded != 1) {
	INVALID("picture 'embedded' flag", embedded);
	embedded = 0;
    }
    picture->embedded = embedded;

    state = PicturePath;
    picture_object = picture;
    return picture;
}

CircleObject *SchematicSink::read_circle(const char buf[])
{
    int x1, y1;
    int radius;
    int color;
    int circle_width, circle_space, circle_length;
    int fill_width, angle1, pitch1, angle2, pitch2;
    int circle_end;
    int circle_type;
    int circle_fill;

    if (release_ver <= VERSION_20000704) {
	/* The old geda file format, i.e. releases 20000704 and older,
	 * does not handle the line type and the filling of the box
	 * object. They are set to default. */
	if (sscanf(buf, "V %d %d %d %d\n",
		   &x1, &y1, &radius, &color) != 4) {
	    error(_("failed to parse circle object"));
	    return NULL;
	}

	circle_width = 0;
	circle_end = 0;
	circle_type = 0;
	circle_length = -1;
	circle_space = -1;

	circle_fill = 0;
	fill_width = 0;
	angle1 = -1;
	pitch1 = -1;
	angle2 = -1;
	pitch2 = -1;
    } else {
	/* The current line format to describe a circle is a space
	 * separated list of characters and numbers in plain ASCII on
	 * a single line. The meaning of each item is described in the
	 * file format documentation. */
	if (sscanf(buf, "V %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
		   &x1, &y1, &radius, &color,
		   &circle_width, &circle_end, &circle_type,
		   &circle_length, &circle_space, &circle_fill,
		   &fill_width, &angle1, &pitch1, &angle2, &pitch2) != 15) {
	    error(_("failed to parse circle object"));
	    return NULL;
	}
    }

    if (radius == 0)
	warn(_("circle has zero radius"));
    else if (radius < 0) {
	error(_("circle has negative radius (%1)").arg(int_to_str(radius)));
	radius = 0;
    }

    CircleObject *circle = new CircleObject(
	Metrum3(x1, y1, 0), radius, drawing);
    validate_color(circle->color = color, circle);
    validate_line(circle->line, circle_width, circle_end,
		  circle_type, circle_length, circle_space);
    validate_fill(circle->fill, circle_fill, fill_width,
		  angle1, pitch1, angle2, pitch2);
    return circle;
}

static void o_complex_remove_promotable_attribs(ComponentObject *object)
{
/*
    if (!toplevel->attribute_promotion) / * controlled through rc file * /
	return;

    foreach (Object *a_object,
	     o_attrib_find_floating_attribs(object->complex->prim_objs)) {
	/ * Is it an attribute we want to promote? * /
	if (!o_complex_is_eligible_attribute(toplevel, a_object))
	    continue;

	if (toplevel->keep_invisible) {
	    / * Hide promotable attributes * /
	    a_object->set_visibility(INVISIBLE);
	} else {
	    / * Delete promotable attributes * /
	    object->complex->prim_objs->removeObject(a_object);
	    a_object->s_delete();
	}
    }

    o_bounds_invalidate(object);
*/
}

ComponentObject *SchematicSink::read_complex(const char buf[])
{
    char type;
    int x1, y1;
    int angle;

    char basename[strlen(buf) + 1];

    int selectable;
    int mirror;

    if (sscanf(buf, "%c %d %d %d %d %d %s\n", &type,
	       &x1, &y1, &selectable, &angle, &mirror, basename) != 7) {
	error(_("failed to parse complex object"));
	return NULL;
    }

    ComponentObject *component =
	new ComponentObject(Metrum3(x1, y1, 0), drawing);
    /** component->type = type; **/

    /* Delete or hide attributes eligible for promotion inside the complex */
    o_complex_remove_promotable_attribs(component);

    if (selectable != 0 && selectable != 1) {
	INVALID("component 'selectable' flag", selectable);
	selectable = 0;
    }
    component->selectable = selectable;

    if (angle != 0 && angle != 90 && angle != 180 && angle != 270) {
	error(_("invalid component rotation (%1 degrees)")
	      .arg(int_to_str(angle)));
	angle = 0;
    }
    component->angle = angle / 90;

    if (mirror != 0 && mirror != 1) {
	INVALID("component mirror flag", mirror);
	mirror = 0;
    }
    component->mirror = mirror;

    if (strncmp(basename, "EMBEDDED", 8) == 0) {
	component->is_embedded = true;
	component->basename = basename + 8;
    } else {
	/* component->clib = s_clib_get_symbol_by_name(basename); */
	component->basename = basename;

	component->contents = shared.component(basename);
	if (!component->contents)
	    messages.warning(_("Component not found: %1")
				 .arg(component->basename));
    }

    return component;
}

TextObject *SchematicSink::read_text(const char buf[])
{
    int x, y;
    int color;
    int size;
    int visibility;
    int show_name_value;
    int angle;
    int alignment;
    int num_lines = 0;

    if (fileformat_ver >= 1) {
	if (sscanf(buf, "T %d %d %d %d %d %d %d %d %d\n",
		   &x, &y, &color, &size, &visibility, &show_name_value,
		   &angle, &alignment, &num_lines) != 9) {
	    error(_("failed to parse text object"));
	    return NULL;
	}
    } else if (release_ver < VERSION_20000220) {
	/* yes, above less than (not less than and equal) is correct.
	 * The format change occurred in 20000220 */
	if (sscanf(buf, "T %d %d %d %d %d %d %d\n",
		   &x, &y, &color, &size, &visibility, &show_name_value,
		   &angle) != 7) {
	    error(_("failed to parse text object"));
	    return NULL;
	}
	alignment = 0; /* older versions didn't have this */
	num_lines = 1; /* only support a single line */
    } else {
	if (sscanf(buf, "T %d %d %d %d %d %d %d %d\n",
		   &x, &y, &color, &size, &visibility, &show_name_value,
		   &angle, &alignment) != 8) {
	    error(_("failed to parse text object"));
	    return NULL;
	}
	num_lines = 1; /* only support a single line */
    }

    TextObject *text = new TextObject(Metrum3(x, y, 0),
				      List<String>(), drawing);
    validate_color(text->color = color, text);

    if (size == 0)
	warn(_("text has zero size"));
    else if (size < 0) {
	error(_("text has negative size (%1)").arg(int_to_str(size)));
	size = 0;
    }
    text->size = size;

    if (visibility != 0 && visibility != 1) {
	INVALID("text visibility flag", visibility);
	visibility = 0;
    }
    text->visibility = visibility;

    if (alignment < 0 || alignment >= SchematicEnum::alignment.size()) {
	INVALID("text alignment", alignment);
	alignment = 0;
    }
    text->alignment = alignment;

    if (angle != 0 && angle != 90 && angle != 180 && angle != 270) {
	error(_("invalid text angle (%1 degrees)").arg(int_to_str(angle)));
	angle = 0;
    }
    text->angle = angle / 90;

    if (show_name_value < 0 ||
	show_name_value >= SchematicEnum::show_name_value.size()) {
	INVALID("text name/value visibility", show_name_value);
	show_name_value = 0;
    }
    text->show_name_value = show_name_value;

    if (num_lines <= 0)
	error(_("text has invalid number of lines (%1)")
	      .arg(int_to_str(num_lines)));
    else {
	state = Text;
	text_object = text;
	this->num_lines = num_lines;
    }
    return text;
}

Path1Object *SchematicSink::read_path(const char buf[])
{
    int color;
    int line_width, line_space, line_length;
    int line_end;
    int line_type;
    int fill_type, fill_width, angle1, pitch1, angle2, pitch2;
    int num_lines = 0;

    /* The current path format to describe a line is a space separated
     * list of characters and numbers in plain ASCII on a single path.
     * The meaning of each item is described in the file format
     * documentation. */

    if (sscanf(buf, "H %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
	       &color, &line_width, &line_end, &line_type,
	       &line_length, &line_space, &fill_type, &fill_width, &angle1,
	       &pitch1, &angle2, &pitch2, &num_lines) != 13) {
	error(_("failed to parse path object"));
	return NULL;
    }

    Path1Object *path = new Path1Object(drawing);

    validate_color(path->color = color, path);

    validate_line(path->line, line_width, line_end,
		  line_type, line_length, line_space);
    validate_fill(path->fill, fill_type, fill_width,
		  angle1, pitch1, angle2, pitch2);

    if (num_lines <= 0)
	error(_("path has invalid number of data lines (%1)")
	      .arg(int_to_str(num_lines)));
    else {
	state = Path;
	this->num_lines = num_lines;
    }
    return path;
}

NetObject *SchematicSink::read_pin(const char buf[])
{
    int x1, y1;
    int x2, y2;
    int color;
    int pin_type;
    int whichend;

    if (release_ver <= VERSION_20020825) {
	if (sscanf(buf, "P %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color) != 5) {
	    error(_("failed to parse pin object"));
	    return NULL;
	}
	pin_type = 0;
	whichend = -1;
    } else {
	if (sscanf(buf, "P %d %d %d %d %d %d %d\n",
		   &x1, &y1, &x2, &y2, &color, &pin_type, &whichend) != 7) {
	    error(_("failed to parse pin object"));
	    return NULL;
	}
    }

/** if (toplevel->override_pin_color != -1)
	color = toplevel->override_pin_color; **/

    NetObject *pin = new NetObject(Metrum3(x1, y1, 0),
				   Metrum3(x2, y2, 0), drawing);
    if (pin->p0 == pin->p1)
	warn(_("pin has zero length"));

    validate_color(pin->color = color, pin);

    if (pin_type != 0 && pin_type != -1) {
	INVALID("pin type", pin_type);
	pin_type = 0;
    }
    pin->is_bus = pin_type;

    pin->is_pin = true;

    if (whichend == -1) {
	warn(_("Found a pin which did not have the whichone field set. "
	       "Verify and correct manually."));
	whichend = 0;
    } else if (whichend != 0 && whichend != 1) {
	INVALID("pin 'whichend' flag", whichend);
	whichend = 0;
    }
    pin->is_inverted = whichend;
    if (pin->is_inverted)
	std::swap(pin->p0, pin->p1);

    return pin;
}

CircleObject *SchematicSink::read_arc(const char buf[])
{
    int x1, y1;
    int radius;
    int start_angle, end_angle;
    int color;
    int arc_width, arc_length, arc_space;
    int arc_type;
    int arc_end;

    /* Depending on the version of the file format used to describe this arc,
     * the buffer is parsed differently.  The unknown parameters of the less
     * restrictive - the oldest - file format are set to common values */

    if (release_ver <= VERSION_20000704) {
	if (sscanf(buf, "A %d %d %d %d %d %d",
		   &x1, &y1, &radius, &start_angle, &end_angle, &color) != 6) {
	    error(_("failed to parse arc object"));
	    return NULL;
	}

	arc_width = 0;
	arc_end = 0;
	arc_type = 0;
	arc_space = -1;
	arc_length = -1;
    } else {
	if (sscanf(buf, "A %d %d %d %d %d %d %d %d %d %d %d",
		   &x1, &y1, &radius, &start_angle, &end_angle, &color,
		   &arc_width, &arc_end, &arc_type, &arc_length, &arc_space)
	    != 11) {
	    error(_("failed to parse arc object"));
	    return NULL;
	}
    }

    if (radius == 0)
	warn(_("arc has zero radius"));
    else if (radius < 0) {
	error(_("arc has negative radius (%1)").arg(int_to_str(radius)));
	radius = 0;
    }

    CircleObject *arc = new CircleObject(Metrum3(x1, y1, 0), radius, drawing);
    arc->is_arc = true;
    arc->startangle = start_angle;
    arc->sweepangle = end_angle;
    validate_color(arc->color = color, arc);
    validate_line(arc->line, arc_width, arc_end,
		  arc_type, arc_length, arc_space);
    return arc;
}



void SchematicSink::check_symversion()
{
    if (!last_complex)
	return;

    /* verify symbol version (not file format but rather contents) */

    /*
    check_symversion1(
	last_complex->search_inherited_attribs_by_name("symversion", 0),
	last_complex->search_attached_attribs_by_name("symversion", 0),
	last_complex->search_object_attribs_by_name/w_default("refdes", 0,
							      "unknown"),
	last_complex->name, last_complex->complex_basename);
    */

    last_complex = NULL;  /* no longer need to check */
}

void SchematicSink::check_symversion1(
    const String &inside, bool inside_present,
    const String &outside, bool outside_present,
    const String &refdes, const String &name, const String &complex_basename)
{
    double inside_value = -1.0;
    double outside_value = -1.0;
    bool ok;

    if (inside_present) {
	inside_value = str_to_double(inside, ok);
	if (!ok) {
	    warn(_("symbol version parse error on refdes %1: "
		   "could not parse symbol file symversion=%2")
		 .arg(refdes, inside));
	    return;
	}
    }

    if (outside_present) {
	outside_value = str_to_double(outside, ok);
	if (!ok) {
	    warn(_("symbol version parse error on refdes %1: "
		   "could not parse attached symversion=%2")
		 .arg(refdes, outside));
	    return;
	}
    }

    check_symversion2(inside_value, inside_present,
		      outside_value, outside_present,
		      refdes, name, complex_basename);
}

void SchematicSink::check_symversion2(
    double inside_value, bool inside_present,
    double outside_value, bool outside_present,
    const String &refdes, const String &name, const String &complex_basename)
{
#if DEBUG
    printf("%s:\n\tinside: %.1f outside: %.1f\n\n",
	   name, inside_value, outside_value);
#endif

    if (!inside_present) {
	if (outside_present)
	    /* No symversion inside, but a version is outside,
	     * this is a weird case */
	    warn(_("symbol version oddity on refdes %1: "
		   "symversion=%2 attached to instantiated symbol, "
		   "but no symversion= inside symbol file")
		 .arg(refdes, double_to_str(outside_value)));

	/* If symversion= is not present anywhere,
	 * the symbol is legacy and versioned okay */
	return;
    }

    if (outside_present) {
	if (outside_value > inside_value) {
	    /* outside value is greater than inside value,
	     * this is weird case */
	    warn(_("symbol version oddity on refdes %1: "
		   "instantiated symbol is newer than symbol in library")
		 .arg(refdes));
	    return;
	}

	if (outside_value == inside_value)
	    /* if inside_value and outside_value match,
	     * then symbol versions are okay */
	    return;
    }

    /* inside & not outside is a valid case, means symbol in library
     * is newer */
    /* also if inside_value is greater than outside_value, then symbol
     * in library is newer */

    warn(_("symbol version mismatch on refdes %1 (%2): "
	   "symbol in library is newer than instantiated symbol")
	 .arg(refdes, complex_basename));

    /* break up the version values into major.minor numbers */
    double inside_major = floor(inside_value);
    double outside_major = outside_present ? floor(outside_value) : 0.;

#if DEBUG
    printf("i: %f %f %f\n", inside_value, inside_major);
    printf("o: %f %f %f\n", outside_value, outside_major);
#endif

    warn((inside_major > outside_major
	      ? _("MAJOR VERSION CHANGE (file %1, instantiated %2)")
	      : _("Minor version change (file %1, instantiated %2)"))
	 .arg(double_to_str(inside_value),
	      double_to_str(outside_value)));
}

String decode_base64(const List<String> &data, bool &ok)
{
    ok = true;
    return String(); /* TODO */
}

void SchematicSink::readLine(const char *line)
{
    //printf("==> Reading '%s'\n", line);
    lineno++;
    String file_content;
    int itemsread;

    switch (state) {
    case Base:
	break;
    case Text:
	text_object->text.append(line);
	if (text_object->text.count() == num_lines) {
	    text_object->update();
	    text_object = 0;
	    num_lines = 0;
	    state = Base;
	}
	return;
    case PicturePath:
	picture_object->filename = line;
	if (*line == '\0')
	    warn(_("image has no filename"));
	if (picture_object->embedded)
	    state = PictureData;
	else {
	    picture_object = 0;
	    state = Base;
	}
	return;
    case PictureData:
	if (line[0] != '.' || line[1] != '\0') {
	    encoded_picture.append(line);
	    return;
	}

	bool ok;
	file_content = decode_base64(encoded_picture, ok);
	encoded_picture.clear();

	if (!ok) {
	    error(_("failed to load embedded image: base64 decoding failed"));
	    picture_object->embedded = false;
	}

	picture_object = 0;
	state = Base;
	/* validate data? */
	return;
    case Path:
	path_object->data.append(line);
	if (path_object->data.count() == num_lines) {
	    path_object = 0;
	    num_lines = 0;
	    state = Base;
	    /* validate data? */
	}
	return;
    }

    /* Do we need to check the symbol version?  Yes, but only if
     * 1) the last object read was a complex and
     * 2) the next object isn't the start of attributes.
     * If the next object is the start of attributes, then check the
     * symbol version after the attributes have been read in, see the
     * STARTATTACH_ATTR case */
    if (*line != '{')
	check_symversion();

    Object *object = 0;

    if (attribute_list) {
	switch (*line) {
	case 'T': {
	    TextObject *text = read_text(line);
	    text->pos -= attribute_list->origin();
	    attribute_list->addObject(text);
	    break;
	}
	case '}':
	    /* by now we have finished reading all the attributes */
	    /* did we just finish attaching to a complex object? */
	    check_symversion();

	    /* slots only apply to complex objects */
	    /*
	    if (object_to_attach_to != NULL &&
		(object_to_attach_to->type == OBJ_COMPLEX ||
		 object_to_attach_to->type == OBJ_PLACEHOLDER)) {
		s_slot_update_object (object_to_attach_to);
	    }
	    */
	    object_to_attach_to = NULL;
	    attribute_list = 0;
	    break;
	default:
	    error(_("tried to attach a non-text item as an attribute"));
	}
	return;
    }

    switch (*line) {
    case 'L': object_to_attach_to = object = read_line(line); break;
    case 'N': object_to_attach_to = object = read_net(line); break;
    case 'U': object_to_attach_to = object = read_bus(line); break;
    case 'B': object_to_attach_to = object = read_box(line); break;
    case 'G': object_to_attach_to = object = read_picture(line); break;
    case 'V': object_to_attach_to = object = read_circle(line); break;

    case 'C': /* OBJ_COMPLEX */
    case 'X': /* OBJ_PLACEHOLDER */
	object_to_attach_to = object = last_complex = complex_to_embed_to = read_complex(line);
	/* last_complex is used for verifying symversion attribute */
	break;

    case 'T': object_to_attach_to = object = read_text(line); break;
    case 'H': object_to_attach_to = object = read_path(line); break;
    case 'P': pins.append(read_pin(line)); object_to_attach_to = object = pins.back(); break;
    case 'A': object_to_attach_to = object = read_arc(line); break;

    case '{': /* STARTATTACH_ATTR */
	/* first is the fp */
	/* 2nd is the object to get the attributes */
	if (object_to_attach_to == NULL) {
	    error(_("unexpected attach symbol start marker"));
	    break;
	}

	attribute_list = object_to_attach_to->children();

	if (!attribute_list) {
	    error(_("object type does not support attaching attributes"));
	    attribute_list = drawing->rootContext();
	}
	break;

    case '[': /* START_EMBEDDED */
	if (!complex_to_embed_to) {
	    error(_("unexpected embedded symbol start marker"));
	    break;
	}
	//stack.append(new Drawing...);
	break;

    case ']': /* END_EMBEDDED */
	if (stack.count() < 2) {
	    error(_("unexpected embedded symbol end marker"));
	    break;
	}
	stack.pop_back();
/*
  world_get_complex_bounds(toplevel, o_current, &left, &top, &right, &bottom);
  o_current->w_left = left;
  o_current->w_top = top;
  o_current->w_right = right;
  o_current->w_bottom = bottom;
  o_current->w_bounds_valid = TRUE;
 */
	break;

    case '}':
	error(_("unexpected attach symbol end marker"));
	break;

    case 'F': /* INFO_FONT */
	/* NOP */
	break;

    case '#': /* COMMENT */
	/* do nothing */
	break;

    case 'v': /* VERSION_CHAR */
	itemsread = sscanf(line, "v %u %u\n",
			   &release_ver, &fileformat_ver);

	if (itemsread == 0) {
	    error(_("failed to parse version from buffer"));
	    break;
	}

	/* 20030921 was the last version which did not have a
	 * fileformat version.  The below latter test should not
	 * happen, but it is here just in in case. */
	if (release_ver <= VERSION_20030921 || itemsread == 1)
	    fileformat_ver = 0;

	if (fileformat_ver == 0)
	    warn(_("old format sym/sch file--"
		   "please run g[sym|sch]update on it"));
	break;

    default:
	error(_("object type not recognized (0x%1)")
	      .arg(int_to_str(*line, 16)));
	object_to_attach_to = NULL;
    }

    if (object)
	drawing->rootContext()->addObject(object);

    if (object_to_attach_to != complex_to_embed_to)
	complex_to_embed_to = 0;
}

void SchematicSink::done()
{
    switch (state) {
    case Base:
	/* sup */
	break;
    case Text:
	error(_("unexpected end-of-file after %d lines").arg(
		  int_to_str(text_object->text.count())));
	break;
    case PicturePath:
	break;
    case PictureData:
	break;
    case Path:
	error(_("unexpected end-of-file when reading path"));
	break;
    }

    if (attribute_list)
	error(_("unexpected end-of-file in attribute list"));

    /* Was the very last thing we read a complex and has it not been
     * checked yet?  This would happen if the complex is at the very
     * end of the file and had no attached attributes */
    check_symversion();

    if (release_ver <= VERSION_20020825)
	guess_pin_orientation();
}

void SchematicSink::guess_pin_orientation()
{
}

Shared::Sink *new_schematic_sink(Drawing::FileType fileType,
				 Library *library, Font1 *font,
				 FileItem *fileItem, Messages &messages)
{
    DrawingItem *item = new DrawingItem(fileType, library, fileItem);
    fileItem->assignVirtualItem(item);

    Drawing *drawing = item->drawing();
    drawing->setBaseBase(10);
    drawing->setBaseExponent(3);

    // if (font && fileType == Drawing::LetterFile)
    //     font->addCharacter(drawing);

    return new SchematicSink(drawing, messages);
}
