/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "sch_text.h"
#include "icons.h"
#include "hub.h"
#include "action.h"
#include "painter.h"
#include "shared.h"
#include "dirmodel.h"
#include "view.h"
#include "xml_writer.h"
#include "sch_attribute.h"


class TextObject::TextEmbeddingLayer : public EmbeddingLayer
{
public:
    TextEmbeddingLayer(const Metrum3 &pos, Metrum width, Metrum height,
		       int angle, int alignment, const Embedding &embedding);
    int angle() const {
	return the_angle;
    }
    bool mirror() const {
	return the_mirror;
    }
    Metrum3 transform(const Metrum3 &pos) const {
	Metrum3 res = mirror() ? Metrum3(-pos.x(), pos.y(), pos.z()) : pos;

	switch (angle()) {
	case 1:
	    res = Metrum3(-res.y(), res.x(), res.z());
	    break;
	case 2:
	    res = Metrum3(-res.x(), -res.y(), res.z());
	    break;
	case 3:
	    res = Metrum3(res.y(), -res.x(), res.z());
	    break;
	}
	return this->pos + res;
    }
private:
    Metrum3 pos;
    int the_angle;
    bool the_mirror;
};

TextObject::TextEmbeddingLayer::TextEmbeddingLayer(
    const Metrum3 &pos, Metrum width, Metrum height,
    int angle, int alignment, const Embedding &embedding)
    : pos(pos), the_angle(angle), the_mirror(false)
{
    Metrum x_ofs(width * 0), y_ofs(height * 0);

    if (alignment == 2 || alignment == 5 || alignment == 8) /* top */
	y_ofs = -height;
    if (alignment == 1 || alignment == 4 || alignment == 7) /* middle */
	y_ofs = -height / 2;
    if (alignment == 6 || alignment == 7 || alignment == 8) /* right */
	x_ofs = -width;
    if (alignment == 3 || alignment == 4 || alignment == 5) /* center */
	x_ofs = -width / 2;

    if ((embedding.angle() % 2 == embedding.angle()) ==
	(embedding.angle() % 2 != embedding.mirror() && angle % 2)) {
	/* I'm confused.  I need a truth table:
	 *   0 1 2 3     0 1 2 3
	 * 0 0 0 1 1   0 0 0 1 1
	 * 1 0 1 1 0   1 1 0 0 1
	 * 2 0 0 1 1   2 0 0 1 1
	 * 3 0 1 1 0   3 1 0 0 1 */

	the_angle = (angle + 2) % 4;
	y_ofs += height;
	x_ofs += width;
    }

    if (embedding.mirror()) {
	the_mirror = true;
	x_ofs = width - x_ofs;
    }

    switch (angle) {
    case 0: this->pos.rx() += x_ofs; this->pos.ry() += y_ofs; break;
    case 1: this->pos.rx() -= y_ofs; this->pos.ry() += x_ofs; break;
    case 2: this->pos.rx() -= x_ofs; this->pos.ry() -= y_ofs; break;
    case 3: this->pos.rx() += y_ofs; this->pos.ry() -= x_ofs; break;
    }
}

class TextObject::LetterEmbeddingLayer final : public EmbeddingLayer
{
public:
    LetterEmbeddingLayer(const Metrum3 &pos, int size) : pos(pos), size(size) {
    }
    Metrum transform(Metrum m) const {
	return m * size / 2;
    }
    Metrum3 transform(const Metrum3 &pos) const {
	return this->pos + pos * size / 2;
    }
private:
    Metrum3 const pos;
    int const size;
};

/****************************************************************************/

TextMajorMode::TextMajorMode(Hub &hub) : MajorMode(hub), minorMode(hub, *this)
{
    setCaption(_("Text"));
    setIcon(icons::font);
    setMinorMode(&minorMode);

/** textEdit = new QTextEdit;

    QVBoxLayout *layout = new QVBoxLayout(widget());
    layout->addWidget(textEdit); **/
}

void TextMajorMode::accept(MinorMode *minor_mode)
{
    /** String::fromQString(textEdit->toPlainText()) **/
    List<String> text = String("Hello world").split('\n');
    if (!text.isEmpty() && text.last().isEmpty())
	text.removeLast();

    hub.drawing()->performActions(
	List<Action *>() << new NewObjectAction(
	    new TextObject(minorMode.point(), text, hub.drawing())),
	icons::add, _("New text"));
}

void TextMajorMode::cancel()
{
    setMinorMode(&minorMode, true);
}

void TextMajorMode::paint(Painter &painter)
{
    MajorMode::paint(painter);

    if (hub.pointerActive()) {
	painter.setPen(0x7f, 0x7f, 0x7f);
	//painter.paintText(minorMode.point(), textEdit->toPlainText());

	painter.setBlackPen();
	painter.paintPoint(minorMode.point());
    }
}

List<ActionLine *> TextMajorMode::lines(ActionPoint *handle)
{
    return List<ActionLine *>();
}

/****************************************************************************/

TextObject::TextObject(const Metrum3 &pos, const List<String> &text,
		       Drawing *drawing)
    : Object(drawing), pos(pos), color(9), size(10), visibility(true),
      show_name_value(0), angle(0), alignment(0), text(text),
      width(0), height(0),
      handle(_("position"), this)
{
    update();
}

Metrum3 TextObject::relativePos() const
{
    return pos;
}

String TextObject::name() const
{
    return _("text");
}

Icon *TextObject::icon() const
{
    return icons::font;
}

void TextObject::update()
{
    letters.clear();
    emb_layers.deleteAll();
    emb_layers.clear();

    if (!visibility)
	return;

    List<String> shown_text = text;

    if (!text.isEmpty() && show_name_value != 0) {
	int pos = text[0].indexOf('=');
	if (pos != -1 && show_name_value == 1)
	    shown_text[0] = text[0].mid(pos + 1);
	if (pos != -1 && show_name_value == 2)
	    shown_text[0] = text[0].left(pos);
    }

    height = 26 * size/2 * (text.count() * 2 - 1);
    width = 0;

    Metrum line_feed = 52 * size/2 * (text.count() - 1);

    foreach (String line, shown_text) {
	Metrum char_feed(0);

	foreach (wchar_t ch, line.decode()) {
	    Metrum3 ch_pos(char_feed, line_feed, 0);

	    Drawing *d = shared.font()->character(ch);

	    if (d) {
		letters.append(d);
		emb_layers.append(new LetterEmbeddingLayer(ch_pos, size));
		char_feed += d->letterWidth() * size/2;
	    }
	}

	if (char_feed > width)
	    width = char_feed;
	line_feed -= 52 * size/2;
    }
}

void TextObject::paint(Painter &painter, const Embedding &embedding)
{
    TextEmbeddingLayer el(absolutePos(), width, height,
			  angle, alignment, embedding);
    Embedding outer_emb(embedding, &el);

    for (unsigned int i = 0; i < letters.count(); i++) {
	Embedding emb(outer_emb, emb_layers[i]);

	foreach (Object *ob, letters[i]->objects())
	    ob->paint(painter, emb);
    }
}

void TextObject::paintShadow(const Metrum3 &pos, const ActionPoint *handle,
			     Painter &painter)
{
    Metrum3 hpos(absolutePos());

    if (handle == &this->handle)
	hpos = pos;

    painter.setPen(0x80, 0x80, 0x80);
    //painter.paintText(hpos, text);
}

List<ActionPoint *> TextObject::handles()
{
    return List<ActionPoint *>() << &handle;
}

Metrum3 TextObject::handlePos(const ActionPoint *handle) const
{
    if (handle == &this->handle)
	return absolutePos();

    return Metrum3();
}

List<ActionLine *> TextObject::handleLines(const ActionPoint *handle)
{
    return List<ActionLine *>();
}

void TextObject::moveHandle(const Metrum3 &pos, const ActionPoint *handle)
{
    if (handle == &this->handle)
	this->pos = pos - context()->origin();
}

Metrum3 TextObject::lineBase(const ActionLine *line) const
{
    return Metrum3();
}

Metrum3 TextObject::lineTangent(const ActionLine *line) const
{
    return Metrum3(1, 0, 0);
}

ScreenDist TextObject::proximity(const ScreenPos &point, View *view,
				 const Embedding &embedding) const
{
    ScreenDist least_dist = (point - view->map(
				 embedding.transform(absolutePos()))).length();
    TextEmbeddingLayer el(absolutePos(), width, height,
			  angle, alignment, embedding);
    Embedding outer_emb(embedding, &el);

    for (unsigned int i = 0; i < letters.count(); i++) {
	Embedding emb(outer_emb, emb_layers[i]);

	foreach (Object *ob, letters[i]->objects()) {
	    ScreenDist dist = ob->proximity(point, view, emb);
	    if (dist < least_dist || least_dist < 0)
		least_dist = dist;
	}
    }

    return least_dist;
}

bool TextObject::within(const ScreenRect &rect, View *view,
			bool partially, const Embedding &embedding) const
{
    bool res = rect.contains(view->map(embedding.transform(absolutePos())));
    TextEmbeddingLayer el(absolutePos(), width, height,
			  angle, alignment, embedding);
    Embedding outer_emb(embedding, &el);

    for (unsigned int i = 0; i < letters.count(); i++) {
	Embedding emb(outer_emb, emb_layers[i]);

	foreach (Object *ob, letters[i]->objects())
	    if (res != partially &&
		ob->within(rect, view, partially, emb) == partially)
		res = partially;
    }

    return res;
}

void TextObject::write(XMLWriter *writer)
{
    writer->startElement("text");
    writer->writeAttribute("position", drawing()->format(pos));
    writer->writeAttribute("color", SchematicEnum::color[color]);
    writer->writeAttribute("size", int_to_str(size));
    writer->writeAttribute("visible", SchematicEnum::boolean[visibility]);
    writer->writeAttribute("show",
	SchematicEnum::show_name_value[show_name_value]);
    writer->writeAttribute("angle", int_to_str(angle));
    writer->writeAttribute("alignment", SchematicEnum::alignment[alignment]);

    foreach (String line, text) {
	writer->startElement("line", true);
	writer->writeCharacterData(line);
	writer->endElement();
    }

    writer->endElement();
}

List<Attribute *> TextObject::attributes() const
{
    return List<Attribute *>() << schematicAttributes->color
			       << schematicAttributes->textSize
			       << schematicAttributes->visibility
			       << schematicAttributes->showNameValue
			       << schematicAttributes->angle
			       << schematicAttributes->alignment;
}

Variant TextObject::attribute(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(color);

    if (attribute == schematicAttributes->textSize)
	return Variant::fromInt(size);

    if (attribute == schematicAttributes->visibility)
	return Variant::fromBool(visibility);

    if (attribute == schematicAttributes->showNameValue)
	return Variant::fromInt(show_name_value);

    if (attribute == schematicAttributes->angle)
	return Variant::fromInt(angle);

    if (attribute == schematicAttributes->alignment)
	return Variant::fromInt(alignment);

    return Variant();
}

Variant TextObject::attributeDefault(const Attribute *attribute) const
{
    if (attribute == schematicAttributes->color)
	return Variant::fromInt(9);

    if (attribute == schematicAttributes->textSize)
	return Variant::fromInt(10);

    if (attribute == schematicAttributes->visibility)
	return Variant::fromBool(true);

    if (attribute == schematicAttributes->showNameValue)
	return Variant::fromInt(0);

    if (attribute == schematicAttributes->angle)
	return Variant::fromInt(0);

    if (attribute == schematicAttributes->alignment)
	return Variant::fromInt(0);

    return Variant();
}

void TextObject::setAttribute(const Variant &value,
			      const Attribute *attribute)
{
    if (attribute == schematicAttributes->color)
	color = value.toInt();

    if (attribute == schematicAttributes->textSize) {
	size = value.toInt();
	update();
    }

    if (attribute == schematicAttributes->visibility) {
	visibility = value.toBool();
	update();
    }

    if (attribute == schematicAttributes->showNameValue) {
	show_name_value = value.toInt();
	update();
    }

    if (attribute == schematicAttributes->angle) {
	angle = value.toInt();
	update();
    }

    if (attribute == schematicAttributes->alignment) {
	alignment = value.toInt();
	update();
    }
}
