/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "shared.h"
#include "util/path.h"
#include "dirmodel.h"
#include "filemonitor.h"
#include "icons.h"
#include "hub.h"

Shared shared;


class SharedPrivate
{
    StandardItem false_item, true_item;
public:
    SharedPrivate();
    ~SharedPrivate();
    List<Hub *> hubs;

    int initialization_state;
    RootInterior project_root, library_root;

    List<Library *> libraries;
    List<String> library_paths;
    List<Font1 *> fonts;
    List<String> font_paths;

    std::map<String, DirectoryItem *> directory_items;
    std::map<String, FileItem *> file_items;

    FileMonitor file_monitor;

    Library *library_for_path(const String &path) const;
    Font1 *font_for_path(const String &path) const;

    StandardInterior boolean_enum;
};

SharedPrivate::SharedPrivate()
    : false_item(_("false"), icons::tree_leaf),
      true_item(_("true"), icons::tree_leaf), initialization_state(0),
      boolean_enum(List<Item *>() << &false_item << &true_item)
{
}

SharedPrivate::~SharedPrivate()
{
    libraries.deleteAll();
    fonts.deleteAll();

    for (std::map<String, DirectoryItem *>::const_iterator i =
	     directory_items.begin(); i != directory_items.end(); i++)
	delete (*i).second;
    for (std::map<String, FileItem *>::const_iterator i =
	     file_items.begin(); i != file_items.end(); i++)
	delete (*i).second;
}

Library *SharedPrivate::library_for_path(const String &path) const
{
    Library *library = 0;
    String library_path;

    for (unsigned int i = 0; i < libraries.count(); i++)
	if (path_contains(library_paths[i], path) &&
	    library_paths[i].length() > library_path.length()) {
	    library = libraries[i];
	    library_path = library_paths[i];
	}

    return library;
}

Font1 *SharedPrivate::font_for_path(const String &path) const
{
    Font1 *font = 0;
    String font_path;

    for (unsigned int i = 0; i < fonts.count(); i++)
	if (path_contains(font_paths[i], path) &&
	    font_paths[i].length() > font_path.length()) {
	    font = fonts[i];
	    font_path = font_paths[i];
	}

    return font;
}

/****************************************************************************/

Shared::Shared() : p(0)
{
}

void Shared::initialize()
{
    assert(!p);
    p = new SharedPrivate();
}

Shared::~Shared()
{
    assert(!p);
}

void Shared::requestQuit()
{
    foreach (Hub *hub, p->hubs)
	hub->requestClose();
}

void Shared::registerHub(Hub *hub)
{
    if (!p->hubs.contains(hub))
	p->hubs.append(hub);
}

void Shared::unregisterHub(Hub *hub)
{
    p->hubs.removeAll(hub);

    if (p->hubs.empty()) {
	delete p;
	p = 0;
    }
}

void Shared::drawingStatusChanged(Drawing *drawing)
{
    foreach (Hub *hub, p->hubs)
	if (hub->drawing() == drawing)
	    hub->drawingStatusChanged();
}

ItemInterior *Shared::projectDirs() const
{
    return &p->project_root;
}

ItemInterior *Shared::library() const
{
    return &p->library_root;
}

bool Shared::moreInitialization(String &message)
{
    if (p->initialization_state == -1)
	return false;

    switch (p->initialization_state++) {
    case 0:
	message = _("Reading project directories...");
	return true;
    case 1: {
	List<String> project_dirs; // = settings.value("projectDirectories");

	if (project_dirs.empty())
	    project_dirs << home_dir() + "/.xi/projects";

	p->project_root.replacePaths(0, p->project_root.count(), project_dirs);

	foreach (Hub *hub, p->hubs)
	    hub->updateProjectDirectories();

	message = _("Reading library...");
	return true;
    }
    case 2: {
	List<String> library_dirs; // = settings.value("libraryDirectories");

	if (library_dirs.empty())
	    library_dirs << home_dir() + "/.xi/library";

	foreach (String dir, library_dirs) {
	    p->libraries.append(new Library);
	    p->library_paths.append(dir);
	}

	p->library_root.replacePaths(0, p->library_root.count(), library_dirs);
	/* fallthrough */
    }
    default:
	p->initialization_state = -1;
	return false;
    }
}

Drawing *Shared::component(const String &name) const
{
    foreach (Library *library, p->libraries) {
	Drawing *component = library->component(name);
	if (component)
	    return component;
    }

    return 0;
}

Font1 *Shared::font()
{
    return fileItem(sanitize_path(home_dir() + "/.xi/font.xml"))
	->thisFont();
}

FileItem *Shared::fileItem(const String &path)
{
    assert(path_is_sane(path));

    FileItem *&item = p->file_items[path];

    if (!item)
	item = new FileItem(
	    path, p->library_for_path(path), p->font_for_path(path));

    return item;
}

DirectoryItem *Shared::directoryItem(const String &path)
{
    assert(path_is_sane(path));

    DirectoryItem *&item = p->directory_items[path];

    if (!item)
	item = new DirectoryItem(
	    path, p->library_for_path(path) || p->font_for_path(path));

    return item;
}

FileMonitor *Shared::fileMonitor() const
{
    return &p->file_monitor;
}

ItemInterior &Shared::booleanEnum() const
{
    return p->boolean_enum;
}

Shared::Sink::~Sink()
{
    /* virtualize */
}

void Shared::readFile(const String &path, Sink *sink, Messages &messages)
{
    if (!sink) {
	assert(!messages.empty());
	return;
    }

    int fd = open(path.data(), O_RDONLY);

    if (fd == -1)
	messages.error("open: " + String(strerror(errno)));
    else {
	char buf[BUFSIZ];
	ssize_t count;

	do {
	    count = read(fd, buf, BUFSIZ);

	    if (count == -1) {
		messages.error("read: " + String(strerror(errno)));
		break;
	    }
	} while (sink->read(buf, count) && count != 0);

	if (close(fd) == -1)
	    messages.error("close: " + String(strerror(errno)));
    }

    delete sink;
}
