/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdio.h>
#include <errno.h>
#include <err.h>
#define __STDC_LIMIT_MACROS
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <limits.h>
#include <wctype.h>
#include "util/string.h"
#include "../config.h"
#include "../gettext.h"

#define BUF_SIZE 256

static char buf[BUF_SIZE];


void String::dump() const
{
    fprintf(stderr, "-> '%s'\n", data());
}

String String::encode(const std::wstring &str)
{
    size_t len = str.length() * 16 + 1;
    char buf[len];
    const wchar_t *src = str.data();

    if ((len = wcsrtombs(buf, &src, len, NULL)) == (size_t)-1)
	err(EXIT_FAILURE, "wcsrtombs");

    return String(buf, len);
}

String String::encodeUTF8(const std::wstring &str)
{
    return encode(str);
}

std::wstring String::decode() const
{
    size_t len = length() * 4 + 1;
    wchar_t buf[len];
    const char *src = data();

    if ((len = mbsrtowcs(buf, &src, len, NULL)) == (size_t)-1)
	err(EXIT_FAILURE, "mbsrtowcs");

    return std::wstring(buf, len);
}

std::wstring String::decodeUTF8() const
{
    return decode();
}

String String::arg(const String &a1) const
{
    size_type pos;
    String s = *this;

    pos = s.find("%1");
    s.replace(pos, 2, a1);

    return s;
}

String String::arg(const String &a1, const String &a2) const
{
    size_type pos;
    String s = *this;

    pos = s.find("%1");
    s.replace(pos, 2, a1);
    pos = s.find("%2", pos + a1.length());
    s.replace(pos, 2, a2);

    return s;
}

String String::capitalize() const /* TODO */
{
    if (empty())
	return String();

    std::wstring str = decode();
    str[0] = towupper(str[0]);
    return encode(str);
}

int String::count(char c) const
{
    size_type pos = 0, result = 0;
    while ((pos = find(c, pos)) != npos) {
	result++;
	pos++;
    }
    return result;
}

int String::count(const String &str) const
{
    size_type pos = 0, result = 0;
    while ((pos = find(str, pos)) != npos) {
	result++;
	pos += str.length();
    }
    return result;
}

String &String::remove(char c)
{
    size_type pos = 0;
    while ((pos = find(c, pos)) != npos)
	erase(pos, 1);
    return *this;
}

String &String::remove(const String &str)
{
    size_type pos = 0;
    while ((pos = find(str, pos)) != npos)
	erase(pos, str.length());
    return *this;
}

List<String> String::split(char sep) const
{
    size_type a = 0, b;
    List<String> l;

    while ((b = find(sep, a)) != npos) {
	l.append(substr(a, b - a));
	a = b + 1;
    }

    l.append(substr(a));
    return l;
}

List<String> String::split(const String &sep) const
{
    size_type a = 0, b;
    List<String> l;

    while ((b = find(sep, a)) != npos) {
	l.append(substr(a, b - a));
	a = b + sep.length();
    }

    l.append(substr(a));
    return l;
}

String String::trimmed() const /* TODO */
{
    std::wstring str = decode();
    size_type a = 0, b = str.length();

    while (b > 0 && iswspace(str[b - 1]))
	b--;
    while (a < b && iswspace(str[a]))
	a++;

    return encode(str.substr(a, b - a));
}

/****************************************************************************/

String join_strings(const List<String> &list, char sep)
{
    if (list.empty())
	return String();

    size_t len = 0;
    String s;
    bool append_sep = false;

    foreach (String str, list)
	len += str.length();
    s.reserve(len + list.count() - 1);

    foreach (String str, list) {
	if (append_sep)
	    s.push_back(sep);
	else
	    append_sep = true;
	s.append(str);
    }

    return s;
}

String join_strings(const List<String> &list, const String &sep)
{
    if (list.empty())
	return String();

    size_t len = 0;
    String s;
    bool append_sep = false;

    foreach (String str, list)
	len += str.length();
    s.reserve(len + (list.count() - 1) * sep.length());

    foreach (String str, list) {
	if (append_sep)
	    s.append(sep);
	else
	    append_sep = true;
	s.append(str);
    }

    return s;
}

int str_to_int(const String &str, bool &ok, int base)
{
    const char *nptr = str.data();
    char *endptr;
    long int val;

    errno = 0;
    val = strtol(nptr, &endptr, base);

    if (errno != 0) {
	warn("strtol: %s", nptr);
	ok = false;
	return 0;
    }

    if (*nptr == '\0' || *endptr != '\0') {
	warnx("str_to_int: No value given");
	ok = false;
	return 0;
    }

    if (val < INT_MIN || val > INT_MAX) {
	warnx("str_to_int: %s: Value out of range for int", nptr);
	ok = false;
	return 0;
    }

    ok = true;
    return val;
}

int64_t str_to_int64(const String &str, bool &ok, int base)
{
    const char *nptr = str.data();
    char *endptr;
    intmax_t val;

    errno = 0;
    val = strtoimax(nptr, &endptr, base);

    if (errno != 0) {
	warn("strtoimax: %s", nptr);
	ok = false;
	return 0;
    }

    if (*nptr == '\0' || *endptr != '\0') {
	warnx("str_to_int64: No value given");
	ok = false;
	return 0;
    }

    if (val < INT64_MIN || val > INT64_MAX) {
	warnx("str_to_int64: %s: Value out of range for int64_t", nptr);
	ok = false;
	return 0;
    }

    ok = true;
    return val;
}

float str_to_float(const String &str, bool &ok)
{
    const char *nptr = str.data();
    char *endptr;
    float val;

    errno = 0;
    val = strtof(nptr, &endptr);

    if (errno != 0) {
	warn("strtof: %s", nptr);
	ok = false;
	return 0;
    }

    if (*nptr == '\0' || *endptr != '\0') {
	warnx("str_to_float: No value given");
	ok = false;
	return 0;
    }

    ok = true;
    return val;
}

double str_to_double(const String &str, bool &ok)
{
    const char *nptr = str.data();
    char *endptr;
    double val;

    errno = 0;
    val = strtod(nptr, &endptr);

    if (errno != 0) {
	warn("strtod: %s", nptr);
	ok = false;
	return 0;
    }

    if (*nptr == '\0' || *endptr != '\0') {
	warnx("str_to_double: No value given");
	ok = false;
	return 0;
    }

    ok = true;
    return val;
}

String int_to_str(int val, int base)
{
    int len;

    if (base == 10)
	len = snprintf(buf, BUF_SIZE, "%d", val);
    else if (base == 16)
	len = snprintf(buf, BUF_SIZE, "%x", val);
    else
	errx(EXIT_FAILURE, "int_to_str: %d: Invalid base", base);

    if (len >= BUF_SIZE)
	errx(EXIT_FAILURE, "snprintf: Buffer too small");

    return String(buf, len);
}

String int64_to_str(int64_t val, int base)
{
    int len;

    if (base == 10)
	len = snprintf(buf, BUF_SIZE, "%" PRId64, val);
    else if (base == 16)
	len = snprintf(buf, BUF_SIZE, "%" PRIx64, val);
    else
	errx(EXIT_FAILURE, "int64_to_str: %d: Invalid base", base);

    if (len >= BUF_SIZE)
	errx(EXIT_FAILURE, "snprintf: Buffer too small");

    return String(buf, len);
}

String float_to_str(float val)
{
    int len = snprintf(buf, BUF_SIZE, "%f", val);

    if (len >= BUF_SIZE)
	errx(EXIT_FAILURE, "snprintf: Buffer too small");

    return String(buf, len);
}

String double_to_str(double val)
{
    int len = snprintf(buf, BUF_SIZE, "%f", val);

    if (len >= BUF_SIZE)
	errx(EXIT_FAILURE, "snprintf: Buffer too small");

    return String(buf, len);
}

String _(const char *msgid)
{
    return String(gettext(msgid));
}
