/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <string.h>
#include <expat.h>
#include "xml.h"
#include "util/metrum3.h"
#include "messages.h"
#include "drawing.h"


class XMLSink::Internal
{
    XMLHandler *const root_handler;
    List<XMLElement *> stack;

    static void start(void *userData, const XML_Char *name,
				      const XML_Char **atts) {
	List<String> att_names, att_values;
	for (unsigned int i = 0; atts[i] != 0; ) {
	    att_names.append(atts[i++]);
	    att_values.append(atts[i++]);
	}

	Internal *i = static_cast<Internal *>(userData);
	XMLElement *element = new XMLElement(
	    name, att_names, att_values, i->messages);

	if (!i->stack.isEmpty()) {
	    element->setNamespaceURI(i->stack.last()->namespaceURI());
	    i->stack.last()->handler()->element(element, i->messages);
	} else
	    i->root_handler->element(element, i->messages);
	i->stack.append(element);
    }
    static void end(void *userData, const XML_Char */*name*/) {
	Internal *i = static_cast<Internal *>(userData);

	i->stack.last()->handler()->done(i->messages);
	delete i->stack.last();
	i->stack.removeLast();
    }
    static void charhndl(void *userData, const XML_Char *s, int len) {
	Internal *i = static_cast<Internal *>(userData);
	i->stack.last()->handler()->characterData(String(s, len), i->messages);
    }
public:
    XML_Parser const p;
    Messages &messages;

    Internal(XMLHandler *rootHandler, Messages &messages)
	: root_handler(rootHandler),
	  p(XML_ParserCreateNS(0, NSSEP)), messages(messages) {
	XML_SetUserData(p, this);
	XML_SetElementHandler(p, start, end);
	XML_SetCharacterDataHandler(p, charhndl);
    }
    ~Internal() {
	XML_ParserFree(p);
    }
};


XMLSink::XMLSink(XMLHandler *rootHandler, Messages &messages)
    : i(new Internal(rootHandler, messages))
{
}

XMLSink::~XMLSink()
{
    delete i;
}

bool XMLSink::read(const void *buf, ssize_t count)
{
    void *xbuf = XML_GetBuffer(i->p, count);

    if (count != 0) {
	if (xbuf == 0) {
	    i->messages.error(_("cannot allocate parser buffer"));
	    return false;
	}

	memcpy(xbuf, buf, count);
    }

    if (XML_ParseBuffer(i->p, count, count == 0) == 0) {
	i->messages.append(Message(Message::Error,
				   XML_ErrorString(XML_GetErrorCode(i->p)),
				   XML_GetCurrentLineNumber(i->p) - 1,
				   XML_GetCurrentColumnNumber(i->p)));
	return false;
    }

    return true;
}

/****************************************************************************/

void XMLHandler::element(XMLElement *element, Messages &messages)
{
    messages.error(_("Unexpected element: %1").arg(element->name()));
    element->ignore();
}

void XMLHandler::characterData(const String &data, Messages &messages)
{
    if (!data.trimmed().isEmpty())
	messages.error(_("Unexpected character data: %1").arg(data));
}

void XMLHandler::done(Messages &/*messages*/)
{
}

/****************************************************************************/

class NullXMLHandler : public XMLHandler
{
public:
    void element(XMLElement *element, Messages &messages);
    void characterData(const String &data, Messages &messages);
};

void NullXMLHandler::element(XMLElement *element, Messages &/*messages*/)
{
    element->ignore();
}

void NullXMLHandler::characterData(const String &/*data*/,
				   Messages &/*messages*/)
{
}

/****************************************************************************/

XMLElement::XMLElement(const String &name,
		       const List<String> &att_names,
		       const List<String> &att_values,
		       Messages &messages)
    : the_name(name), att_names(att_names), att_values(att_values),
      the_handler(new XMLHandler()), messages(messages)
{
}

XMLElement::~XMLElement()
{
    foreach (String att_name, att_names)
	messages.error(_("Unknown attribute: %1").arg(att_name));
}

String XMLElement::namespaceURI() const
{
    return uri;
}

void XMLElement::setNamespaceURI(const String &uri)
{
    this->uri = uri;
}

String XMLElement::name() const
{
    if (!uri.isEmpty() && the_name.startsWith(uri + NSSEP))
	return the_name.mid(uri.length() + 1);

    return the_name;
}

bool XMLElement::selectAttribute(const String &name)
{
    int i = att_names.indexOf(name);

    if (i == -1)
	return false;

    att_names.removeAt(i);
    att_value = att_values.takeAt(i);
    return true;
}

void XMLElement::ignore()
{
    att_names.clear();
    att_values.clear();
    setHandler(new NullXMLHandler());
}

String XMLElement::value() const
{
    return att_value;
}

int XMLElement::toInt(int base)
{
    bool ok;
    int i = str_to_int(att_value, ok, base);
    if (!ok)
	messages.error(_("Not an integer value: %1").arg(att_value));
    return i;
}

int64_t XMLElement::toInt64(int base)
{
    bool ok;
    int64_t l = str_to_int64(att_value, ok, base);
    if (!ok)
	messages.error(_("Not an integer value: %1").arg(att_value));
    return l;
}

float XMLElement::toFloat()
{
    bool ok;
    float f = str_to_float(att_value, ok);
    if (!ok)
	messages.error(_("Not a float value: %1").arg(att_value));
    return f;
}

double XMLElement::toDouble()
{
    bool ok;
    double d = str_to_double(att_value, ok);
    if (!ok)
	messages.error(_("Not a float value: %1").arg(att_value));
    return d;
}

Metrum XMLElement::toMetrum(Drawing *drawing)
{
    bool ok;
    Metrum m = drawing->parseMetrum(att_value, ok);
    if (!ok)
	messages.error(_("Not a metric value: %1").arg(att_value));
    return m;
}

Metrum3 XMLElement::toMetrum3(Drawing *drawing)
{
    bool ok;
    Metrum3 m = drawing->parseMetrum3(att_value, ok);
    if (!ok)
	messages.error(_("Not a 3D metric value: %1").arg(att_value));
    return m;
}

int XMLElement::lookup(const List<String> &values)
{
    int i;

    if ((i = values.indexOf(att_value)) != -1)
	return i;

    messages.error(_("Unknown value: %1").arg(att_value));
    return 0;
}

XMLHandler *XMLElement::handler() const
{
    return the_handler;
}

void XMLElement::setHandler(XMLHandler *handler)
{
    the_handler = handler;
}
