/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <algorithm>
#include "action.h"
#include "util/encoding.h"

using namespace Toolkit;


/*! \class Toolkit::ActionSet
 *  \brief Do the actual work when an Action has been triggered.
 *
 *  An action set is a callback object which handles the triggering of
 *  one or more actions.  When one of the associated actions has been
 *  triggered, it calls ActionSet::actionTriggered on the action set
 *  which should do whatever is expected as a consequence.
 *
 *  In addition, the action set keeps a list of associated actions so
 *  they can be registered with a single call to
 *  InputWindowHook::registerActionSet.
 *
 *  A typical action set might look like this:
 *
 *  \code{.cpp}
 *  class ClipboardActionSet : public Toolkit::ActionSet
 *  {
 *  public:
 *      ClipboardActionSet();
 *      void actionTriggered(Toolkit::Action *action);
 *
 *      Toolkit::Action cutAction;
 *      Toolkit::Action copyAction;
 *      Toolkit::Action pasteAction;
 *  };
 *
 *  ClipboardActionSet::ClipboardActionSet() : cutAction(*this),
 *                                             copyAction(*this),
 *                                             pasteAction(*this)
 *  {
 *      cutAction.setText(_("&Cut"));
 *      cutAction.setHotkey(_("C-x"));
 *      cutAction.setIcon(icons::cut);
 *
 *      copyAction...
 *  }
 *
 *  void ClipboardActionSet::actionTriggered(Toolkit::Action *action)
 *  {
 *      if (action == &cutAction) {
 *          ...
 *          return;
 *      }
 *      ...
 *  }
 *  \endcode  */

/*! \brief Destruct an action set.
 *
 *  \note Make sure any associated Action%s are deleted first.  */

ActionSet::~ActionSet()
{
    assert(_actions.empty());
}

/*! \brief Called whenever an action associated with this action set
 *         is triggered.
 *
 *  \param[in] action The action which has been triggered.  */

void ActionSet::actionTriggered(Action *action)
{
    /* relax */
}

/*! \brief Return the actions associated with the action set.  */

std::vector<Action *> ActionSet::actions() const
{
    return _actions;
}

/****************************************************************************/

/*! \class Toolkit::Action::Observer
 *  \brief Observe appearance and status changes of an Action.
 *
 *  Whenever the code providing an action decides to change the
 *  action's state, all objects representing that action are notified
 *  about the change so they can update their appearance accordingly.
 *  In order to be informed about these changes, create an instance of
 *  Observer which registers with the action and will receive calls to
 *  Observer::actionChanged every time the action changes for the
 *  duration of the observer's lifespan.
 *
 *  You will typically want to inherit Observer in addition to your
 *  usual inheritance, like this:
 *
 *  \code{.cpp}
 *  class MyWidget : public Toolkit::Widget, public Toolkit::Action::Observer
 *  {
 *  public:
 *      MyWidget(Toolkit::Connection &conn, Toolkit::Action &action);
 *      void actionChanged();
 *      ...
 *  };
 *  \endcode
 *
 *  \note If you instantiate an observer, make sure it is deleted
 *        before deleting the observed action.  It is an error for an
 *        observer to exist after the action has been deleted.  */

/*! \brief Construct an Observer and register it with the observed Action.
 *
 *  \param[in] action The action which should be observed.  */

Action::Observer::Observer(Action &action) : action(action)
{
    assert(std::find(action.observers.begin(),
		     action.observers.end(), this) == action.observers.end());
    action.observers.push_back(this);
}

/*! \brief Unregister and destruct an Observer.  */

Action::Observer::~Observer()
{
    std::vector<Observer *>::iterator p = std::find(
	action.observers.begin(), action.observers.end(), this);
    assert(p != action.observers.end());
    action.observers.erase(p);
}

/*! \fn Toolkit::Action::Observer::actionChanged
 *  \brief Called every time the observed Action's properties change.
 *
 *  The observed properties include:
 *
 *  * Action::text()
 *  * Action::mnemonic()
 *  * Action::hotkey()
 *  * Action::icon()
 *  * Action::menu()
 *  * Action::isActive()
 *  * Action::disableMessage()  */

/****************************************************************************/

/*! \class Toolkit::Action
 *  \brief Control appearance and state of an action the user can
 *         perform.
 *
 *  An Action represents something the user can select from a menu bar
 *  or toolbar.  It provides a single place to set the properties of
 *  the action---appearance, whether the action is disabled, for what
 *  reason, and whether it is currently activated---and updates all
 *  associated menu and toolbar widgets accordingly.
 *
 *  Whenever one of the widgets representing an action is activated,
 *  it calls Action::trigger on the action.  Creating a subclass of
 *  Action for each action just to override that method would be
 *  tedious, so one object inheriting from ActionSet can handle the
 *  activation of multiple actions.  What actions you organize in an
 *  action set is purely a design decision---you may put all actions
 *  into a single action set if you want to.
 *
 *  An action needs to be registered with the InputWindowHook so the
 *  user is able to trigger it using the hotkey.  To make this easier,
 *  the action set keeps a list of associated actions which are
 *  registered together via InputWindowHook::registerActionSet.
 *
 *  See the description of ActionSet for a code example.  */

/*! \brief Construct an action.
 *
 *  \param[in] actionSet ActionSet handling this action.  Any calls to
 *                       Action::trigger will be relayed to this
 *                       action set.  */

Action::Action(ActionSet &actionSet)
    : action_set(actionSet), _mnemonic(-1), _icon(0), _menu(0),
      is_active(false), is_disabled(false)
{
    assert(std::find(action_set._actions.begin(),
		     action_set._actions.end(), this)
	   == action_set._actions.end());
    action_set._actions.push_back(this);
}

/*! \brief Destruct an action.
 *
 *  \note Make sure any Observer%s of the action are deleted
 *        first.  */

Action::~Action()
{
    assert(observers.empty());

    std::vector<Action *>::iterator p = std::find(
	action_set._actions.begin(), action_set._actions.end(), this);
    assert(p != action_set._actions.end());
    action_set._actions.erase(p);
}

void Action::update_observers()
{
    std::vector<Observer *> cached_observers = observers;

    for (std::vector<Observer *>::const_iterator i = cached_observers.begin();
	 i != cached_observers.end(); i++)
	(*i)->actionChanged();
}

/*! \brief Return the action's caption.
 *
 *  Any ampersands interpreted by Action::setText will have been
 *  stripped.  If you want to assign the caption to an action,
 *  retrieve both caption and mnemonic and set them using
 *  Action::setLiteralText.  */

std::string Action::text() const
{
    return _text;
}

/*! \brief Return the action's mnemonic.
 *
 *  The mnemonic letter is the the one which activates the action in
 *  an open menu.  It may be shown underlined.
 *
 *  \return The index of the mnemonic letter in the caption, starting
 *          with \c 0, or \c -1 if no mnemonic letter is defined.  */

int Action::mnemonic() const
{
    return _mnemonic;
}

/*! \brief Set the action's caption and mnemonic.
 *
 *  A character preceded with an '&' is considered the mnemonic letter
 *  and may be underlined by the displaying widget.  There can only be
 *  one mnemonic letter at a time.  To include a literal ampersand in
 *  the caption, escape it with another ampersand ("&&").
 *
 *  If you want to set a string as the caption without ampersands
 *  being treated specially, use Action::setLiteralText instead.  */

void Action::setText(const std::string &text)
{
    std::wstring w = assert_decode(text);
    int m = -1;

    for (std::wstring::size_type p = w.find(L'&'); p != w.npos;
	 p = w.find(L'&', p + 1)) {
	w.erase(p, 1);
	assert(p < w.length());
	if (w[p] != L'&') {
	    assert(m == -1);
	    m = p;
	}
    }

    std::string t = assert_encode(w);
    if (t == _text && m == _mnemonic)
	return;

    _text = t;
    _mnemonic = m;
    update_observers();
}

/*! \brief Set the action's caption and mnemonic ignoring any special
 *         characters.
 *
 *  Use this function if you don't want ampersands to be treated
 *  specially---for example, if you are using a file name as the
 *  caption.
 *
 *  \param[in] text The caption.
 *
 *  \param[in] mnemonic The index of the mnemonic letter, starting
 *                      with \c 0.  If omitted or \c -1, no mnemonic
 *                      will be used.  */

void Action::setLiteralText(const std::string &text, int mnemonic)
{
    std::wstring w = assert_decode(text);
    assert(mnemonic >= -1);
    assert(mnemonic < (int)w.size());

    if (text == _text && mnemonic == _mnemonic)
	return;

    _text = text;
    _mnemonic = mnemonic;
    update_observers();
}

/*! \brief Return the action's hotkey.  */

Hotkey Action::hotkey() const
{
    return _hotkey;
}

/*! \brief Set the action's hotkey.  */

void Action::setHotkey(const Hotkey &hotkey)
{
    if (hotkey == _hotkey)
	return;

    _hotkey = hotkey;
    update_observers();
}

/*! \brief Return the action's icon.  */

Icon *Action::icon() const
{
    return _icon;
}

/*! \brief Set the action's icon.
 *
 *  In a MenuItem, the icon will be shown along with the text.  In a
 *  ToolbarButton, just the icon will be shown.  */

void Action::setIcon(Icon *icon)
{
    if (icon == _icon)
	return;

    _icon = icon;
    update_observers();
}

/*! \brief Return the menu invoked by the action.  */

Menu *Action::menu() const
{
    return _menu;
}

/*! \brief Set the menu invoked by the action.
 *
 *  The menu will be shown when the pointer hovers over the MenuItem
 *  or the user clicks on the ToolbarButton.  */

void Action::setMenu(Menu *menu)
{
    if (menu == _menu)
	return;

    _menu = menu;
    update_observers();
}

/*! \brief Return whether the action is currently active.  */

bool Action::isActive() const
{
    return is_active;
}

/*! \brief Set whether the action is currently active.
 *
 *  A ToolbarButton will have a pressed-down appearance while its
 *  action is active.  */

void Action::setActive(bool active)
{
    if (active == is_active)
	return;

    is_active = active;
    update_observers();
}

/*! \brief Return whether the action is disabled.  */

bool Action::isDisabled() const
{
    return is_disabled;
}

/*! \brief Return an explanation message if the action is disabled, or
 *         an empty string otherwise.  */

std::string Action::disableMessage() const
{
    return disable_message;
}

/*! \brief Enable the action.  */

void Action::enable()
{
    if (!is_disabled)
	return;

    is_disabled = false;
    disable_message.clear();
    update_observers();
}

/*! \brief Disable the action and set a message explaining why it is
 *         disabled.  */

void Action::disable(const std::string &message)
{
    if (is_disabled && message == disable_message)
	return;

    is_disabled = true;
    disable_message = message;
    update_observers();
}

/*! \brief Trigger the action.
 *
 *  This call will be relayed to the ActionSet::actionTriggered method
 *  of the action set provided at construction time.  */

void Action::trigger()
{
    action_set.actionTriggered(this);
}
