/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <algorithm>
#include "container1.h"

using namespace Toolkit;


Container::Container(Connection &conn) : ContainerWidget(conn),
					 _spacing(0), _padding(0),
					 _expand(true), _background(-1)
{
}

Container::~Container()
{
    for (std::vector<Control *>::const_iterator i = ordered_controls.begin();
	 i != ordered_controls.end(); i++)
	delete *i;
}

std::vector<ContainerWidget::Control *> Container::controls() const
{
    return ordered_controls;
}

int Container::spacing() const
{
    return _spacing;
}

void Container::setSpacing(int spacing)
{
    if (spacing == _spacing)
	return;

    _spacing = spacing;
    updateSizeRequirements();
    requestLayout();
}

int Container::padding() const
{
    return _padding;
}

void Container::setPadding(int padding)
{
    if (padding == _padding)
	return;

    _padding = padding;
    updateSizeRequirements();
    requestLayout();
}

bool Container::expand() const
{
    return _expand;
}

void Container::setExpand(bool expand)
{
    if (expand == _expand)
	return;

    _expand = expand;
    requestLayout();
}

unsigned long Container::background() const
{
    return _background;
}

void Container::setBackground(unsigned long background)
{
    _background = background;
}

void Container::addWidget(Widget *widget)
{
    Control *control = new Control(*this);
    ordered_controls.push_back(control);
    control->setWidget(widget);
    updateSizeRequirements();
    requestLayout();
}

void Container::insertWidget(unsigned int position, Widget *widget)
{
    Control *control = new Control(*this);
    ordered_controls.insert(ordered_controls.begin() + position, control);
    control->setWidget(widget);
    updateSizeRequirements();
    requestLayout();
}

void Container::removeWidget(Widget *widget)
{
    WidgetControl *control = widget->control();
    std::vector<Control *>::iterator i =
	std::find(ordered_controls.begin(), ordered_controls.end(), control);

    assert(control);
    assert(i != ordered_controls.end());

    ordered_controls.erase(i);
    delete control;
    updateSizeRequirements();
    requestLayout();
}

bool Container::widgetEnabled(Widget *widget) const
{
    Control *control = static_cast<Control *>(widget->control());
    assert(std::find(ordered_controls.begin(),
		     ordered_controls.end(), control)
	       != ordered_controls.end());
    return control->isEnabled();
}

void Container::setWidgetEnabled(bool enabled, Widget *widget)
{
    Control *control = static_cast<Control *>(widget->control());
    assert(std::find(ordered_controls.begin(),
		     ordered_controls.end(), control)
	       != ordered_controls.end());
    control->setEnabled(enabled);
    updateSizeRequirements();
    requestLayout();
}

void Container::sizeRequirementsChanged(WidgetControl *control)
{
    updateSizeRequirements();
    requestLayout();
}

void Container::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    if (background() != (unsigned long)-1) {
	gc().setForeground(background());
	pixmap.fillRectangle(gc(), x, y, width(), height());
    }

    ContainerWidget::paint(region, pixmap);
}

/****************************************************************************/

VContainer::VContainer(Connection &conn) : Container(conn),
					   min_width(0), min_height(0),
					   max_width(0), expanding(0)
{
}

void VContainer::updateSizeRequirements()
{
    std::vector<Control *> controls = this->controls();
    min_width = 0;
    min_height = padding() * 2 - spacing();
    max_width = -1;
    expanding = 0;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	if (!(*i)->isEnabled())
	    continue;
	SizeRequirements reqs = (*i)->sizeRequirements();

	if (reqs.minWidth > min_width)
	    min_width = reqs.minWidth;
	min_height += reqs.minHeight + spacing();
	if (reqs.maxWidth != -1 && (
		max_width == -1 || reqs.maxWidth < max_width))
	    max_width = reqs.maxWidth;
	if (reqs.maxHeight != reqs.minHeight)
	    expanding++;
    }

    min_width += padding() * 2;
    if (max_width != -1)
	max_width += padding() * 2;

    SizeRequirements reqs;
    reqs.minWidth = min_width;
    reqs.minHeight = min_height;
    reqs.maxWidth = max_width;
    if (!expanding)
	reqs.maxHeight = min_height;
    setSizeRequirements(reqs);
}

void VContainer::updateLayout()
{
    std::vector<Control *> controls = this->controls();
    int used = 0;
    int available = height() - min_height;
    int interested = expanding;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	if (!(*i)->isEnabled())
	    continue;
	SizeRequirements reqs = (*i)->sizeRequirements();
	Region dimensions(x() + padding(),
			  y() + used + padding(),
			  width() - padding() * 2,
			  reqs.minHeight);

	if (reqs.maxHeight != reqs.minHeight && expand()) {
	    int extra = (available - 1) / interested + 1;
	    dimensions.height += extra;
	    available -= extra;
	    interested--;
	}

	(*i)->setDimensions(dimensions);
	used += dimensions.height + spacing();
    }
}

/****************************************************************************/

HContainer::HContainer(Connection &conn) : Container(conn),
					   min_width(0), min_height(0),
					   max_height(0), expanding(0)
{
}

void HContainer::updateSizeRequirements()
{
    std::vector<Control *> controls = this->controls();
    min_width = padding() * 2 - spacing();
    min_height = 0;
    max_height = -1;
    expanding = 0;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	if (!(*i)->isEnabled())
	    continue;
	SizeRequirements reqs = (*i)->sizeRequirements();

	min_width += reqs.minWidth + spacing();
	if (reqs.minHeight > min_height)
	    min_height = reqs.minHeight;
	if (reqs.maxWidth != reqs.minWidth)
	    expanding++;
	if (reqs.maxHeight != -1 && (
		max_height == -1 || reqs.maxHeight < max_height))
	    max_height = reqs.maxHeight;
    }

    min_height += padding() * 2;
    if (max_height != -1)
	max_height += padding() * 2;

    SizeRequirements reqs;
    reqs.minWidth = min_width;
    reqs.minHeight = min_height;
    if (!expanding)
	reqs.maxWidth = min_width;
    reqs.maxHeight = max_height;
    setSizeRequirements(reqs);
}

void HContainer::updateLayout()
{
    std::vector<Control *> controls = this->controls();
    int used = 0;
    int available = width() - min_width;
    int interested = expanding;

    for (std::vector<Control *>::const_iterator i = controls.begin();
	 i != controls.end(); i++) {
	if (!(*i)->isEnabled())
	    continue;
	SizeRequirements reqs = (*i)->sizeRequirements();
	Region dimensions(x() + used + padding(),
			  y() + padding(),
			  reqs.minWidth,
			  height() - padding() * 2);

	if (reqs.maxWidth != reqs.minWidth && expand()) {
	    int extra = (available - 1) / interested + 1;
	    dimensions.width += extra;
	    available -= extra;
	    interested--;
	}

	(*i)->setDimensions(dimensions);
	used += dimensions.width + spacing();
    }
}
