/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include <sstream>
#include <algorithm>
#include "dockboxw.h"
#include "../connection.h"
#include "../icon.h"
#include "logic.h"
#include "configuration.h"

#define DOCKBOX_TITLE_PADDING_X 5
#define DOCKBOX_TITLE_PADDING_Y 3
#define DOCKBOX_TAB_PADDING_X 5
#define DOCKBOX_TAB_PADDING_Y 3

using namespace Toolkit::Docks;


DockboxWidget::DockboxWidget(Connection &conn, Logic &logic)
    : ContainerWidget(conn), logic(logic), active_dock(0), dock_control(*this)
{
    update_size_requirements();
}

std::vector<Dock *> DockboxWidget::docks() const
{
    return _docks;
}

void DockboxWidget::addDock(Dock *dock)
{
    if (std::find(_docks.begin(), _docks.end(), dock) != _docks.end())
	return;

    _docks.push_back(dock);
    assert(!logic.containingDockbox[dock]);
    logic.containingDockbox[dock] = this;
    dock->action.setActive(true);

    if (docks().size() == 1)
	setActiveDock(dock);

    update_size_requirements();
    if (docks().size() == 2)
	requestLayout();
}

void DockboxWidget::removeDock(Dock *dock)
{
    std::vector<Dock *>::iterator i = std::find(_docks.begin(),
						_docks.end(), dock);
    assert(i != _docks.end());

    i = _docks.erase(i);
    logic.containingDockbox[dock] = 0;
    dock->action.setActive(false);

    if (dock == active_dock) {
	if (i != _docks.end())
	    setActiveDock(*i);
	else if (!_docks.empty())
	    setActiveDock(_docks.front());
	else
	    setActiveDock(0);
    }

    update_size_requirements();
    if (docks().size() == 1)
	requestLayout();
}

Dock *DockboxWidget::activeDock() const
{
    return active_dock;
}

void DockboxWidget::setActiveDock(Dock *dock)
{
    if (dock == active_dock)
	return;

    dock_control.setWidget(dock ? dock->widget : 0);
    active_dock = dock;
    invalidate();
}

void DockboxWidget::readConfiguration(Configuration &config)
{
    if (config.listFirst())
	do {
	    addDock(logic.lookupDock(config.readLastToken()));
	} while (config.listNext());
}

void DockboxWidget::writeConfiguration(std::ostringstream &stream) const
{
    stream << "(";
    for (std::vector<Dock *>::const_iterator i = _docks.begin();
	 i != _docks.end(); i++) {
	if (i != _docks.begin())
	    stream << ", ";
	stream << (*i)->id;
    }
    stream << ")";
}

void DockboxWidget::sizeRequirementsChanged(WidgetControl */*control*/)
{
    update_size_requirements();
}

void DockboxWidget::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    int title_height = conn.fontSet().ascent() +
		       conn.fontSet().descent() +
		       DOCKBOX_TITLE_PADDING_Y * 2;
    int tab_height = conn.fontSet().ascent() +
		     conn.fontSet().descent() +
		     DOCKBOX_TAB_PADDING_Y * 2;

    if (docks().size() < 2)
	tab_height = 0;

    unsigned long bg_pixel, text_pixel;

    if (logic.focusHandler.focussedControl() == control()) {
	bg_pixel = colors().dockFocussedTitle.pixel;
	text_pixel = colors().dockFocussedTitleText.pixel;
    } else {
	bg_pixel = colors().dockTitle.pixel;
	text_pixel = colors().dockTitleText.pixel;
    }

    gc().setForeground(bg_pixel);
    pixmap.fillRectangle(gc(), x, y, width(), title_height);

    gc().setForeground(colors().dockPlasma.pixel);
    pixmap.drawPoint(gc(), x, y);
    pixmap.drawPoint(gc(), x + width() - 1, y);

    gc().setForeground(text_pixel);
    pixmap.wcDrawString(conn.fontSet(), gc(),
			x + DOCKBOX_TITLE_PADDING_X,
			y + DOCKBOX_TITLE_PADDING_Y + conn.fontSet().ascent(),
			conn.ellipsis(conn.fontSet(), active_dock->title,
				      width() - DOCKBOX_TITLE_PADDING_X * 2));

    gc().setForeground(active_dock->backgroundColor);
    pixmap.fillRectangle(gc(), x, y + title_height,
			       width(), height() - title_height - tab_height);

    ContainerWidget::paint(region, pixmap);

    if (docks().size() < 2)
	return;

    int tab_x = 0;

    unsigned int i;
    for (i = 0; i < docks().size(); i++) {
	int tab_width = (width() - tab_x) / (docks().size() - i);

	if (docks()[i] == active_dock) {
	    gc().setForeground(docks()[i]->backgroundColor);
	    pixmap.fillRectangle(gc(), x + tab_x, y + height() - tab_height,
				 tab_width, tab_height);

	    gc().setForeground(colors().dockPlasma.pixel);
	    pixmap.drawPoint(gc(), x + tab_x, y + height() - 1);
	    pixmap.drawPoint(gc(), x + tab_x + tab_width - 1,
				   y + height() - 1);

	    gc().setForeground(docks()[i]->textColor);
	} else
	    gc().setForeground(colors().dockPlasmaText.pixel);

	Icon *icon = docks()[i]->icon;

	int icon_x = x + tab_x + DOCKBOX_TAB_PADDING_X;
	int icon_y = y + height() - (tab_height + icon->height()) / 2;

	pixmap.wcDrawString(conn.fontSet(), gc(),
			    icon_x + icon->width() + DOCKBOX_TAB_PADDING_X,
			    y + height() - tab_height + DOCKBOX_TAB_PADDING_Y +
				conn.fontSet().ascent(),
			    conn.ellipsis(conn.fontSet(), docks()[i]->title,
					  tab_width - icon->width() -
					      DOCKBOX_TAB_PADDING_X * 3));

	X::Drawable::copyArea(icon->pixmap(conn, docks()[i] == active_dock
				  ? docks()[i]->backgroundColor
				  : colors().dockPlasma.pixel),
			      pixmap, gc(), 0, 0,
			      icon->width(), icon->height(), icon_x, icon_y);

	tab_x += tab_width;
    }
}

void DockboxWidget::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    int title_height = conn.fontSet().ascent() +
		       conn.fontSet().descent() +
		       DOCKBOX_TITLE_PADDING_Y * 2;

    if (y >= this->y() && y < this->y() + title_height &&
	button != Button4 && button != Button5)
	logic.focusHandler.setFocussedControl(
	    static_cast<FocussableControl *>(control()));

    if (docks().size() > 1) {
	int tab_height = conn.fontSet().ascent() +
			 conn.fontSet().descent() +
			 DOCKBOX_TAB_PADDING_Y * 2;

	if (y >= this->y() + height() - tab_height &&
	    y < this->y() + height()) {

	    if (button == Button4) {
		std::vector<Dock *>::const_iterator i =
		    std::find(_docks.begin(), _docks.end(), active_dock);
		if (i == _docks.begin() || i == _docks.end())
		    setActiveDock(docks().back());
		else
		    setActiveDock(*(i - 1));
	    } else if (button == Button5) {
		std::vector<Dock *>::const_iterator i =
		    std::find(_docks.begin(), _docks.end(), active_dock);
		if (i == _docks.end() - 1 || i == _docks.end())
		    setActiveDock(docks().front());
		else
		    setActiveDock(*(i + 1));
	    } else {
		int tab_x = 0;
		unsigned int i;

		for (i = 0; i < docks().size(); i++) {
		    int tab_width = (width() - tab_x) / (docks().size() - i);

		    if (x >= this->x() + tab_x &&
			x < this->x() + tab_x + tab_width) {
			setActiveDock(docks()[i]);
			logic.focusHandler.setFocussedControl(
			    static_cast<FocussableControl *>(control()));
			break;
		    }

		    tab_x += tab_width;
		}
	    }
	}
    }

    ContainerWidget::buttonPressEvent(time, x, y, x_root, y_root,
				      state, button);
}

void DockboxWidget::updateLayout()
{
    int title_height = conn.fontSet().ascent() +
		       conn.fontSet().descent() +
		       DOCKBOX_TITLE_PADDING_Y * 2;
    int tab_height = conn.fontSet().ascent() +
		     conn.fontSet().descent() +
		     DOCKBOX_TAB_PADDING_Y * 2;

    if (docks().size() < 2)
	tab_height = 0;

    dock_control.setDimensions(Region(x(), y() + title_height, width(),
				      height() - title_height - tab_height));
}

void DockboxWidget::update_size_requirements()
{
    SizeRequirements reqs;
    std::vector<Dock *> cached_docks = docks();

    for (std::vector<Dock *>::const_iterator i = cached_docks.begin();
	 i != cached_docks.end(); i++) {
	SizeRequirements r = (*i)->widget->sizeRequirements();

	if (r.minWidth > reqs.minWidth)
	    reqs.minWidth = r.minWidth;
	if (r.minHeight > reqs.minHeight)
	    reqs.minHeight = r.minHeight;
	if (r.maxWidth != -1 && (
		reqs.maxWidth == -1 || r.maxWidth < reqs.maxWidth))
	    reqs.maxWidth = r.maxWidth;
	if (r.maxHeight != -1 && (
		reqs.maxHeight == -1 || r.maxHeight < reqs.maxHeight))
	    reqs.maxHeight = r.maxHeight;
    }

    if (reqs.minWidth < DOCKBOX_TITLE_PADDING_X * 2)
	reqs.minWidth = DOCKBOX_TITLE_PADDING_X * 2;

    reqs.minHeight += conn.fontSet().ascent() +
		      conn.fontSet().descent() +
		      DOCKBOX_TITLE_PADDING_Y * 2;
    if (reqs.maxHeight != -1)
	reqs.maxHeight += conn.fontSet().ascent() +
			  conn.fontSet().descent() +
			  DOCKBOX_TITLE_PADDING_Y * 2;

    if (_docks.size() > 1) {
	if (reqs.minWidth < DOCKBOX_TAB_PADDING_X * 2)
	    reqs.minWidth = DOCKBOX_TAB_PADDING_X * 2;

	reqs.minHeight += conn.fontSet().ascent() +
			  conn.fontSet().descent() +
			  DOCKBOX_TAB_PADDING_Y * 2;
	if (reqs.maxHeight != -1)
	    reqs.maxHeight += conn.fontSet().ascent() +
			      conn.fontSet().descent() +
			      DOCKBOX_TAB_PADDING_Y * 2;
    }

    if (reqs.maxWidth != -1 && reqs.minWidth > reqs.maxWidth)
	reqs.maxWidth = reqs.minWidth;
    if (reqs.maxHeight != -1 && reqs.minHeight > reqs.maxHeight)
	reqs.maxHeight = reqs.minHeight;

    setSizeRequirements(reqs);
}
