/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <math.h>
#include <sstream>
#include "layoutw.h"
#include "columnw.h"
#include "logic.h"
#include "configuration.h"

#define INNER_PLASMA_X 6
#define INNER_PLASMA_Y 6

using namespace Toolkit::Docks;


class LayoutWidget::ResizeOperation
{
public:
    ResizeOperation(int grip, unsigned int stored_width)
	: grip(grip), stored_width(stored_width) {
    }
    int grip;
    unsigned int stored_width;
};


static int fixup_split_pos_x(int splitPos, int width,
			     const Toolkit::SizeRequirements &left,
			     const Toolkit::SizeRequirements &right)
{
    if (right.maxWidth != -1 &&
	splitPos < width - right.maxWidth)
	splitPos = width - right.maxWidth;
    if (splitPos > width - right.minWidth)
	splitPos = width - right.minWidth;

    if (left.maxWidth != -1 &&
	splitPos > left.maxWidth)
	splitPos = left.maxWidth;
    if (splitPos < left.minWidth)
	splitPos = left.minWidth;

    return splitPos;
}


LayoutWidget::LayoutWidget(Connection &conn, Logic &logic)
    : ContainerWidget(conn), logic(logic),
      left(new ColumnWidget(conn, logic)),
      right(new ColumnWidget(conn, logic)),
      left_control(*this), right_control(*this),
      left_share(.25), right_share(.25),
      left_width(0), right_width(0),
      left_resize(0), right_resize(0),
      mainControl(*this, logic.focusHandler)
{
    left_control.setWidget(left);
    right_control.setWidget(right);
}

LayoutWidget::~LayoutWidget()
{
    delete left_resize;
    delete right_resize;
    delete left;
    delete right;
}

void LayoutWidget::readConfiguration(Configuration &config)
{
    if (!config.listFirst())
	throw Configuration::Error();
    left->readConfiguration(config);
    config.consumeSpace();
    left_share = Configuration::toDouble(config.readLastToken());
    if (config.listNext())
	throw Configuration::Error();

    config.consumeSpace();

    if (!config.listFirst())
	throw Configuration::Error();
    right->readConfiguration(config);
    config.consumeSpace();
    right_share = Configuration::toDouble(config.readLastToken());
    if (config.listNext())
	throw Configuration::Error();

    requestLayout();
}

void LayoutWidget::writeConfiguration(std::ostringstream &stream) const
{
    stream << "(";
    left->writeConfiguration(stream);
    stream << " " << left_share << ") (";
    right->writeConfiguration(stream);
    stream << " " << right_share << ")";
}

void LayoutWidget::resetConfiguration()
{
    left->resetConfiguration();
    right->resetConfiguration();
    requestLayout();
}

void LayoutWidget::updateLayout()
{
    unsigned int net_width = width() - INNER_PLASMA_X * 2;
    left_width = lrint(net_width * left_share);
    right_width = lrint(net_width * right_share);

    SizeRequirements total_reqs(left->sizeRequirements());
    total_reqs.addX(right->sizeRequirements());

    int total_width = fixup_split_pos_x(
	left_width + right_width, net_width,
	total_reqs, mainControl.sizeRequirements());

    left_width = fixup_split_pos_x(
	left_width, total_width,
	left->sizeRequirements(), right->sizeRequirements());
    right_width = total_width - left_width;

    mainControl.setDimensions(
	Region(x() + (left->docks.empty() ? 0 : left_width + INNER_PLASMA_X),
	       y(),
	       width()
		   - (left->docks.empty() ? 0 : left_width + INNER_PLASMA_X)
		   - (right->docks.empty() ? 0 : right_width + INNER_PLASMA_X),
	       height()));

    left_control.setEnabled(!left->docks.empty());
    left_control.setDimensions(Region(x(), y(), left_width, height()));

    right_control.setEnabled(!right->docks.empty());
    right_control.setDimensions(Region(x() + width() - right_width, y(),
				       right_width, height()));
}

void LayoutWidget::sizeRequirementsChanged(WidgetControl *control)
{
    SizeRequirements reqs(mainControl.sizeRequirements());
    reqs.addX(left->sizeRequirements());
    reqs.addX(right->sizeRequirements());
    reqs.addX(INNER_PLASMA_X * 2);
    setSizeRequirements(reqs);

    requestLayout();
}

void LayoutWidget::motionEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, char is_hint)
{
    if (left_resize) {
	SizeRequirements reqs(mainControl.sizeRequirements());
	reqs.addX(INNER_PLASMA_X);
	reqs.addX(right->sizeRequirements());

	unsigned int net_width = width() - INNER_PLASMA_X * 2;
	left_width = fixup_split_pos_x(
	    x - this->x() - left_resize->grip,
	    net_width + INNER_PLASMA_X,
	    left->sizeRequirements(), reqs);
	right_width = net_width - left_width - fixup_split_pos_x(
	    net_width - left_width - left_resize->stored_width,
	    net_width - left_width,
	    mainControl.sizeRequirements(), right->sizeRequirements());

	left_share = (double)left_width / net_width;
	right_share = (double)right_width / net_width;
	requestLayout();
    }

    if (right_resize) {
	SizeRequirements reqs(mainControl.sizeRequirements());
	reqs.addX(INNER_PLASMA_X);
	reqs.addX(left->sizeRequirements());

	unsigned int net_width = width() - INNER_PLASMA_X * 2;
	right_width = net_width + INNER_PLASMA_X - fixup_split_pos_x(
	    x - this->x() - right_resize->grip,
	    net_width + INNER_PLASMA_X,
	    reqs, right->sizeRequirements());
	left_width = fixup_split_pos_x(
	    right_resize->stored_width, net_width - right_width,
	    left->sizeRequirements(), mainControl.sizeRequirements());

	left_share = (double)left_width / net_width;
	right_share = (double)right_width / net_width;
	requestLayout();
    }
}

void LayoutWidget::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    if (x >= this->x() + left_width &&
	x < this->x() + left_width + INNER_PLASMA_X) {
	left_resize = new ResizeOperation(
	    x - (this->x() + left_width), right_width);
	window().defineCursor(logic.horzResizeCursor());
    }
    if (x >= this->x() + this->width() - right_width - INNER_PLASMA_X &&
	x < this->x() + this->width() - right_width) {
	right_resize = new ResizeOperation(
	    x - (this->x() + this->width() - right_width - INNER_PLASMA_X),
	    left_width);
	window().defineCursor(logic.horzResizeCursor());
    }
}

void LayoutWidget::buttonReleaseEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    if (left_resize) {
	delete left_resize;
	left_resize = 0;
	window().undefineCursor();
    }
    if (right_resize) {
	delete right_resize;
	right_resize = 0;
	window().undefineCursor();
    }
}
