/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <err.h>
#include <assert.h>
#include <X11/cursorfont.h>
#include <sstream>
#include "logic.h"
#include "../connection.h"
#include "dockboxw.h"
#include "configuration.h"

using namespace Toolkit::Docks;


Logic::Logic(Connection &conn)
    : main_widget(conn), layout_widget(conn, *this),
      horz_resize_cursor(
	  conn.display().createFontCursor(XC_sb_h_double_arrow)),
      vert_resize_cursor(
	  conn.display().createFontCursor(XC_sb_v_double_arrow)),
      focusHandler(layout_widget.mainControl)
{
    main_widget.mainControl.setWidget(&layout_widget);
}

Logic::~Logic()
{
    assert(docks_by_id.empty());
    horz_resize_cursor.free();
    vert_resize_cursor.free();
}

void Logic::addDock(Dock *dock)
{
    assert(docks_by_id.find(dock->id) == docks_by_id.end());
    docks_by_id[dock->id] = dock;
}

void Logic::removeDock(Dock *dock)
{
    std::map<std::string, Dock *>::iterator i = docks_by_id.find(dock->id);
    assert(i != docks_by_id.end());
    assert((*i).second == dock);
    docks_by_id.erase(i);
}

Toolkit::Widget *Logic::widget()
{
    return &main_widget;
}

Toolkit::WidgetControl *Logic::mainControl()
{
    return &layout_widget.mainControl;
}

Dock *Logic::focussedDock() const
{
    FocussableControl *control = focusHandler.focussedControl();
    if (control == &layout_widget.mainControl)
	return 0;

    return static_cast<DockboxWidget *>(control->widget())->activeDock();
}

void Logic::setFocussedDock(Dock *dock)
{
    if (!dock) {
	focusHandler.setFocussedControl(&layout_widget.mainControl);
	return;
    }

    DockboxWidget *dockbox = containingDockbox[dock];
    if (!dockbox)
	return; /* not yet implemented */

    dockbox->setActiveDock(dock);
    focusHandler.setFocussedControl(
	static_cast<FocussableControl *>(dockbox->control()));
}

std::string Logic::configuration() const
{
    std::ostringstream stream;
    layout_widget.writeConfiguration(stream);
    stream << " () () ()";
    return stream.str();
}

void Logic::setConfiguration(const std::string &configuration)
{
    try {
	Configuration config(configuration);
	layout_widget.resetConfiguration();
	layout_widget.readConfiguration(config);
	config.consumeSpace();
	if (config.listFirst())
	    throw Configuration::Error();
	config.consumeSpace();
	if (config.listFirst())
	    throw Configuration::Error();
	config.consumeSpace();
	if (config.listFirst())
	    throw Configuration::Error();
	if (!config.fullyConsumed())
	    throw Configuration::Error();
    } catch (Configuration::Error &e) {
	warnx("Logic::setConfiguration: Bad configuration string");
	layout_widget.resetConfiguration();
    }
}

Dock *Logic::lookupDock(const std::string &id)
{
    return docks_by_id.at(id);
}

X::Cursor Logic::horzResizeCursor() const
{
    return horz_resize_cursor;
}

X::Cursor Logic::vertResizeCursor() const
{
    return vert_resize_cursor;
}
