/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <err.h>
#include <assert.h>
#include <algorithm>
#include "hooks.h"
#include "connection.h"
#include "selection.h"
#include "statusbar.h"
#include "action.h"

using namespace Toolkit;

static XGCValues gc_defaults =
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };


class Toolkit::WindowWidgetControl : public WidgetControl
{
    X::GC _gc;
    WidgetWindowHook *const _hook;
public:
    WindowWidgetControl(WidgetWindowHook *hook)
      : WidgetControl(hook->conn), _hook(hook) {
	_gc = window().createGC(GCGraphicsExposures, &gc_defaults);
    }
    ~WindowWidgetControl() {
	_gc.free();
    }

    X::Window window() const {
	return _hook->window();
    }
    X::GC gc() const {
	return _gc;
    }
    WidgetWindowHook *hook() const {
	return _hook;
    }

    void focus(Focussable *focussable) {
	/* ignore */
    }
    void focussablesChanged() {
	/* ignore */
    }
    void setDimensions(const Region &dimensions) {
	WidgetControl::setDimensions(dimensions);
    }
private:
    void sizeRequirementsChanged() {
	SizeRequirements reqs(sizeRequirements());
	assert(reqs.minWidth != -1);
	assert(reqs.minHeight != -1); /* min... == -1 is deprecated */

	XSizeHints size_hints;
	size_hints.flags = PMinSize;
	size_hints.min_width = reqs.minWidth;
	size_hints.min_height = reqs.minHeight;
	if (reqs.maxWidth != -1 && reqs.maxHeight != -1) {
	    size_hints.flags |= PMaxSize;
	    size_hints.max_width = reqs.maxWidth;
	    size_hints.max_height = reqs.maxHeight;
	}
	window().setWMNormalHints(&size_hints);
    }
private:
    void propagateChildLayoutRequest() {
	assert(!_hook->is_updating0);
    }
};


Destructible::~Destructible()
{
    /* virtualize */
}

WindowHook::WindowHook(Connection &conn, X::Window window,
		       Destructible *destructible)
    : _window(window), destructible(destructible), conn(conn)
{
    conn.addHook(this);
}

WindowHook::~WindowHook()
{
    conn.removeHook(this);
}

X::Window WindowHook::window() const
{
    return _window;
}

void WindowHook::update()
{
    /* relax */
}

void WindowHook::keyPress(X::Window root, X::Window subwindow,
			  const XKeyEvent &event)
{
    /* relax */
}

void WindowHook::keyRelease(X::Window root, X::Window subwindow,
			    const XKeyEvent &event)
{
    /* relax */
}

void WindowHook::buttonPress(X::Window root, X::Window subwindow,
			     const XButtonEvent &event)
{
    /* relax */
}

void WindowHook::buttonRelease(X::Window root, X::Window subwindow,
			       const XButtonEvent &event)
{
    /* relax */
}

void WindowHook::motionNotify(X::Window root, X::Window subwindow,
			      const XMotionEvent &event)
{
    /* relax */
}

void WindowHook::enterNotify(X::Window root, X::Window subwindow,
			     const XCrossingEvent &event)
{
    /* relax */
}

void WindowHook::leaveNotify(X::Window root, X::Window subwindow,
			     const XCrossingEvent &event)
{
    /* relax */
}

void WindowHook::focusIn(const XFocusChangeEvent &event)
{
    /* relax */
}

void WindowHook::focusOut(const XFocusChangeEvent &event)
{
    /* relax */
}

void WindowHook::expose(const XExposeEvent &event)
{
    /* relax */
}

void WindowHook::visibilityNotify(const XVisibilityEvent &event)
{
    /* relax */
}

void WindowHook::destroyNotify(const XDestroyWindowEvent &event)
{
    delete destructible;
}

void WindowHook::unmapNotify(const XUnmapEvent &event)
{
    /* relax */
}

void WindowHook::mapNotify(const XMapEvent &event)
{
    /* relax */
}

void WindowHook::reparentNotify(const XReparentEvent &event)
{
    /* relax */
}

void WindowHook::configureNotify(const XConfigureEvent &event)
{
    /* relax */
}

void WindowHook::selectionClear(const XSelectionClearEvent &event)
{
    /* relax */
}

void WindowHook::selectionRequest(X::Window requestor,
				  const XSelectionRequestEvent &event)
{
    /* relax */
}

void WindowHook::selectionNotify(const XSelectionEvent &event)
{
    /* relax */
}

void WindowHook::deleteWindow(const XClientMessageEvent &event)
{
    /* relax */
}

/****************************************************************************/

SelectionWindowHook::SelectionWindowHook(Connection &conn, X::Window window,
					 Destructible *destructible)
    : WindowHook(conn, window, destructible)
{
}

SelectionWindowHook::~SelectionWindowHook()
{
    for (std::map<Atom, SelectionHook *>::const_iterator i =
	     selections.begin(); i != selections.end(); i++)
	delete (*i).second;

    for (std::map<Atom, ConversionHook *>::const_iterator i =
	     conversions.begin(); i != conversions.end(); i++)
	delete (*i).second;
}

void SelectionWindowHook::clear_selection(Atom selection, Time time)
{
    SelectionHook *hook = selections[selection];
    if (!hook)
	return;

    selections[selection] = 0;
    hook->clear(time);
    delete hook;
}

void SelectionWindowHook::cancel_conversion(Atom selection, Time time)
{
    ConversionHook *hook = conversions[selection];
    if (!hook)
	return;

    conversions[selection] = 0;
    hook->cancel(time);
    delete hook;
}

void SelectionWindowHook::select(SelectionHook *hook)
{
    clear_selection(hook->selection, hook->time);
    assert(selections[hook->selection] == 0);

    selections[hook->selection] = hook;
    conn.display().setSelectionOwner(hook->selection, window(), hook->time);

    if (conn.display().getSelectionOwner(hook->selection) != window())
	clear_selection(hook->selection, hook->time);
}

void SelectionWindowHook::convertSelection(ConversionHook *hook)
{
    cancel_conversion(hook->selection, hook->time);
    assert(conversions[hook->selection] == 0);

    conversions[hook->selection] = hook;
    conn.display().convertSelection(hook->selection, hook->target,
				    hook->selection, window(), hook->time);
}

void SelectionWindowHook::selectionClear(const XSelectionClearEvent &event)
{
    clear_selection(event.selection, event.time);
}

void SelectionWindowHook::selectionRequest(X::Window requestor,
					   const XSelectionRequestEvent &event)
{
    SelectionHook *hook = selections[event.selection];
    if (!hook)
	return;

    XEvent ev;
    ev.type = SelectionNotify;
    ev.xselection.display = event.display;
    ev.xselection.requestor = event.requestor;
    ev.xselection.selection = event.selection;
    ev.xselection.target = event.target;
    ev.xselection.property = hook->request(requestor, event.target,
					   event.property, event.time)
				 ? event.property : None;
    ev.xselection.time = event.time;
    requestor.sendEvent(False, 0l, &ev);
}

void SelectionWindowHook::selectionNotify(const XSelectionEvent &event)
{
    ConversionHook *hook = conversions[event.selection];
    if (!hook)
	return;

    conversions[event.selection] = 0;

    if (event.time < hook->time)
	warnx("Dropping stale SelectionNotify event");
    else {
	if (event.property != None && event.target != hook->target)
	    warnx("Target %ld ('%s') does not match "
		  "requested target %ld ('%s')",
		  event.target, conn.display().getAtomName(event.target),
		  hook->target, conn.display().getAtomName(hook->target));

	hook->notify(event.property, event.time);
    }

    delete hook;
}

/****************************************************************************/

WidgetWindowHook::WidgetWindowHook(Connection &conn, X::Window window,
				   Destructible *destructible,
				   int x, int y, int width, int height,
				   int border_width, WidgetWindowHook *parent,
				   bool holds_menu)
    : SelectionWindowHook(conn, window, destructible),
      _control(new WindowWidgetControl(this)),
      outer_dimensions(x, y, width + border_width * 2,
			     height + border_width * 2),
      inner_dimensions(x + border_width, y + border_width, width, height),
      border_width(border_width), reparented_to(0),
      hovered_widget(0), autograbbed_widget(0),
      parent(parent), holds_menu(holds_menu),
      has_focus(false), focus_handler(0), status_bar(0),
      is_updating0(false), is_updating1(false), is_closed(false),
      is_mapped(false), visibility_state(VisibilityFullyObscured)
{
    _control->setDimensions(Region(0, 0, width, height));
    if (parent)
	parent->add_child(this);
}

WidgetWindowHook::~WidgetWindowHook()
{
    if (parent)
	parent->remove_child(this);
    assert(children.empty());
    delete _control;
}

WidgetControl *WidgetWindowHook::control() const
{
    return _control;
}

Toolkit::Region WidgetWindowHook::outerDimensions() const
{
    return outer_dimensions;
}

Toolkit::Region WidgetWindowHook::innerDimensions() const
{
    return inner_dimensions;
}

void WidgetWindowHook::setOuterDimensions(Region const &dimensions)
{
    /* Don't update outer_dimensions, configureNotify will do that.
       (This implies we cannot skip moveResize even if the dimensions didn't
       change--there might be a pending XMoveResizeWindow in the X queue.) */

    window().moveResize(dimensions.x, dimensions.y,
			dimensions.width - border_width * 2,
			dimensions.height - border_width * 2);
}

void WidgetWindowHook::setInnerDimensions(Region const &dimensions)
{
    /* Don't update inner_dimensions, configureNotify will do that.
       (This implies we cannot skip moveResize even if the dimensions didn't
       change--there might be a pending XMoveResizeWindow in the X queue.) */

    window().moveResize(dimensions.x - border_width,
			dimensions.y - border_width,
			dimensions.width, dimensions.height);
}

unsigned int WidgetWindowHook::borderWidth() const
{
    return border_width;
}

void WidgetWindowHook::set_hovered_widget(Widget *w, Time time,
					  int x, int y, int x_root, int y_root,
					  unsigned int state)
{
    if (w == hovered_widget)
	return;

    if (hovered_widget)
	hovered_widget->leaveEvent(time, x, y, x_root, y_root, state);
    hovered_widget = w;
    if (hovered_widget)
	hovered_widget->enterEvent(time, x, y, x_root, y_root, state);

    if (statusBar())
	statusBar()->setMessage(
	    hovered_widget ? hovered_widget->statusMessage() : std::string());
}

void WidgetWindowHook::add_child(WidgetWindowHook *child)
{
    assert(std::find(children.begin(),
		     children.end(), child) == children.end());
    children.push_back(child);
}

void WidgetWindowHook::remove_child(WidgetWindowHook *child)
{
    std::vector<WidgetWindowHook *>::iterator i =
	std::find(children.begin(), children.end(), child);
    assert(i != children.end());
    children.erase(i);
}

void WidgetWindowHook::set_mnemonic_visibility(bool visible)
{
    if (control()->widget())
	control()->widget()->setMnemonicVisibility(visible);

    for (std::vector<WidgetWindowHook *>::const_iterator i = children.begin();
	 i != children.end(); i++)
	(*i)->set_mnemonic_visibility(visible);
}

StatusBar *WidgetWindowHook::statusBar() const
{
    return parent ? parent->statusBar() : status_bar;
}

void WidgetWindowHook::setStatusBar(StatusBar *statusBar)
{
    if (parent) {
	parent->setStatusBar(statusBar);
	return;
    }
    if (statusBar == status_bar)
	return;

    if (status_bar)
	status_bar->setMessage(std::string());
    status_bar = statusBar;
    if (status_bar)
	status_bar->setMessage(
	    hovered_widget ? hovered_widget->statusMessage() : std::string());

    for (std::vector<WidgetWindowHook *>::const_iterator i = children.begin();
	 i != children.end(); i++)
	(*i)->updateStatusMessage();
}

void WidgetWindowHook::updateStatusMessage()
{
    if (statusBar() && hovered_widget)
	statusBar()->setMessage(hovered_widget->statusMessage());
}

void WidgetWindowHook::update()
{
    assert(!is_updating0);
    is_updating0 = true;

    _control->updateChildLayout();

    if (!is_mapped || visibility_state == VisibilityFullyObscured)
	update_pending.clear();
    else {
	is_updating1 = true;
	while (!update_pending.empty()) {
	    Region region(update_pending.front());
	    update_pending.erase(update_pending.begin());
	    repaint(region);
	}
	is_updating1 = false;
    }

    is_updating0 = false;
    last_exposed_region = Region();
}

Widget *WidgetWindowHook::hoveredWidget() const
{
    return hovered_widget;
}

bool WidgetWindowHook::hasFocus() const
{
    return has_focus;
}

void WidgetWindowHook::close()
{
    if (is_closed)
	return;
    is_closed = true;
    window().unmap();
}

Focussable *WidgetWindowHook::focusHandler() const
{
    return focus_handler;
}

void WidgetWindowHook::setFocusHandler(Focussable *focusHandler)
{
    if (focusHandler == focus_handler)
	return;

    focus_handler = focusHandler;
}

void WidgetWindowHook::keyPress(X::Window root, X::Window subwindow,
				const XKeyEvent &event)
{
    KeySym latin1_keysym = XLookupKeysym(const_cast<XKeyEvent *>(&event), 0);
    if (latin1_keysym == XK_Alt_L || latin1_keysym == XK_Alt_R)
	set_mnemonic_visibility(true);
}

void WidgetWindowHook::keyRelease(X::Window root, X::Window subwindow,
				  const XKeyEvent &event)
{
    KeySym latin1_keysym = XLookupKeysym(const_cast<XKeyEvent *>(&event), 0);
    if (latin1_keysym == XK_Alt_L || latin1_keysym == XK_Alt_R)
	set_mnemonic_visibility(false);
}

void WidgetWindowHook::buttonPress(X::Window root, X::Window subwindow,
				   const XButtonEvent &event)
{
    Widget *w = control()->widget() ?
		control()->widget()->widgetAt(event.x, event.y) : 0;

    if (conn.rootMenu() && !holds_menu && (!w || !w->isMenu())) {
	conn.leaveMenuMode(event.time);
	conn.notifyAutomaticGrab(event.time, 0, this);
	return;
    }

    if (event.state & ButtonMask) {
	if (autograbbed_widget)
	    autograbbed_widget->buttonPressEvent(event.time, event.x, event.y,
						 event.x_root, event.y_root,
						 event.state, event.button);
	return;
    }

    conn.notifyAutomaticGrab(event.time, w, this);

    if (w)
	w->buttonPressEvent(event.time, event.x, event.y,
			    event.x_root, event.y_root,
			    event.state, event.button);
}

void WidgetWindowHook::buttonRelease(X::Window root, X::Window subwindow,
				     const XButtonEvent &event)
{
    Widget *w0 = autograbbed_widget;

    Widget *w = control()->widget() ?
		control()->widget()->widgetAt(event.x, event.y) : 0;

    Widget *w1 = w != autograbbed_widget && w && w->isMenu() ? w : 0;
    Widget *w2 = 0, *w3 = 0;

    /* assuming Button<n> = n, Button<n>Mask = 1 << (n + 7) */
    unsigned int modified_state = event.state & ~(1 << (event.button + 7));

    if ((modified_state & ButtonMask) == 0) {
	if (conn.rootMenu() && !holds_menu && w && !w->isMenu())
	    w = 0;

	if (w != hovered_widget) {
	    w2 = hovered_widget;
	    w3 = hovered_widget = w;

	    if (statusBar())
		statusBar()->setMessage(
		    hovered_widget ?
		    hovered_widget->statusMessage() : std::string());
	}

	conn.notifyAutomaticUngrab(event.time, this);
    }

    if (w0)
	w0->buttonReleaseEvent(event.time, event.x, event.y,
			       event.x_root, event.y_root,
			       event.state, event.button);
    if (w1)
	w1->buttonReleaseEvent(event.time, event.x, event.y,
			       event.x_root, event.y_root,
			       event.state, event.button);
    if (w2)
	w2->leaveEvent(event.time, event.x, event.y,
		       event.x_root, event.y_root, modified_state);
    if (w3)
	w3->enterEvent(event.time, event.x, event.y,
		       event.x_root, event.y_root, modified_state);
}

void WidgetWindowHook::motionNotify(X::Window root, X::Window subwindow,
				    const XMotionEvent &event)
{
    Widget *w = control()->widget() ?
		control()->widget()->widgetAt(event.x, event.y) : 0;
    Widget *p = hovered_widget;

    if (conn.rootMenu() && !holds_menu && (!w || !w->isMenu()))
	w = 0;

    if (event.state & ButtonMask) {
	if (w != autograbbed_widget && (!w || !w->isMenu()))
	    w = 0;

	set_hovered_widget(w,
			   event.time, event.x, event.y,
			   event.x_root, event.y_root, event.state);

	if (autograbbed_widget)
	    autograbbed_widget->motionEvent(event.time, event.x, event.y,
					    event.x_root, event.y_root,
					    event.state, event.is_hint);

	if (p && hovered_widget == p && p != autograbbed_widget)
	    p->motionEvent(event.time, event.x, event.y,
			   event.x_root, event.y_root,
			   event.state, event.is_hint);
    } else {
	set_hovered_widget(w,
			   event.time, event.x, event.y,
			   event.x_root, event.y_root, event.state);

	if (p && hovered_widget == p)
	    p->motionEvent(event.time, event.x, event.y,
			   event.x_root, event.y_root,
			   event.state, event.is_hint);
    }
}

void WidgetWindowHook::enterNotify(X::Window root, X::Window subwindow,
				   const XCrossingEvent &event)
{
    Widget *w = control()->widget() ?
		control()->widget()->widgetAt(event.x, event.y) : 0;

    if (conn.rootMenu() && !holds_menu && (!w || !w->isMenu()))
	w = 0;
    if ((event.state & ButtonMask) && w != autograbbed_widget &&
	(!autograbbed_widget || !autograbbed_widget->isMenu()))
	w = 0;

    set_hovered_widget(w,
		       event.time, event.x, event.y,
		       event.x_root, event.y_root, event.state);
}

void WidgetWindowHook::leaveNotify(X::Window root, X::Window subwindow,
				   const XCrossingEvent &event)
{
    set_hovered_widget(0,
		       event.time, event.x, event.y,
		       event.x_root, event.y_root, event.state);
}

void WidgetWindowHook::focusIn(const XFocusChangeEvent &event)
{
    if (event.detail != NotifyNonlinear)
	return;

    if (has_focus)
	return;
    has_focus = true;
    if (focus_handler)
	focus_handler->focusInEvent();
}

void WidgetWindowHook::focusOut(const XFocusChangeEvent &event)
{
    if (event.detail != NotifyNonlinear)
	return;

    conn.leaveMenuMode(CurrentTime);
    set_mnemonic_visibility(false);

    if (!has_focus)
	return;
    has_focus = false;
    if (focus_handler)
	focus_handler->focusOutEvent();
}

void WidgetWindowHook::expose(const XExposeEvent &event)
{
    /* If there has been another exposure since the last update, it's
     * probably a window being dragged.  Merge exposure areas to avoid
     * fragmentation of the update queue. */

    last_exposed_region = last_exposed_region.hull(
	Region(event.x, event.y, event.width, event.height));

    invalidate(last_exposed_region);
}

void WidgetWindowHook::visibilityNotify(const XVisibilityEvent &event)
{
    visibility_state = event.state;
}

void WidgetWindowHook::unmapNotify(const XUnmapEvent &event)
{
    is_mapped = false;
    visibility_state = VisibilityFullyObscured;
    set_hovered_widget(0, CurrentTime, 0, 0, 0, 0, 0);

    if (is_closed)
	window().destroy();
}

void WidgetWindowHook::mapNotify(const XMapEvent &event)
{
    is_mapped = true;
}

void WidgetWindowHook::reparentNotify(const XReparentEvent &event)
{
    reparented_to = event.parent;
}

void WidgetWindowHook::configureNotify(const XConfigureEvent &event)
{
    Region old_pos(inner_dimensions);

    if (reparented_to && event.above != reparented_to) {
	/* The synthetic configure event does not arrive in time for
	 * the first update, so take width and height from the early
	 * event but throw away x and y */
	outer_dimensions = Region(outer_dimensions.x, outer_dimensions.y,
				  event.width + event.border_width * 2,
				  event.height + event.border_width * 2);
	inner_dimensions = Region(inner_dimensions.x, inner_dimensions.y,
				  event.width, event.height);
    } else {
	outer_dimensions = Region(event.x, event.y,
				  event.width + event.border_width * 2,
				  event.height + event.border_width * 2);
	inner_dimensions = Region(event.x + event.border_width,
				  event.y + event.border_width,
				  event.width, event.height);
    }
    border_width = event.border_width;

    _control->setDimensions(Region(0, 0, event.width, event.height));
    if (_control->widget() && (inner_dimensions.x != old_pos.x ||
			       inner_dimensions.y != old_pos.y))
	_control->widget()->windowPositionChanged();
}

void WidgetWindowHook::deleteWindow(const XClientMessageEvent &event)
{
    close();
}

void WidgetWindowHook::invalidate(const Region &region)
{
    /* this is not technically a problem, but indicates bad code */
    assert(!is_updating1);

    if (region.x < 0)
	return invalidate(Region(0, region.y,
				 region.width + region.x, region.height));
    if (region.y < 0)
	return invalidate(Region(region.x, 0,
				 region.width, region.height + region.y));
    if (region.width <= 0 || region.height <= 0)
	return;

    for (std::vector<Region>::iterator i = update_pending.begin();
	 i != update_pending.end(); i++) {
	if ((*i).contains(region))
	    return;
	if (region.contains(*i)) {
	    *i = region;
	    return;
	}
    }

    update_pending.push_back(region);
}

void WidgetWindowHook::repaint(const Region &region)
{
    if (!control()->widget())
	return;

    X::GC gc = window().createGC(GCGraphicsExposures, &gc_defaults);
    X::Pixmap pixmap = window().createPixmap(
	region.width, region.height, conn.visualInfo().depth);

    control()->widget()->paint(region, pixmap);
    X::Drawable::copyArea(pixmap, window(), gc, 0, 0,
			  region.width, region.height, region.x, region.y);

    gc.free();
    pixmap.free();
}

void WidgetWindowHook::withdrawWidget(Widget *widget)
{
    /* Passed widget may be in destruction */

    if (widget == hovered_widget)
	hovered_widget = 0;

    if (widget == autograbbed_widget)
	autograbbed_widget = 0;
}

/****************************************************************************/

InputWindowHook::InputWindowHook(Connection &conn, X::Window window,
				 Destructible *destructible,
				 int x, int y, int width, int height,
				 int border_width)
    : WidgetWindowHook(conn, window, destructible,
		       x, y, width, height, border_width, 0, false)
{
    if ((input_context = X::InputContext(
	     XCreateIC(conn.inputMethod(),
		       XNInputStyle, conn.ximStyle(),
		       XNClientWindow, (int)window,
		       XNFocusWindow, (int)window, NULL))) == NULL)
	errx(EXIT_FAILURE, "XCreateIC: Input context could not be created");
}

void InputWindowHook::registerActionSet(ActionSet *actionSet)
{
    assert(std::find(action_sets.begin(),
		     action_sets.end(), actionSet) == action_sets.end());
    action_sets.push_back(actionSet);
}

void InputWindowHook::unregisterActionSet(ActionSet *actionSet)
{
    std::vector<ActionSet *>::iterator i = std::find(
	action_sets.begin(), action_sets.end(), actionSet);
    assert(i != action_sets.end());
    action_sets.erase(i);
}

void InputWindowHook::keyPress(X::Window root, X::Window subwindow,
			       const XKeyEvent &event)
{
    KeySym keysym = NoSymbol;
    Status status = XLookupNone;
    std::string str;

    str = input_context.mbLookupString(&event, &keysym, &status);
    if (status != XLookupChars && status != XLookupBoth)
	str = std::string();
    if (status != XLookupKeySym && status != XLookupBoth)
	keysym = NoSymbol;

    if (focusHandler())
	focusHandler()->keyPressEvent(
	    event.time, event.x, event.y, event.x_root, event.y_root,
	    event.state, event.keycode, str, keysym);

    WidgetWindowHook::keyPress(root, subwindow, event);

    Hotkey hotkey(incomplete_hotkey, event.state, keysym,
		  XLookupKeysym(const_cast<XKeyEvent *>(&event), 0));
    incomplete_hotkey.clear();
    if (statusBar())
	statusBar()->setIncompleteHotkey(incomplete_hotkey);

    if (hotkey.isNull() || hotkey.isInvalid())
	return;

    Action *match = 0;

    for (std::vector<ActionSet *>::const_iterator i = action_sets.begin();
	 i != action_sets.end(); i++) {
	std::vector<Action *> actions = (*i)->actions();

	for (std::vector<Action *>::const_iterator j = actions.begin();
	     j != actions.end(); j++)
	    if ((*j)->hotkey().equals(hotkey))
		match = *j;
	    else if ((*j)->hotkey().startsWith(hotkey)) {
		incomplete_hotkey = hotkey;
		if (statusBar())
		    statusBar()->setIncompleteHotkey(incomplete_hotkey);
		return;
	    }
    }

    if (match && !match->isDisabled())
	match->trigger();
}

void InputWindowHook::keyRelease(X::Window root, X::Window subwindow,
				 const XKeyEvent &event)
{
    if (focusHandler())
	focusHandler()->keyReleaseEvent(
	    event.time, event.x, event.y, event.x_root, event.y_root,
	    event.state, event.keycode);

    WidgetWindowHook::keyRelease(root, subwindow, event);
}

void InputWindowHook::focusIn(const XFocusChangeEvent &event)
{
    if (event.detail != NotifyNonlinear)
	return;

    if (!hasFocus())
	input_context.setFocus();
    WidgetWindowHook::focusIn(event);
}

void InputWindowHook::focusOut(const XFocusChangeEvent &event)
{
    if (event.detail != NotifyNonlinear)
	return;

    bool had_focus = hasFocus();
    WidgetWindowHook::focusOut(event);
    if (had_focus)
	input_context.unsetFocus();
}
