/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#ifndef TOOLKIT_HOOKS_H
#define TOOLKIT_HOOKS_H

#include <map>
#include "util/region.h"
#include "x_abstraction.h"
#include "hotkey.h"

namespace Toolkit {

class Connection;
class SelectionHook;
class ConversionHook;
class WindowWidgetControl; /* internal */
class Widget;
class Focussable;
class StatusBar;
class WidgetControl;
class ActionSet;

class Destructible
{
public:
    virtual ~Destructible();
};

class WindowHook
{
    X::Window _window;
    Destructible *const destructible;
public:
    WindowHook(Connection &conn, X::Window window, Destructible *destructible);
    virtual ~WindowHook();
    Connection &conn;
    X::Window window() const;

    virtual void update();

    virtual void keyPress(X::Window root, X::Window subwindow,
			  const XKeyEvent &event);
    virtual void keyRelease(X::Window root, X::Window subwindow,
			    const XKeyEvent &event);
    virtual void buttonPress(X::Window root, X::Window subwindow,
			     const XButtonEvent &event);
    virtual void buttonRelease(X::Window root, X::Window subwindow,
			       const XButtonEvent &event);
    virtual void motionNotify(X::Window root, X::Window subwindow,
			      const XMotionEvent &event);
    virtual void enterNotify(X::Window root, X::Window subwindow,
			     const XCrossingEvent &event);
    virtual void leaveNotify(X::Window root, X::Window subwindow,
			     const XCrossingEvent &event);
    virtual void focusIn(const XFocusChangeEvent &event);
    virtual void focusOut(const XFocusChangeEvent &event);
    virtual void expose(const XExposeEvent &event);
    virtual void visibilityNotify(const XVisibilityEvent &event);
    virtual void destroyNotify(const XDestroyWindowEvent &event);
    virtual void unmapNotify(const XUnmapEvent &event);
    virtual void mapNotify(const XMapEvent &event);
    virtual void reparentNotify(const XReparentEvent &event);
    virtual void configureNotify(const XConfigureEvent &event);

    virtual void selectionClear(const XSelectionClearEvent &event);
    virtual void selectionRequest(X::Window requestor,
				  const XSelectionRequestEvent &event);
    virtual void selectionNotify(const XSelectionEvent &event);

    virtual void deleteWindow(const XClientMessageEvent &event);
};

class SelectionWindowHook : public WindowHook
{
    std::map<Atom, SelectionHook *> selections;
    std::map<Atom, ConversionHook *> conversions;
    void clear_selection(Atom selection, Time time);
    void cancel_conversion(Atom selection, Time time);
public:
    SelectionWindowHook(Connection &conn, X::Window window,
			Destructible *destructible);
    ~SelectionWindowHook();

    void select(SelectionHook *hook);
    void convertSelection(ConversionHook *hook);

    void selectionClear(const XSelectionClearEvent &event);
    void selectionRequest(X::Window requestor,
			  const XSelectionRequestEvent &event);
    void selectionNotify(const XSelectionEvent &event);
};

class WidgetWindowHook : public SelectionWindowHook
{
    friend class Connection; /* writes autograbbed_widget */
    friend class WindowWidgetControl; /* reads is_updating0 */

    WindowWidgetControl *const _control;
    Region outer_dimensions, inner_dimensions;
    unsigned int border_width;
    Window reparented_to;
    Widget *hovered_widget;
    Widget *autograbbed_widget;
    WidgetWindowHook *const parent;
    std::vector<WidgetWindowHook *> children;
    bool const holds_menu;
    bool has_focus;
    Focussable *focus_handler;
    StatusBar *status_bar;
    std::vector<Region> update_pending;
    Region last_exposed_region;
    bool is_updating0, is_updating1;
    bool is_closed;
    bool is_mapped;
    int visibility_state;
    void set_hovered_widget(Widget *w, Time time, int x, int y,
			    int x_root, int y_root, unsigned int state);
    void add_child(WidgetWindowHook *child);
    void remove_child(WidgetWindowHook *child);
    void set_mnemonic_visibility(bool visible);
    void repaint(const Region &region);

public:
    WidgetWindowHook(Connection &conn, X::Window window,
		     Destructible *destructible,
		     int x, int y, int width, int height, int border_width,
		     WidgetWindowHook *parent, bool holds_menu);
    ~WidgetWindowHook();
    WidgetControl *control() const;
    Region outerDimensions() const;
    Region innerDimensions() const;
    void setOuterDimensions(Region const &dimensions);
    void setInnerDimensions(Region const &dimensions);
    unsigned int borderWidth() const;
    Widget *hoveredWidget() const;
    void close();

    void invalidate(const Region &region);
    void withdrawWidget(Widget *widget);

    Focussable *focusHandler() const;
    void setFocusHandler(Focussable *focusHandler);

    StatusBar *statusBar() const;
    void setStatusBar(StatusBar *statusBar);
    void updateStatusMessage();

    void update();

    void keyPress(X::Window root, X::Window subwindow,
		  const XKeyEvent &event);
    void keyRelease(X::Window root, X::Window subwindow,
		    const XKeyEvent &event);
    void buttonPress(X::Window root, X::Window subwindow,
		     const XButtonEvent &event);
    void buttonRelease(X::Window root, X::Window subwindow,
		       const XButtonEvent &event);
    void motionNotify(X::Window root, X::Window subwindow,
		      const XMotionEvent &event);
    void enterNotify(X::Window root, X::Window subwindow,
		     const XCrossingEvent &event);
    void leaveNotify(X::Window root, X::Window subwindow,
		     const XCrossingEvent &event);
    void focusIn(const XFocusChangeEvent &event);
    void focusOut(const XFocusChangeEvent &event);
    void expose(const XExposeEvent &event);
    void visibilityNotify(const XVisibilityEvent &event);
    void unmapNotify(const XUnmapEvent &event);
    void mapNotify(const XMapEvent &event);
    void reparentNotify(const XReparentEvent &event);
    void configureNotify(const XConfigureEvent &event);

    void deleteWindow(const XClientMessageEvent &event);
protected:
    bool hasFocus() const;
};

class InputWindowHook : public WidgetWindowHook
{
    X::InputContext input_context;
    std::vector<ActionSet *> action_sets;
    Hotkey incomplete_hotkey;
public:
    InputWindowHook(Connection &conn, X::Window window,
		    Destructible *destructible,
		    int x, int y, int width, int height, int border_width);
    void registerActionSet(ActionSet *actionSet);
    void unregisterActionSet(ActionSet *actionSet);

    void keyPress(X::Window root, X::Window subwindow,
		  const XKeyEvent &event);
    void keyRelease(X::Window root, X::Window subwindow,
		    const XKeyEvent &event);
    void focusIn(const XFocusChangeEvent &event);
    void focusOut(const XFocusChangeEvent &event);
};

} /* namespace Toolkit */

#endif
