/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <stdlib.h>
#include <stdint.h>
#include <err.h>
#include "icon.h"
#include "connection.h"

struct icon_data {
    unsigned int width;
    unsigned int height;
    unsigned int bytes_per_pixel; /* 3:RGB, 4:RGBA */
    unsigned char pixel_data[];
};

static XGCValues gc_defaults =
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

using namespace Toolkit;


class Toolkit::IconPixmap
{
    Connection *const _conn;
    unsigned long const _pixel;
    bool const _disabled;
    X::Pixmap _pixmap;
public:
    IconPixmap(Connection *conn, unsigned long pixel, bool disabled,
	       struct icon_data const *icon_data);
    ~IconPixmap();
    Connection *conn() const;
    unsigned long pixel() const;
    bool disabled() const;
    X::Pixmap pixmap() const;
};

static uint8_t shift_width(unsigned long mask)
{
    if (mask == 0)
	return 0;

    uint8_t i = 0;
    while (!((mask >> i) & 1))
	i++;
    return i;
}

IconPixmap::IconPixmap(Connection *conn, unsigned long pixel, bool disabled,
		       struct icon_data const *icon_data)
    : _conn(conn), _pixel(pixel), _disabled(disabled)
{
    XVisualInfo const &vi = conn->visualInfo();
    uint8_t red_shift = shift_width(vi.red_mask);
    uint8_t green_shift = shift_width(vi.green_mask);
    uint8_t blue_shift = shift_width(vi.blue_mask);
    int red_bg = (pixel & vi.red_mask) >> red_shift;
    int green_bg = (pixel & vi.green_mask) >> green_shift;
    int blue_bg = (pixel & vi.blue_mask) >> blue_shift;

    int width = icon_data->width;
    int height = icon_data->height;

    int depth = 24;
    int bits_per_pixel = 32;
    int scanline_pad = 32;

    uint32_t *data = (uint32_t *)calloc(width * height, bits_per_pixel / 8);

    XImage image;
    image.width = width;
    image.height = height;
    image.xoffset = 0;
    image.format = ZPixmap;
    image.data = (char *)data;
    image.byte_order = LSBFirst;
    image.bitmap_unit = bits_per_pixel;
    image.bitmap_bit_order = LSBFirst;
    image.bitmap_pad = scanline_pad;
    image.depth = vi.depth;
    image.bytes_per_line = width * bits_per_pixel / 8;
    image.bits_per_pixel = bits_per_pixel;
    image.red_mask = vi.red_mask;
    image.green_mask = vi.green_mask;
    image.blue_mask = vi.blue_mask;
    image.obdata = 0;

    if (XInitImage(&image) == False)
	errx(EXIT_FAILURE, "XInitImage failed");

    for (int y = 0; y < height; y++)
	for (int x = 0; x < width; x++) {
	    uint8_t red = icon_data->pixel_data[(x + y * width) * 4];
	    uint8_t green = icon_data->pixel_data[(x + y * width) * 4 + 1];
	    uint8_t blue = icon_data->pixel_data[(x + y * width) * 4 + 2];
	    uint8_t alpha = icon_data->pixel_data[(x + y * width) * 4 + 3];

	    if (disabled) {
		red = green = blue =
		    (red * 54 + green * 183 + blue * 19) / 256;
		alpha = alpha / 2;
	    }

	    red = (red_bg * 0xff + (red - red_bg) * alpha) / 0xff;
	    green = (green_bg * 0xff + (green - green_bg) * alpha) / 0xff;
	    blue = (blue_bg * 0xff + (blue - blue_bg) * alpha) / 0xff;

	    data[x + y * width] =
		((red << red_shift) & vi.red_mask) |
		((green << green_shift) & vi.green_mask) |
		((blue << blue_shift) & vi.blue_mask);
	}

    _pixmap = conn->screen().rootWindow().createPixmap(width, height, depth);
    X::GC gc = _pixmap.createGC(GCGraphicsExposures, &gc_defaults);
    _pixmap.putImage(gc, &image, 0, 0, 0, 0, width, height);
    gc.free();

    free(image.data);
}

IconPixmap::~IconPixmap()
{
    _pixmap.free();
}

Connection *IconPixmap::conn() const
{
    return _conn;
}

unsigned long IconPixmap::pixel() const
{
    return _pixel;
}

bool IconPixmap::disabled() const
{
    return _disabled;
}

X::Pixmap IconPixmap::pixmap() const
{
    return _pixmap;
}

/****************************************************************************/

class Toolkit::IconMask
{
    Connection *const _conn;
    X::Pixmap _pixmap;
public:
    IconMask(Connection *conn, struct icon_data const *icon_data);
    ~IconMask();
    Connection *conn() const;
    X::Pixmap pixmap() const;
};

IconMask::IconMask(Connection *conn, struct icon_data const *icon_data)
    : _conn(conn)
{
    int width = icon_data->width;
    int height = icon_data->height;

    int scanline_pad = 32;

    int bytes_per_line = ((width - 1) / scanline_pad + 1) * scanline_pad / 8;
    uint8_t *data = (uint8_t *)calloc(height, bytes_per_line);

    XImage image;
    image.width = width;
    image.height = height;
    image.xoffset = 0;
    image.format = XYBitmap;
    image.data = (char *)data;
    image.byte_order = LSBFirst;
    image.bitmap_unit = 8;
    image.bitmap_bit_order = LSBFirst;
    image.bitmap_pad = scanline_pad;
    image.depth = 1;
    image.bytes_per_line = bytes_per_line;
    image.bits_per_pixel = 1; /* ignored */
    image.red_mask = 0; /* ignored */
    image.green_mask = 0; /* ignored */
    image.blue_mask = 0; /* ignored */
    image.obdata = 0;

    if (XInitImage(&image) == False)
	errx(EXIT_FAILURE, "XInitImage failed");

    for (int y = 0; y < height; y++)
	for (int x = 0; x < width; x++) {
	    uint8_t alpha = icon_data->pixel_data[(x + y * width) * 4 + 3];
	    data[x / 8 + y * bytes_per_line] |= (alpha < 128) << (x % 8);
	}

    _pixmap = conn->screen().rootWindow().createPixmap(width, height, 1);
    X::GC gc = _pixmap.createGC(GCGraphicsExposures, &gc_defaults);
    _pixmap.putImage(gc, &image, 0, 0, 0, 0, width, height);
    gc.free();

    free(image.data);
}

IconMask::~IconMask()
{
    _pixmap.free();
}

Connection *IconMask::conn() const
{
    return _conn;
}

X::Pixmap IconMask::pixmap() const
{
    return _pixmap;
}

/****************************************************************************/

Icon::Icon(void const *icon_data) : icon_data(icon_data)
{
}

Icon::~Icon()
{
    for (std::vector<Toolkit::IconPixmap *>::const_iterator i =
	     _pixmaps.begin(); i != _pixmaps.end(); i++)
	delete *i;

    for (std::vector<Toolkit::IconMask *>::const_iterator i =
	     _masks.begin(); i != _masks.end(); i++)
	delete *i;
}

X::Pixmap Icon::pixmap(Connection &conn, unsigned long pixel, bool disabled)
{
    for (std::vector<Toolkit::IconPixmap *>::const_iterator i =
	     _pixmaps.begin(); i != _pixmaps.end(); i++)
	if ((*i)->conn() == &conn && (*i)->pixel() == pixel &&
	    (*i)->disabled() == disabled)
	    return (*i)->pixmap();

    IconPixmap *pixmap = new IconPixmap(&conn, pixel, disabled,
	static_cast<struct icon_data const *>(icon_data));
    _pixmaps.push_back(pixmap);
    return pixmap->pixmap();
}

X::Pixmap Icon::mask(Connection &conn)
{
    for (std::vector<Toolkit::IconMask *>::const_iterator i =
	     _masks.begin(); i != _masks.end(); i++)
	if ((*i)->conn() == &conn)
	    return (*i)->pixmap();

    IconMask *mask = new IconMask(&conn,
	static_cast<struct icon_data const *>(icon_data));
    _masks.push_back(mask);
    return mask->pixmap();
}

unsigned int Icon::width() const
{
    return static_cast<struct icon_data const *>(icon_data)->width;
}

unsigned int Icon::height() const
{
    return static_cast<struct icon_data const *>(icon_data)->height;
}
