/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "scrolling.h"
#include "connection.h"

#define SCROLLBAR_WIDTH 16
#define SCROLLBAR_RIPPLE_WIDTH 7
#define SCROLLBAR_MIN_TRACK_SIZE 8
#define SCROLLBAR_RIPPLE_MIN_TRACK_SIZE 18

using namespace Toolkit;


ScrollbarHandle::ScrollbarHandle(Scrollbar &scrollbar)
    : ButtonicWidget(scrollbar.conn),
      scrollbar(scrollbar), grab_x(0), grab_y(0), is_grabbed(false)
{
    SizeRequirements reqs;

    switch (scrollbar.type()) {
    case Scrollbar::Top:
    case Scrollbar::Bottom:
	reqs.minWidth = SCROLLBAR_MIN_TRACK_SIZE;
	reqs.minHeight = SCROLLBAR_WIDTH;
	reqs.maxHeight = SCROLLBAR_WIDTH;
	break;
    case Scrollbar::Left:
    case Scrollbar::Right:
	reqs.minWidth = SCROLLBAR_WIDTH;
	reqs.minHeight = SCROLLBAR_MIN_TRACK_SIZE;
	reqs.maxWidth = SCROLLBAR_WIDTH;
	break;
    }

    setSizeRequirements(reqs);
    setDecorationSize(SCROLLBAR_RIPPLE_WIDTH, SCROLLBAR_RIPPLE_WIDTH);
}

bool ScrollbarHandle::isGrabbed() const
{
    return is_grabbed;
}

void ScrollbarHandle::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (button != Button1)
	return;

    grab_x = x - this->x();
    grab_y = y - this->y();
    is_grabbed = true;
}

void ScrollbarHandle::buttonReleaseEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    if (!is_grabbed || button != Button1)
	return;

    is_grabbed = false;
    scrollbar.releaseHandle();
}

void ScrollbarHandle::motionEvent(Time time, int x, int y,
				  int x_root, int y_root,
				  unsigned int state, char is_hint)
{
    if (!is_grabbed)
	return;

    scrollbar.moveHandle(x - this->x() - grab_x,
			 y - this->y() - grab_y);
}

void ScrollbarHandle::paintDecoration(int xpos, int ypos, X::Pixmap pixmap)
{
    switch (scrollbar.type()) {
    case Scrollbar::Top:
    case Scrollbar::Bottom:
	if (width() < SCROLLBAR_RIPPLE_MIN_TRACK_SIZE)
	    break;
	for (int i = 0; i < 3; i++) {
	    gc().setForeground(colors().scrollbarRipple0.pixel);
	    pixmap.drawLine(gc(), xpos + i * 3, ypos,
			    xpos + i * 3, ypos + SCROLLBAR_RIPPLE_WIDTH);
	    gc().setForeground(colors().scrollbarRipple1.pixel);
	    pixmap.drawLine(gc(), xpos + i * 3 + 1, ypos,
			    xpos + i * 3 + 1, ypos + SCROLLBAR_RIPPLE_WIDTH);
	}
	break;
    case Scrollbar::Left:
    case Scrollbar::Right:
	if (height() < SCROLLBAR_RIPPLE_MIN_TRACK_SIZE)
	    break;
	for (int i = 0; i < 3; i++) {
	    gc().setForeground(colors().scrollbarRipple0.pixel);
	    pixmap.drawLine(gc(), xpos, ypos + i * 3,
			    xpos + SCROLLBAR_RIPPLE_WIDTH, ypos + i * 3);
	    gc().setForeground(colors().scrollbarRipple1.pixel);
	    pixmap.drawLine(gc(), xpos, ypos + i * 3 + 1,
			    xpos + SCROLLBAR_RIPPLE_WIDTH, ypos + i * 3 + 1);
	}
	break;
    }
}

ButtonicWidget::CornerRoundness ScrollbarHandle::cornerRoundness() const
{
    return CornerRoundness(1);
}

/****************************************************************************/

ScrollbarButton::ScrollbarButton(Scrollbar &scrollbar)
    : GenericButton(scrollbar.conn), scrollbar(scrollbar), _type(Left)
{
    setDecorationSize(7, 7);

    SizeRequirements reqs;
    reqs.minWidth = SCROLLBAR_WIDTH;
    reqs.minHeight = SCROLLBAR_WIDTH;
    reqs.maxWidth = SCROLLBAR_WIDTH;
    reqs.maxHeight = SCROLLBAR_WIDTH;
    setSizeRequirements(reqs);
}

ScrollbarButton::Type ScrollbarButton::type() const
{
    return _type;
}

void ScrollbarButton::setType(Type type)
{
    if (type == _type)
	return;

    _type = type;
    invalidate();
}

bool ScrollbarButton::insensitive() const
{
    return GenericButton::insensitive();
}

void ScrollbarButton::setInsensitive(bool insensitive)
{
    GenericButton::setInsensitive(insensitive);
}

void ScrollbarButton::paintDecoration(int xpos, int ypos, X::Pixmap pixmap)
{
    if (insensitive())
	gc().setForeground(colors().scrollbarInsensitiveArrow.pixel);
    else
	gc().setForeground(colors().scrollbarArrow.pixel);

    switch (_type) {
    case Left:
	pixmap.drawLine(gc(), xpos + 1, ypos + 3, xpos + 1, ypos + 3);
	pixmap.drawLine(gc(), xpos + 2, ypos + 2, xpos + 2, ypos + 4);
	pixmap.drawLine(gc(), xpos + 3, ypos + 1, xpos + 3, ypos + 5);
	pixmap.drawLine(gc(), xpos + 4, ypos,     xpos + 4, ypos + 6);
	break;
    case Right:
	pixmap.drawLine(gc(), xpos + 2, ypos,     xpos + 2, ypos + 6);
	pixmap.drawLine(gc(), xpos + 3, ypos + 1, xpos + 3, ypos + 5);
	pixmap.drawLine(gc(), xpos + 4, ypos + 2, xpos + 4, ypos + 4);
	pixmap.drawLine(gc(), xpos + 5, ypos + 3, xpos + 5, ypos + 3);
	break;
    case Up:
	pixmap.drawLine(gc(), xpos + 3, ypos + 1, xpos + 3, ypos + 1);
	pixmap.drawLine(gc(), xpos + 2, ypos + 2, xpos + 4, ypos + 2);
	pixmap.drawLine(gc(), xpos + 1, ypos + 3, xpos + 5, ypos + 3);
	pixmap.drawLine(gc(), xpos,     ypos + 4, xpos + 6, ypos + 4);
	break;
    case Down:
	pixmap.drawLine(gc(), xpos,     ypos + 2, xpos + 6, ypos + 2);
	pixmap.drawLine(gc(), xpos + 1, ypos + 3, xpos + 5, ypos + 3);
	pixmap.drawLine(gc(), xpos + 2, ypos + 4, xpos + 4, ypos + 4);
	pixmap.drawLine(gc(), xpos + 3, ypos + 5, xpos + 3, ypos + 5);
	break;
    }
}

ButtonicWidget::CornerRoundness ScrollbarButton::cornerRoundness() const
{
    switch (_type) {
    case Left:
	return CornerRoundness(2, 1, 2, 1);
    case Right:
	return CornerRoundness(1, 2, 1, 2);
    case Up:
	return CornerRoundness(2, 2, 1, 1);
    case Down:
	return CornerRoundness(1, 1, 2, 2);
    };

    return CornerRoundness(2);
}

void ScrollbarButton::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    GenericButton::buttonPressEvent(time, x, y, x_root, y_root, state, button);

    switch (_type) {
    case Left:
    case Up:
	scrollbar.setPosition(scrollbar.position() - scrollbar.lineSize());
	break;
    case Right:
    case Down:
	scrollbar.setPosition(scrollbar.position() + scrollbar.lineSize());
	break;
    }
}

/****************************************************************************/

Scrollbar::Scroller::Scroller(Scrollbar &scrollbar, WidgetControl &control)
    : WidgetControlScroller(control), scrollbar(scrollbar)
{
}

void Scrollbar::Scroller::updateRange()
{
    scrollbar.parameters_changed();
}

void Scrollbar::Scroller::updatePageSize()
{
    scrollbar.parameters_changed();
}

void Scrollbar::Scroller::updatePosition()
{
    scrollbar.parameters_changed();
}

/****************************************************************************/

Scrollbar::Scrollbar(Type type, WidgetControl &control, Connection &conn)
    : ContainerWidget(conn), scroller(*this, control), _type(type),
      dec_wc(*this), inc_wc(*this), handle_wc(*this),
      dec(*this), inc(*this), handle(*this)
{
    SizeRequirements reqs;

    switch (type) {
    case Top:
    case Bottom:
	dec.setType(ScrollbarButton::Left);
	inc.setType(ScrollbarButton::Right);
	reqs.minWidth = SCROLLBAR_WIDTH * 2 + SCROLLBAR_MIN_TRACK_SIZE * 2;
	reqs.minHeight = SCROLLBAR_WIDTH;
	reqs.maxHeight = SCROLLBAR_WIDTH;
	break;
    case Left:
    case Right:
	dec.setType(ScrollbarButton::Up);
	inc.setType(ScrollbarButton::Down);
	reqs.minWidth = SCROLLBAR_WIDTH;
	reqs.minHeight = SCROLLBAR_WIDTH * 2 + SCROLLBAR_MIN_TRACK_SIZE * 2;
	reqs.maxWidth = SCROLLBAR_WIDTH;
	break;
    };

    setSizeRequirements(reqs);

    dec.setInsensitive(true);
    inc.setInsensitive(true);

    dec_wc.setWidget(&dec);
    inc_wc.setWidget(&inc);
    handle_wc.setWidget(&handle);
}

Scrollbar::Type Scrollbar::type() const
{
    return _type;
}

WidgetScroller *Scrollbar::widget_scroller() const
{
    if (!scroller.control.widget())
	return 0;

    if (scroller.control.horzScroller() == &scroller)
	return scroller.control.widget()->horzScroller();

    if (scroller.control.vertScroller() == &scroller)
	return scroller.control.widget()->vertScroller();

    return 0;
}

int Scrollbar::range() const
{
    if (!widget_scroller())
	return 0;

    return widget_scroller()->range();
}

int Scrollbar::pageSize() const
{
    if (!widget_scroller())
	return 0;

    return widget_scroller()->pageSize();
}

int Scrollbar::position() const
{
    if (!widget_scroller())
	return 0;

    return widget_scroller()->position();
}

void Scrollbar::setPosition(int position)
{
    if (!widget_scroller())
	return;

    widget_scroller()->setPosition(position);
}

int Scrollbar::lineSize() const
{
    if (!widget_scroller())
	return 0;

    return widget_scroller()->lineSize();
}

void Scrollbar::updateLayout()
{
    switch (type()) {
    case Top:
    case Bottom:
	dec_wc.setDimensions(
	    Region(x(), y(), SCROLLBAR_WIDTH, SCROLLBAR_WIDTH));
	inc_wc.setDimensions(
	    Region(x() + width() - SCROLLBAR_WIDTH, y(),
		   SCROLLBAR_WIDTH, SCROLLBAR_WIDTH));
	break;
    case Left:
    case Right:
	dec_wc.setDimensions(
	    Region(x(), y(), SCROLLBAR_WIDTH, SCROLLBAR_WIDTH));
	inc_wc.setDimensions(
	    Region(x(), y() + height() - SCROLLBAR_WIDTH,
		   SCROLLBAR_WIDTH, SCROLLBAR_WIDTH));
	break;
    };

    update_handle_dimensions();
}

void Scrollbar::paint1(int t0, int t1, const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    switch (type()) {
    case Top:
	gc().setForeground(colors().scrollbar.pixel);
	pixmap.fillRectangle(gc(), t0, y, t1 - t0, height() - 1);
	gc().setForeground(colors().scrollbarBorder.pixel);
	pixmap.drawLine(gc(), t0, y + height() - 1, t1 - 1, y + height() - 1);
	break;
    case Left:
	gc().setForeground(colors().scrollbar.pixel);
	pixmap.fillRectangle(gc(), x, t0, width() - 1, t1 - t0);
	gc().setForeground(colors().scrollbarBorder.pixel);
	pixmap.drawLine(gc(), x + width() - 1, t0, x + width() - 1, t1 - 1);
	break;
    case Bottom:
	gc().setForeground(colors().scrollbar.pixel);
	pixmap.fillRectangle(gc(), t0, y + 1, t1 - t0, height() - 1);
	gc().setForeground(colors().scrollbarBorder.pixel);
	pixmap.drawLine(gc(), t0, y, t1 - 1, y);
	break;
    case Right:
	gc().setForeground(colors().scrollbar.pixel);
	pixmap.fillRectangle(gc(), x + 1, t0, width() - 1, t1 - t0);
	gc().setForeground(colors().scrollbarBorder.pixel);
	pixmap.drawLine(gc(), x, t0, x, t1 - 1);
	break;
    };
}

void Scrollbar::paint(const Region &region, X::Pixmap pixmap)
{
    ContainerWidget::paint(region, pixmap);

    int x = this->x() - region.x;
    int y = this->y() - region.y;

    switch (type()) {
    case Top:
    case Bottom:
	if (range() == 0 || pageSize() == range()) {
	    paint1(x + SCROLLBAR_WIDTH, x + width() - SCROLLBAR_WIDTH,
		   region, pixmap);
	    break;
	}
	paint1(x + SCROLLBAR_WIDTH, handle_wc.dimensions().x - region.x,
	       region, pixmap);
	paint1(handle_wc.dimensions().x - region.x +
	       handle_wc.dimensions().width, x + width() - SCROLLBAR_WIDTH,
	       region, pixmap);
	break;
    case Left:
    case Right:
	if (range() == 0 || pageSize() == range()) {
	    paint1(y + SCROLLBAR_WIDTH, y + height() - SCROLLBAR_WIDTH,
		   region, pixmap);
	    break;
	}
	paint1(y + SCROLLBAR_WIDTH, handle_wc.dimensions().y - region.y,
	       region, pixmap);
	paint1(handle_wc.dimensions().y - region.y +
	       handle_wc.dimensions().height, y + height() - SCROLLBAR_WIDTH,
	       region, pixmap);
	break;
    };
}

void Scrollbar::buttonPressEvent(
    Time time, int x, int y, int x_root, int y_root,
    unsigned int state, unsigned int button)
{
    bool increase;

    if (type() == Top || type() == Bottom)
	increase = x > handle_wc.dimensions().x;
    else
	increase = y > handle_wc.dimensions().y;

    if (increase)
	setPosition(position() + pageSize());
    else
	setPosition(position() - pageSize());
}

void Scrollbar::moveHandle(int delta_x, int delta_y)
{
    int min, max, pixels0, pixels1;

    if (type() == Top || type() == Bottom) {
	min = x() + SCROLLBAR_WIDTH - handle_wc.dimensions().x;
	max = x() + width() - SCROLLBAR_WIDTH - handle_wc.dimensions().x -
	    handle_size();
	if (delta_x < min)
	    delta_x = min;
	if (delta_x > max)
	    delta_x = max;
	handle_wc.setDimensions(
	    Region(handle_wc.dimensions().x + delta_x, y(),
		   handle_size(),
		   SCROLLBAR_WIDTH));

	pixels0 = width() - 2 * SCROLLBAR_WIDTH - handle_size();
	pixels1 = handle_wc.dimensions().x + delta_x - x() - SCROLLBAR_WIDTH;
    } else {
	min = y() + SCROLLBAR_WIDTH - handle_wc.dimensions().y;
	max = y() + height() - SCROLLBAR_WIDTH - handle_wc.dimensions().y -
	    handle_size();
	if (delta_y < min)
	    delta_y = min;
	if (delta_y > max)
	    delta_y = max;
	handle_wc.setDimensions(
	    Region(x(), handle_wc.dimensions().y + delta_y,
		   SCROLLBAR_WIDTH,
		   handle_size()));

	pixels0 = height() - 2 * SCROLLBAR_WIDTH - handle_size();
	pixels1 = handle_wc.dimensions().y + delta_y - y() - SCROLLBAR_WIDTH;
    }

    invalidate();
    if (pixels0 == 0)
	return;

    setPosition((pixels1 * (range() - pageSize()) + pixels0 / 2) / pixels0);
}

void Scrollbar::releaseHandle()
{
    update_handle_dimensions();
    invalidate();
}

unsigned int Scrollbar::handle_size() const
{
    int size = 0;

    switch (type()) {
    case Top:
    case Bottom:
	if (range() == 0)
	    return width() - 2 * SCROLLBAR_WIDTH;

	size = (width() - 2 * SCROLLBAR_WIDTH - 1) * pageSize() / range() + 1;
	break;
    case Left:
    case Right:
	if (range() == 0)
	    return height() - 2 * SCROLLBAR_WIDTH;

	size = (height() - 2 * SCROLLBAR_WIDTH - 1) * pageSize() / range() + 1;
	break;
    }

    if (size < SCROLLBAR_MIN_TRACK_SIZE)
	return SCROLLBAR_MIN_TRACK_SIZE;
    return size;
}

void Scrollbar::parameters_changed()
{
    dec.setInsensitive(position() == 0);
    inc.setInsensitive(position() + pageSize() == range());
    update_handle_dimensions();
    invalidate();
}

void Scrollbar::update_handle_dimensions()
{
    if (handle.isGrabbed())
	return;

    switch (type()) {
    case Top:
    case Bottom:
	handle_wc.setDimensions(
	    Region(x() + SCROLLBAR_WIDTH +
		   (width() - 2 * SCROLLBAR_WIDTH - handle_size())
		       * position()
		       / (range() - pageSize() ? range() - pageSize() : 1),
		   y(),
		   handle_size(),
		   SCROLLBAR_WIDTH));
	break;
    case Left:
    case Right:
	handle_wc.setDimensions(
	    Region(x(),
		   y() + SCROLLBAR_WIDTH +
		   (height() - 2 * SCROLLBAR_WIDTH - handle_size())
		       * position()
		       / (range() - pageSize() ? range() - pageSize() : 1),
		   SCROLLBAR_WIDTH,
		   handle_size()));
	break;
    };
}

/****************************************************************************/

ScrollPane::MainControl::MainControl(ScrollPane &pane)
    : Control(pane), pane(pane)
{
}

void ScrollPane::MainControl::widgetChanged()
{
    pane.horz_wc.setEnabled(widget() && widget()->horzScroller());
    pane.vert_wc.setEnabled(widget() && widget()->vertScroller());
    pane.nav_wc.setEnabled(widget() && widget()->horzScroller()
				    && widget()->vertScroller());
    pane.sizeRequirementsChanged(this);
    pane.requestLayout();
}

ScrollPane::ScrollPane(Connection &conn)
    : ContainerWidget(conn),
      main_wc(*this), horz_wc(*this), vert_wc(*this), nav_wc(*this),
      horz_sb(Scrollbar::Bottom, main_wc, conn),
      vert_sb(Scrollbar::Right, main_wc, conn), _background(-1)
{
    main_wc.setHorzScroller(&horz_sb.scroller);
    main_wc.setVertScroller(&vert_sb.scroller);
    horz_wc.setWidget(&horz_sb);
    vert_wc.setWidget(&vert_sb);
}

WidgetControl *ScrollPane::mainControl()
{
    return &main_wc;
}

WidgetControl *ScrollPane::navigatorControl()
{
    return &nav_wc;
}

unsigned long ScrollPane::background() const
{
    return _background;
}

void ScrollPane::setBackground(unsigned long background)
{
    _background = background;
}

void ScrollPane::updateLayout()
{
    unsigned int w = width(), h = height();

    if (main_wc.widget() && main_wc.widget()->vertScroller())
	w -= SCROLLBAR_WIDTH;
    if (main_wc.widget() && main_wc.widget()->horzScroller())
	h -= SCROLLBAR_WIDTH;

    main_wc.setDimensions(Region(x(), y(), w, h));
    horz_wc.setDimensions(Region(x(), y() + height() - SCROLLBAR_WIDTH,
				 w, SCROLLBAR_WIDTH));
    vert_wc.setDimensions(Region(x() + width() - SCROLLBAR_WIDTH, y(),
				 SCROLLBAR_WIDTH, h));
    nav_wc.setDimensions(Region(x() + width() - SCROLLBAR_WIDTH,
				y() + height() - SCROLLBAR_WIDTH,
				SCROLLBAR_WIDTH, SCROLLBAR_WIDTH));
}

void ScrollPane::sizeRequirementsChanged(WidgetControl *control)
{
    SizeRequirements reqs(main_wc.sizeRequirements());
    SizeRequirements horz_reqs(horz_wc.sizeRequirements());
    SizeRequirements vert_reqs(vert_wc.sizeRequirements());

    if (horz_wc.isEnabled()) {
	if (reqs.minWidth < horz_reqs.minWidth)
	    reqs.minWidth = horz_reqs.minWidth;
    }
    if (vert_wc.isEnabled()) {
	if (reqs.minHeight < vert_reqs.minHeight)
	    reqs.minHeight = vert_reqs.minHeight;
    }

    if (horz_wc.isEnabled()) {
	reqs.minHeight += horz_reqs.minHeight;
	if (reqs.maxHeight != -1)
	    reqs.maxHeight += horz_reqs.minHeight;
    }
    if (vert_wc.isEnabled()) {
	reqs.minWidth += vert_reqs.minWidth;
	if (reqs.maxWidth != -1)
	    reqs.maxWidth += vert_reqs.minWidth;
    }

    setSizeRequirements(reqs);
}

void ScrollPane::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    if (background() != (unsigned long)-1) {
	gc().setForeground(background());
	pixmap.fillRectangle(gc(), x, y, width(), height());
    }

    ContainerWidget::paint(region, pixmap);
}
