/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "statusbar.h"
#include "timer.h"
#include "connection.h"
#include "icon.h"

#define STATUSBAR_PADDING_X 8
#define STATUSBAR_PADDING_Y 4
#define STATUSBAR_ICON_PADDING_LEFT 0
#define STATUSBAR_ICON_PADDING_RIGHT 4
#define STATUSBAR_CORRECTION_Y -2

using namespace Toolkit;


class Toolkit::StatusBarTimer : public Timer
{
    StatusBar &status_bar;
public:
    StatusBarTimer(StatusBar &statusBar) : status_bar(statusBar) {
    }
    void timeout() {
	status_bar.setMessage(std::string());
    }
};

StatusBar::StatusBar(Connection &conn)
    : Widget(conn), icon(0), _num(0), _denom(0),
      timer(new StatusBarTimer(*this))
{
    SizeRequirements reqs;
    reqs.minWidth = STATUSBAR_PADDING_X * 2;
    reqs.minHeight = reqs.maxHeight =
	conn.fontSet().ascent() +
	conn.fontSet().descent() + STATUSBAR_PADDING_Y * 2;
    setSizeRequirements(reqs);
}

StatusBar::~StatusBar()
{
    delete timer;
}

std::string StatusBar::message() const
{
    return _message;
}

void StatusBar::setMessage(const std::string &message, Icon *icon)
{
    if (timer->isActive() || (message == _message && icon == this->icon))
	return;

    _message = message;
    this->icon = icon;
    if (_denom == 0)
	invalidate();
}

void StatusBar::setTimedMessage(long sec, long usec,
				const std::string &message, Icon *icon)
{
    timer->set(sec, usec, false);
    if (message == _message && icon == this->icon)
	return;

    _message = message;
    this->icon = icon;
    if (_denom == 0)
	invalidate();
}

void StatusBar::setProgress(int num, int denom)
{
    if (num == _num && denom == _denom)
	return;

    _num = num;
    _denom = denom;
    invalidate();
}

void StatusBar::setIncompleteHotkey(const Hotkey &hotkey)
{
    if (hotkey == incomplete_hotkey)
	return;

    incomplete_hotkey = hotkey;
    invalidate();
}

void StatusBar::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    if (_denom == 0) {
	Icon *icon = 0;
	std::string message;

	if (!incomplete_hotkey.isNull())
	    message = incomplete_hotkey.format() + " ...";
	else {
	    icon = this->icon;
	    message = this->message();
	}

	gc().setForeground(colors().statusbar.pixel);
	pixmap.fillRectangle(gc(), x, y, width(), height());

	if (icon)
	    X::Drawable::copyArea(
		icon->pixmap(conn, colors().statusbar.pixel, false),
		pixmap, gc(), 0, 0, icon->width(), icon->height(),
		x + STATUSBAR_PADDING_X + STATUSBAR_ICON_PADDING_LEFT,
		y + STATUSBAR_CORRECTION_Y + (height() - icon->height()) / 2);

	gc().setForeground(colors().statusbarText.pixel);
	pixmap.mbDrawString(conn.fontSet(), gc(),
			    x + STATUSBAR_PADDING_X +
				(icon ? icon->width()
				    + STATUSBAR_ICON_PADDING_LEFT
				    + STATUSBAR_ICON_PADDING_RIGHT : 0),
			    y + STATUSBAR_CORRECTION_Y + STATUSBAR_PADDING_Y +
				conn.fontSet().ascent(), message);
    } else {
	int pbar = width() * _num / _denom;
	gc().setForeground(colors().statusbarProgress.pixel);
	pixmap.fillRectangle(gc(), x, y, pbar, height());
	gc().setForeground(colors().statusbar.pixel);
	pixmap.fillRectangle(gc(), x + pbar, y, width() - pbar, height());
    }
}
