/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "tabbar.h"
#include "model_view.h"
#include "connection.h"
#include "icon.h"
#include "menus.h"

#define TABBAR_TAB_PADDING_X 5
#define TABBAR_TAB_PADDING_Y 3
#define TABBAR_TAB_ICON_PADDING 5

using namespace Toolkit;


class Toolkit::TabBarShadow : public Shadow
{
    TabBar &tab_bar;
    Shadow *buildShadow(Item *item) const {
	return new TabBarShadow(
	    item, const_cast<TabBarShadow *>(this), tab_bar);
    }
public:
    TabBarShadow(ItemInterior *interior, TabBar &tabBar)
	: Shadow(interior), tab_bar(tabBar) {
    }
    TabBarShadow(Item *item, Shadow *parent, TabBar &tabBar)
	: Shadow(item, parent), tab_bar(tabBar) {
    }
    void updateItem() {
	tab_bar.invalidate();
    }
    void updateItems(unsigned int index, unsigned int old_count,
					 unsigned int new_count) {
	if (parent())
	    return;
	if (tab_bar.selectedTab() == 0 && count() != 0)
	    tab_bar.selectTab(shadow(0)->item());
	else if (tab_bar.selectedTabIndex() == NPOS) {
	    if (index < count())
		tab_bar.selectTab(shadow(index)->item());
	    else if (count() != 0)
		tab_bar.selectTab(shadow(count() - 1)->item());
	    else
		tab_bar.selectTab(0);
	}
	tab_bar.invalidate();
    }
};

/****************************************************************************/

TabBar::TabBar(Connection &conn)
    : Widget(conn), _interior(0), main_shadow(0), selected_tab(0)
{
    SizeRequirements reqs;
    reqs.minWidth = TABBAR_TAB_PADDING_X * 2;
    reqs.minHeight = conn.fontSet().ascent() +
		     conn.fontSet().descent() + TABBAR_TAB_PADDING_Y * 2;
    reqs.maxHeight = reqs.minHeight;
    setSizeRequirements(reqs);
}

TabBar::~TabBar()
{
    delete main_shadow;
}

ItemInterior *TabBar::interior() const
{
    return _interior;
}

void TabBar::setInterior(ItemInterior *interior)
{
    if (interior == _interior)
	return;

    delete main_shadow;
    _interior = interior;
    main_shadow = interior ? new TabBarShadow(interior, *this) : 0;
}

Item *TabBar::selectedTab() const
{
    return selected_tab;
}

void TabBar::selectTab(Item *tab)
{
    if (tab == selected_tab || !interior() ||
	interior()->cachedIndexOf(tab) == NPOS)
	return;

    selected_tab = tab;
    invalidate();
    tabSelected();
}

unsigned int TabBar::selectedTabIndex() const
{
    return interior()->cachedIndexOf(selected_tab);
}

void TabBar::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    int tab_x = 0;

    if (!main_shadow)
	return;

    for (unsigned int i = 0; i < main_shadow->count(); i++) {
	int tab_width = (width() - tab_x) / (main_shadow->count() - i);
	Shadow *shadow = main_shadow->shadow(i);

	unsigned long fg, bg;

	if (shadow->item() == selected_tab) {
	    fg = colors().tabbarSelectedText.pixel;
	    bg = colors().tabbarSelected.pixel;
	} else {
	    fg = colors().tabbarUnselectedText.pixel;
	    bg = colors().tabbarUnselected.pixel;
	}

	gc().setForeground(bg);
	pixmap.fillRectangle(gc(), x + tab_x, y, tab_width, height());

	gc().setForeground(colors().tabbarPlasma.pixel);
	pixmap.drawPoint(gc(), x + tab_x, y);
	pixmap.drawPoint(gc(), x + tab_x + tab_width - 1, y);

	Icon *icon = shadow->item()->icon();
	int icon_width = icon ? icon->width() + TABBAR_TAB_ICON_PADDING : 0;

	gc().setForeground(fg);
	pixmap.wcDrawString(conn.fontSet(), gc(),
			    x + tab_x + TABBAR_TAB_PADDING_X + icon_width,
			    y + TABBAR_TAB_PADDING_Y +
				conn.fontSet().ascent(),
			    conn.ellipsis(conn.fontSet(),
					  shadow->item()->caption(),
					  tab_width - icon_width -
					      TABBAR_TAB_PADDING_X * 2));

	if (icon)
	    X::Drawable::copyArea(icon->pixmap(conn, bg), pixmap, gc(), 0, 0,
				  icon->width(), icon->height(),
				  x + tab_x + TABBAR_TAB_PADDING_X,
				  y + (height() - icon->height()) / 2);

	tab_x += tab_width;
    }
}

void TabBar::buttonPressEvent(Time time, int x, int y,
			      int x_root, int y_root,
			      unsigned int state, unsigned int button)
{
    unsigned int i;
    int tab_x = 0;

    if (!main_shadow || main_shadow->count() == 0)
	return;

    for (i = 0; i < main_shadow->count(); i++) {
	int tab_width = (width() - tab_x) / (main_shadow->count() - i);

	if (x >= this->x() + tab_x && x < this->x() + tab_x + tab_width)
	    break;

	tab_x += tab_width;
    }

    switch (button) {
    case Button1:
	selectTab(interior()->cachedItem(i));
	break;
    case Button3:
	Menu *menu = invokeContextMenu(main_shadow->shadow(i));
	if (menu) {
	    conn.enterMenuMode(time, menu);
	    menu->show(time, x_root, y_root, hook());
	}
	break;
    }
}
