/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <assert.h>
#include "historyview.h"
#include "../connection.h"
#include "../icon.h"

#define HISTORYVIEW_PADDING_X 4
#define HISTORYVIEW_PADDING_Y 4

#define HISTORYVIEW_ITEM_PADDING_X0 8
#define HISTORYVIEW_ITEM_PADDING_X1 8
#define HISTORYVIEW_ITEM_PADDING_X2 4
#define HISTORYVIEW_ITEM_PADDING_X3 4
#define HISTORYVIEW_ITEM_PADDING_Y 2
#define HISTORYVIEW_INDENT 15

using namespace Toolkit;


HistoryViewShadow::HistoryViewShadow(ItemInterior *interior, HistoryView &view)
    : Shadow(interior), view(view),
      latest_child(0), anchor_x(0), anchor_y(0),
      visible_children_count(0), visible_children_height(0)
{
}

HistoryViewShadow::HistoryViewShadow(Item *item, Shadow *parent,
				     HistoryView &view)
    : Shadow(item, parent), view(view),
      latest_child(0), anchor_x(0), anchor_y(0),
      visible_children_count(0), visible_children_height(0)
{
}

Shadow *HistoryViewShadow::buildShadow(Item *item) const
{
    return new HistoryViewShadow(
	item, const_cast<HistoryViewShadow *>(this), view);
}

HistoryViewShadow *HistoryViewShadow::latestChild() const
{
    return latest_child;
}

void HistoryViewShadow::setLatestChild(HistoryViewShadow *latestChild)
{
    if (latestChild == latest_child)
	return;

    latest_child = latestChild;
    assert(indexOf(latest_child) != NPOS);
    view.requestLayout();
}

void HistoryViewShadow::updateItem()
{
    view.invalidate();
}

void HistoryViewShadow::updateItems(unsigned int index, unsigned int old_count,
							unsigned int new_count)
{
    if (latest_child && indexOf(latest_child) == NPOS)
	latest_child = 0;
    if (!latest_child && count() != 0)
	setLatestChild(shadow(count() - 1));

    view.requestLayout();
}

HistoryViewShadow *HistoryViewShadow::parent() const
{
    return static_cast<HistoryViewShadow *>(Shadow::parent());
}

HistoryViewShadow *HistoryViewShadow::shadow(unsigned int index) const
{
    return static_cast<HistoryViewShadow *>(Shadow::shadow(index));
}

HistoryViewShadow *HistoryViewShadow::find(int x, int y)
{
    if (sensitive_region.contains(x, y))
	return this;

    for (unsigned int i = 0; i < count(); i++) {
	HistoryViewShadow *res = shadow(i)->find(x, y);
	if (res)
	    return res;
    }

    return 0;
}

HistoryViewShadow *HistoryViewShadow::nth(unsigned int index) const
{
    for (unsigned int i = 0; i < count(); i++) {
	HistoryViewShadow *sh = shadow(i);
	if (sh == latest_child)
	    continue;

	if (index == 0)
	    return sh;
	index--;

	if (index < sh->visibleChildrenCount())
	    return sh->nth(index);
	index -= sh->visibleChildrenCount();
    }

    if (index == 0)
	return latest_child;
    index--;

    assert(index < latest_child->visibleChildrenCount());
    return latest_child->nth(index);
}

unsigned int HistoryViewShadow::anchorX() const
{
    return anchor_x;
}

unsigned int HistoryViewShadow::anchorY() const
{
    return anchor_y;
}

unsigned int HistoryViewShadow::height() const
{
    return _height;
}

unsigned int HistoryViewShadow::visibleChildrenCount() const
{
    return visible_children_count;
}

unsigned int HistoryViewShadow::visibleChildrenHeight() const
{
    return visible_children_height;
}

void HistoryViewShadow::relayout(unsigned int x, unsigned int y)
{
    anchor_x = x;
    anchor_y = y;

    if (!parent()) {
	sensitive_region = Region();
	_indent = 0;
	_height = 0;
    } else {
	_indent = HISTORYVIEW_INDENT;
	_height = view.conn.fontSet().ascent() +
		  view.conn.fontSet().descent() +
		  HISTORYVIEW_ITEM_PADDING_Y * 2;

	sensitive_region =
	    Region(anchorX() + HISTORYVIEW_ITEM_PADDING_X0
			     + HISTORYVIEW_ITEM_PADDING_X1,
		   -anchorY() - _height,
		   view.width() - anchorX() - HISTORYVIEW_PADDING_X * 2
					    - HISTORYVIEW_ITEM_PADDING_X0
					    - HISTORYVIEW_ITEM_PADDING_X1,
		   _height);
    }
    visible_children_count = 0;
    visible_children_height = 0;

    for (unsigned int i = 0; i < count(); i++) {
	HistoryViewShadow *sh = shadow(i);
	if (sh == latest_child)
	    continue;
	sh->relayout(anchor_x + _indent,
		     anchor_y + _height + visible_children_height);
	visible_children_count += sh->visibleChildrenCount() + 1;
	visible_children_height += sh->height() +
				   sh->visibleChildrenHeight();
    }

    if (latest_child) {
	latest_child->relayout(anchor_x,
			       anchor_y + _height + visible_children_height);
	visible_children_count += latest_child->visibleChildrenCount() + 1;
	visible_children_height += latest_child->height() +
				   latest_child->visibleChildrenHeight();
    }
}

void HistoryViewShadow::paintChildren(int origin_x, int origin_y,
				      X::Pixmap pixmap)
{
    for (unsigned int i = 0; i < count(); i++) {
	shadow(i)->paint0(origin_x, origin_y, pixmap);
	shadow(i)->paintChildren(origin_x, origin_y, pixmap);
	shadow(i)->paint1(origin_x, origin_y, pixmap);
    }
}

void HistoryViewShadow::paint0(int origin_x, int origin_y, X::Pixmap pixmap)
{
    Region region = sensitive_region;

    unsigned long background;

    if (view.isSelected(this)) {
	background = view.colors().viewSelected.pixel;
	view.gc().setForeground(background);
	pixmap.fillRectangle(view.gc(),
			     origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0
						 + HISTORYVIEW_ITEM_PADDING_X1,
			     origin_y - anchor_y - _height,
			     region.width,
			     _height);
	view.gc().setForeground(view.colors().viewSelectedText.pixel);
    } else {
	background = view.colors().viewBackground.pixel;
	view.gc().setForeground(background);
	pixmap.fillRectangle(view.gc(),
			     origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0
						 + HISTORYVIEW_ITEM_PADDING_X1,
			     origin_y - anchor_y - _height,
			     region.width,
			     _height);
	view.gc().setForeground(view.colors().viewItemText.pixel);
    }

    Icon *icon = item()->icon();

    X::Drawable::copyArea(icon->pixmap(view.conn, background),
			  pixmap, view.gc(),
			  0, 0, icon->width(), icon->height(),
			  origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0
					      + HISTORYVIEW_ITEM_PADDING_X1
					      + HISTORYVIEW_ITEM_PADDING_X2,
			  origin_y - anchor_y - _height +
			      (_height - icon->height()) / 2);

    pixmap.wcDrawString(item()->isBold() ? view.conn.fontSetBold()
					 : view.conn.fontSet(), view.gc(),
			origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0 +
					      HISTORYVIEW_ITEM_PADDING_X1 +
					      HISTORYVIEW_ITEM_PADDING_X2 +
			      icon->width() + HISTORYVIEW_ITEM_PADDING_X3,
			origin_y - anchor_y - _height +
			    (_height + view.conn.fontSet().ascent()
				     - view.conn.fontSet().descent()) / 2,
			item()->caption());

    XGCValues values;

    if (parent()->parent()) {
	values.foreground = view.colors().viewLines.pixel;
	values.line_width = 2;
	view.gc().change(GCForeground | GCLineWidth, &values);

	if (this != parent()->latestChild())
	    pixmap.drawLine(
		view.gc(),
		origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0
				    - HISTORYVIEW_INDENT,
		origin_y - anchor_y - _height + _height / 2
				    + HISTORYVIEW_INDENT * 2 / 3,
		origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0,
		origin_y - anchor_y - _height + _height / 2);
	else
	    pixmap.drawLine(view.gc(),
		origin_x + parent()->anchorX() + HISTORYVIEW_ITEM_PADDING_X0,
		origin_y - parent()->anchorY() - parent()->height() +
		    parent()->height() / 2,
		origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0,
		origin_y - anchor_y - _height + _height / 2);

	values.line_width = 0;
	view.gc().change(GCLineWidth, &values);
    }

/*  values.foreground = view.colors().viewLines.pixel;
    values.line_width = 0;
    values.line_style = LineOnOffDash;
    values.dash_offset = 1;
    values.dashes = 1;
    view.gc().change(GCForeground | GCLineWidth | GCLineStyle |
		     GCDashOffset | GCDashList, &values);

    pixmap.drawLine(view.gc(),
		    origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0,
		    origin_y - anchor_y - _height + _height / 2,
		    origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0
					+ HISTORYVIEW_ITEM_PADDING_X1 - 1,
		    origin_y - anchor_y - _height + _height / 2);

    values.line_style = LineSolid;
    view.gc().change(GCLineStyle, &values); */
}

void HistoryViewShadow::paint1(int origin_x, int origin_y, X::Pixmap pixmap)
{
    X::GC gc(view.gc());

    gc.setForeground(view.colors().viewBackground.pixel);
    pixmap.fillArc(gc,
		   origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0 - 4,
		   origin_y - anchor_y - _height / 2 - 4,
		   7, 7, 0, 0x5a00);
    gc.setForeground(view.colors().viewSelectedText.pixel);
    pixmap.drawArc(gc,
		   origin_x + anchor_x + HISTORYVIEW_ITEM_PADDING_X0 - 4,
		   origin_y - anchor_y - _height / 2 - 4,
		   7, 7, 0, 0x5a00);
}

/****************************************************************************/

class HistoryView::Scroller : public WidgetScroller
{
    HistoryView &view;
public:
    Scroller(HistoryView &view) : WidgetScroller(view), view(view) {
    }
    int lineSize() const {
	return 1;
    }
    void setRange(int range) {
	WidgetScroller::setRange(range);
    }
    void setPageSize(int pageSize) {
	WidgetScroller::setPageSize(pageSize);
    }
private:
    void updatePosition() {
	view.scroll_position_changed();
	view.invalidate();
    }
};

/****************************************************************************/

HistoryView::HistoryView(Connection &conn)
    : Widget(conn, 0, vert_scroller = new Scroller(*this)),
      root_interior(0), root_shadow(0), latest_shadow(0),
      origin_x(0), origin_y(0)
{
    SizeRequirements reqs;
    reqs.minWidth = HISTORYVIEW_PADDING_X * 2;
    reqs.minHeight = HISTORYVIEW_PADDING_Y * 2;
    setSizeRequirements(reqs);

    setFocussables(std::vector<Focussable *>(1, this));
}

HistoryView::~HistoryView()
{
    delete root_shadow;
    delete vert_scroller;
}

ItemInterior *HistoryView::rootInterior() const
{
    return root_interior;
}

void HistoryView::setRootInterior(ItemInterior *interior)
{
    if (interior == root_interior)
	return;

    root_interior = interior;

    delete root_shadow;
    root_shadow = interior ? new HistoryViewShadow(interior, *this) : 0;

    requestLayout();
}

HistoryViewShadow *HistoryView::rootShadow() const
{
    return root_shadow;
}

HistoryViewShadow *HistoryView::latestShadow() const
{
    return latest_shadow;
}

void HistoryView::setLatestShadow(HistoryViewShadow *latestShadow)
{
    if (latestShadow == latest_shadow)
	return;

    latest_shadow = latestShadow;
    for (HistoryViewShadow *s = latest_shadow;
	 s && s->parent(); s = s->parent())
	s->parent()->setLatestChild(s);
}

void HistoryView::scrollTo(HistoryViewShadow *shadow)
{
    int index = -1;
    HistoryViewShadow *parent = shadow->parent();

    do {
	unsigned int i = shadow == parent->latestChild()
	    ? parent->count() : parent->indexOf(shadow);
	while (i-- > 0) {
	    HistoryViewShadow *sh = parent->shadow(i);
	    if (sh != parent->latestChild())
		index += sh->visibleChildrenCount() + 1;
	}
	index += 1;
	shadow = parent;
	parent = shadow->parent();
    } while (parent);

    index = root_shadow->visibleChildrenCount() - index - 1;
    if (index >= vertScroller()->position() + vertScroller()->pageSize())
	vertScroller()->setPosition(index - vertScroller()->pageSize() + 1);
    if (vertScroller()->position() > index)
	vertScroller()->setPosition(index);
}

void HistoryView::updateLayout()
{
    /* structure or dimensions have changed */

    if (root_shadow)
	root_shadow->relayout(0, 0);

    if (!root_shadow || root_shadow->visibleChildrenCount() == 0)
	vert_scroller->setRange(0);
    else {
	vert_scroller->setRange(root_shadow->visibleChildrenCount());
	vert_scroller->setPageSize((height() - HISTORYVIEW_PADDING_Y * 2)
				   * root_shadow->visibleChildrenCount()
				   / root_shadow->visibleChildrenHeight());
    }

    scroll_position_changed();
}

void HistoryView::scroll_position_changed()
{
    origin_x = HISTORYVIEW_PADDING_X;
    origin_y = HISTORYVIEW_PADDING_Y;

    if (root_shadow && root_shadow->visibleChildrenCount() != 0) {
	HistoryViewShadow *first =
	    root_shadow->nth(root_shadow->visibleChildrenCount()
				 - vertScroller()->position() - 1);
	origin_y += first->anchorY() + first->height();
    }
}

void HistoryView::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().viewBackground.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    if (!root_shadow || root_shadow->count() == 0)
	return;

    XRectangle rectangle;
    rectangle.x = x;
    rectangle.y = y;
    rectangle.width = width();
    rectangle.height = height();

    assert(root_shadow->visibleChildrenCount() != 0);
    HistoryViewShadow *first =
	root_shadow->nth(root_shadow->visibleChildrenCount()
			     - vertScroller()->position() - 1);

    gc().setClipRectangles(0, 0, &rectangle, 1, Unsorted);
    root_shadow->paintChildren(
	x + HISTORYVIEW_PADDING_X,
	y + HISTORYVIEW_PADDING_Y + first->anchorY() + first->height(),
	pixmap);
    gc().setClipMask(None);
}

void HistoryView::buttonPressEvent(Time time, int x, int y,
				   int x_root, int y_root,
				   unsigned int state, unsigned int button)
{
    switch (button) {
    case Button4:
	vertScroller()->setPosition(
	    vertScroller()->position() -
	    vertScroller()->pageSize() * 120 / height());
	break;
    case Button5:
	vertScroller()->setPosition(
	    vertScroller()->position() +
	    vertScroller()->pageSize() * 120 / height());
	break;
    default:
	if (control())
	    control()->focus(this);

	HistoryViewShadow *shadow =
	    root_shadow ? root_shadow->find(x - origin_x - this->x(),
					    y - origin_y - this->y()) : 0;
	if (shadow)
	    setSelected(shadow);
    }
}
