/* Copyright (C) 2013 Roland Lutz

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "view.h"
#include "../connection.h"

using namespace Toolkit;


ViewShadow::ViewShadow(ItemInterior *interior, View &view)
    : Shadow(interior), view(view), is_expanded(true),
      anchor_x(0), anchor_y(0),
      visible_children_count(0), visible_children_height(0)
{
}

ViewShadow::ViewShadow(Item *item, Shadow *parent, View &view)
    : Shadow(item, parent), view(view), is_expanded(!parent->parent()),
      anchor_x(0), anchor_y(0),
      visible_children_count(0), visible_children_height(0)
{
}

Shadow *ViewShadow::buildShadow(Item *item) const
{
    return new ViewShadow(item, const_cast<ViewShadow *>(this), view);
}

void ViewShadow::updateItem()
{
    view.invalidate();
}

void ViewShadow::updateItems(unsigned int index, unsigned int old_count,
						 unsigned int new_count)
{
    view.requestLayout();
}

ViewShadow *ViewShadow::parent() const
{
    return static_cast<ViewShadow *>(Shadow::parent());
}

ViewShadow *ViewShadow::shadow(unsigned int index) const
{
    return static_cast<ViewShadow *>(Shadow::shadow(index));
}

ViewShadow *ViewShadow::last() const
{
    return count() == 0 ? 0 : shadow(count() - 1);
}

ViewShadow *ViewShadow::find(int x, int y, RegionType region)
{
    if ((region == SensitiveRegion ? _layout.sensitiveRegion :
	 region == PlupRegion ? _layout.plupRegion :
	 region == EditorRegion ? _layout.editorRegion :
	 Region()).contains(x, y))
	return this;

    if (is_expanded)
	for (unsigned int i = 0; i < count(); i++) {
	    ViewShadow *res = shadow(i)->find(x, y, region);
	    if (res)
		return res;
	}

    return 0;
}

ViewShadow *ViewShadow::nth(unsigned int index) const
{
    for (unsigned int i = 0; i < count(); i++) {
	ViewShadow *sh = shadow(i);

	if (index == 0)
	    return sh;
	index--;

	if (index < sh->visibleChildrenCount())
	    return sh->nth(index);
	index -= sh->visibleChildrenCount();
    }

    return 0;
}

ViewShadow *ViewShadow::next(ViewShadow *sh)
{
    if (sh) {
	unsigned int i = indexOf(sh);
	if (i == NPOS)
	    return 0;
	if (i + 1 < count())
	    return shadow(i + 1);
    } else if (is_expanded && count() != 0)
	return shadow(0);

    return parent() ? parent()->next(this) : 0;
}

ViewShadow *ViewShadow::previous(ViewShadow *sh)
{
    if (!sh)
	return parent() ? parent()->previous(this) : 0;

    unsigned int i = indexOf(sh);
    if (i == NPOS)
	return 0;
    if (i != 0)
	return shadow(i - 1)->lastDescendant();
    return parent() ? this : 0;
}

ViewShadow *ViewShadow::lastDescendant()
{
    if (count() == 0 || !is_expanded)
	return this;

    return last()->lastDescendant();
}

bool ViewShadow::isExpanded() const
{
    return is_expanded;
}

void ViewShadow::setExpanded(bool expanded)
{
    if (!parent() || expanded == is_expanded)
	return;

    is_expanded = expanded;
    view.requestLayout();
}

unsigned int ViewShadow::anchorX() const
{
    return anchor_x;
}

unsigned int ViewShadow::anchorY() const
{
    return anchor_y;
}

const ViewShadowLayout &ViewShadow::layout() const
{
    return _layout;
}

unsigned int ViewShadow::visibleChildrenCount() const
{
    return visible_children_count;
}

unsigned int ViewShadow::visibleChildrenHeight() const
{
    return visible_children_height;
}

void ViewShadow::relayout(unsigned int x, unsigned int y)
{
    anchor_x = x;
    anchor_y = y;
    _layout = parent() ? view.layoutShadow(this) : ViewShadowLayout();
    visible_children_count = 0;
    visible_children_height = 0;

    if (!is_expanded)
	return;

    for (unsigned int i = 0; i < count(); i++) {
	ViewShadow *sh = shadow(i);
	sh->relayout(anchor_x + _layout.indent,
		     anchor_y + _layout.height + visible_children_height);
	visible_children_count += sh->visibleChildrenCount() + 1;
	visible_children_height += sh->layout().height +
				   sh->visibleChildrenHeight();
    }
}

void ViewShadow::paint(int origin_x, int origin_y, X::Pixmap pixmap)
{
    if (parent())
	view.paintShadow(this, origin_x, origin_y, pixmap);

    if (is_expanded)
	for (unsigned int i = 0; i < count(); i++)
	    shadow(i)->paint(origin_x, origin_y, pixmap);
}

/****************************************************************************/

class View::Scroller : public WidgetScroller
{
    View &view;
public:
    Scroller(View &view) : WidgetScroller(view), view(view) {
    }
    int lineSize() const {
	return 1;
    }
    void setRange(int range) {
	WidgetScroller::setRange(range);
    }
    void setPageSize(int pageSize) {
	WidgetScroller::setPageSize(pageSize);
    }
private:
    void updatePosition() {
	view.scroll_position_changed();
	view.invalidate();
    }
};

/****************************************************************************/

View::View(int padding_x, int padding_y, Connection &conn)
    : Widget(conn, 0, vert_scroller = new Scroller(*this)),
      root_interior(0), root_shadow(0), highlighted_shadow(0),
      origin_x(0), origin_y(0), padding_x(padding_x), padding_y(padding_y)
{
    SizeRequirements reqs;
    reqs.minWidth = padding_x * 2;
    reqs.minHeight = padding_y * 2;
    setSizeRequirements(reqs);
}

View::~View()
{
    delete root_shadow;
    delete vert_scroller;
}

ItemInterior *View::rootInterior() const
{
    return root_interior;
}

void View::setRootInterior(ItemInterior *interior)
{
    if (interior == root_interior)
	return;

    root_interior = interior;

    delete root_shadow;
    root_shadow = interior ? new ViewShadow(interior, *this) : 0;

    requestLayout();
}

ViewShadow *View::rootShadow() const
{
    return root_shadow;
}

ViewShadow *View::highlightedShadow() const
{
    return highlighted_shadow;
}

void View::scrollTo(ViewShadow *shadow)
{
    int index = -1;
    ViewShadow *parent = shadow->parent();

    do {
	for (unsigned int i = 0; i < parent->count(); i++) {
	    ViewShadow *s = parent->shadow(i);
	    if (s == shadow)
		break;
	    index += s->visibleChildrenCount() + 1;
	}
	index += 1;
	shadow = parent;
	shadow->setExpanded(true);
	parent = shadow->parent();
    } while (parent);

    if (index >= vertScroller()->position() + vertScroller()->pageSize())
	vertScroller()->setPosition(index - vertScroller()->pageSize() + 1);
    if (vertScroller()->position() > index)
	vertScroller()->setPosition(index);
}

void View::updateLayout()
{
    /* structure or dimensions have changed */

    if (root_shadow)
	root_shadow->relayout(0, 0);

    if (!root_shadow || root_shadow->visibleChildrenCount() == 0)
	vert_scroller->setRange(0);
    else {
	vert_scroller->setRange(root_shadow->visibleChildrenCount());
	vert_scroller->setPageSize((height() - padding_y * 2)
				   * root_shadow->visibleChildrenCount()
				   / root_shadow->visibleChildrenHeight());
    }

    scroll_position_changed();
}

void View::scroll_position_changed()
{
    origin_x = padding_x;
    origin_y = padding_y;

    if (root_shadow && root_shadow->visibleChildrenCount() != 0)
	origin_y -= root_shadow->nth(vertScroller()->position())->anchorY();
}

ViewShadow *View::findShadow(int x, int y, ViewShadow::RegionType region)
{
    return root_shadow ? root_shadow->find(
	x - origin_x - this->x(),
	y - origin_y - this->y(), region) : 0;
}

void View::paint(const Region &region, X::Pixmap pixmap)
{
    int x = this->x() - region.x;
    int y = this->y() - region.y;

    gc().setForeground(colors().viewBackground.pixel);
    pixmap.fillRectangle(gc(), x, y, width(), height());

    if (!root_shadow || root_shadow->count() == 0)
	return;

    XRectangle rectangle;
    rectangle.x = x;
    rectangle.y = y;
    rectangle.width = width();
    rectangle.height = height();

    gc().setClipRectangles(0, 0, &rectangle, 1, Unsorted);
    root_shadow->paint(x + origin_x, y + origin_y, pixmap);
    gc().setClipMask(None);
}

void View::enterEvent(Time time, int x, int y, int x_root, int y_root,
		      unsigned int state/*, char is_hint*/)
{
    ViewShadow *shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (shadow == highlighted_shadow)
	return;

    highlighted_shadow = shadow;
    highlightedShadowChanged();
    invalidate();
}

void View::leaveEvent(Time time, int x, int y, int x_root, int y_root,
		      unsigned int state/*, char is_hint*/)
{
    if (!highlighted_shadow)
	return;

    highlighted_shadow = 0;
    highlightedShadowChanged();
    invalidate();
}

void View::motionEvent(Time time, int x, int y, int x_root, int y_root,
		       unsigned int state, char is_hint)
{
    ViewShadow *shadow = findShadow(x, y, ViewShadow::SensitiveRegion);
    if (shadow == highlighted_shadow)
	return;

    highlighted_shadow = shadow;
    highlightedShadowChanged();
    invalidate();
}

void View::buttonPressEvent(Time time, int x, int y, int x_root, int y_root,
			    unsigned int state, unsigned int button)
{
    if (button == Button4)
	vertScroller()->setPosition(
	    vertScroller()->position() -
	    vertScroller()->pageSize() * 120 / height());

    if (button == Button5)
	vertScroller()->setPosition(
	    vertScroller()->position() +
	    vertScroller()->pageSize() * 120 / height());
}

int View::originX() const
{
    return origin_x;
}

int View::originY() const
{
    return origin_y;
}

void View::highlightedShadowChanged()
{
    /* relax */
}
